// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.DaVinci
{
    public struct ColourXMLParameters
    {
        public string Format;
        public float Red;
        public float Green;
        public float Blue;
    }

    public class Colour
    {
        public enum RGBValueFormat
        {
            Byte8,
            UInt16,
            Float32
        }

        protected float _Red;
        protected float _Green;
        protected float _Blue;

        #region "Properties"

        public float Red
        {
            get
            {
                return _Red;
            }
            set
            {
                CheckColourValue("red", value);
                _Red = value;
            }
        }

        public float Green
        {
            get
            {
                return _Green;
            }
            set
            {
                CheckColourValue("green", value);
                _Green = value;
            }
        }

        public float Blue
        {
            get
            {
                return _Blue;
            }
            set
            {
                CheckColourValue("blue", value);
                _Blue = value;
            }
        }

        #endregion

        public Colour()
        {
            _Red = 0.0f;
            _Green = 0.0f;
            _Blue = 0.0f;
        }

        public Colour(float red, float green, float blue)
            : this()
        {
            Red = red;
            Green = green;
            Blue = blue;
        }

        public static Colour From16BitRGB(int red, int green, int blue)
        {
            return new Colour(((float)red) / 65535.0f, ((float)green) / 65535.0f, ((float)blue) / 65535.0f);
        }

        public static Colour From8BitRGB(int red, int green, int blue)
        {
            return new Colour(((float)red) / 255.0f, ((float)green) / 255.0f, ((float)blue) / 255.0f);
        }

        protected void CheckColourValue(string channelName, float newValue)
        {
            if ((newValue < 0.0f) || (newValue > 1.0f))
            {
                throw new ArgumentOutOfRangeException("The value for the " + channelName + " channel must be greater than zero but less than one.");
            }
        }

        public static RGBValueFormat RGBValueFormatFromString(string formatString)
        {
            string fString = formatString.Trim().ToLower();

            switch (fString)
            {
                case "byte8":
                    return RGBValueFormat.Byte8;
                    break;
                case "unsignedinteger16":
                    return RGBValueFormat.UInt16;
                    break;
                case "float32":
                    return RGBValueFormat.Float32;
                    break;
            }

            throw new ArgumentException("The string '" + formatString + "' could not be converted to a recognized format for representing RGB values.");
        }

        public static Colour FromXMLParameters(ColourXMLParameters xmlParams)
        {
            try
            {
                Colour.RGBValueFormat vFormat = RGBValueFormatFromString(xmlParams.Format);
                switch (vFormat)
                {
                    case Colour.RGBValueFormat.Byte8:
                        return From8BitRGB((int)xmlParams.Red, (int)xmlParams.Green, (int)xmlParams.Blue);
                        break;
                    case Colour.RGBValueFormat.UInt16:
                        return From16BitRGB((int)xmlParams.Red, (int)xmlParams.Green, (int)xmlParams.Blue);
                        break;
                    case Colour.RGBValueFormat.Float32:
                        return new Colour(xmlParams.Red, xmlParams.Green, xmlParams.Blue);
                        break;
                }
            }
            catch (ArgumentException argEx)
            {
                throw new ArgumentException("Could not create a colour from the passed XML parameters. Verify that the Format value is valid.", argEx);
            }
            catch (Exception ex)
            {
                throw new ArgumentException("Could not create a colour from the passed XML parameters.", ex);
            }
            throw new ArgumentException("An unexpected result occurred attempting to create a colour. Please verify that the XML definition conforms to the required specifications.");
        }
    }
}
