// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.DaVinci
{
    public enum CustomChannelReferenceType
    {
        None,
        BandByName,
        ChannelByName,
        SpecialBand
    }

    public class CustomChannelReference
    {
        protected CustomChannelReferenceType _ReferenceType;
        protected string _ReferenceName;
        protected string _ScriptTextToReplace;
        protected string _ScriptObjectName;
        protected const string _DefaultName = "_invalid_reference_name_";

        #region Properties

        public CustomChannelReferenceType ReferenceType
        {
            get
            {
                return _ReferenceType;
            }
            set
            {
                _ReferenceType = value;
                GenerateScriptObjectName();
            }
        }

        public string ReferenceName
        {
            get
            {
                return _ReferenceName;
            }
            set
            {
                _ReferenceName = value;
                GenerateScriptObjectName();
            }
        }

        public string ScriptTextToReplace
        {
            get
            {
                return _ScriptTextToReplace;
            }
            set
            {
                _ScriptTextToReplace = value;
            }
        }

        public string ScriptObjectName
        {
            get
            {
                return _ScriptObjectName;
            }
        }

        #endregion

        public CustomChannelReference()
        {
            _ReferenceType = CustomChannelReferenceType.None;
            _ReferenceName = _DefaultName;
            _ScriptTextToReplace = "";
            _ScriptObjectName = _DefaultName;
        }

        protected void GenerateScriptObjectName()
        {
            StringBuilder builder = new StringBuilder();
            switch (ReferenceType)
            {
                case CustomChannelReferenceType.BandByName:
                    builder.Append("bbn_");
                    break;
                case CustomChannelReferenceType.ChannelByName:
                    builder.Append("cbn_");
                    break;
                case CustomChannelReferenceType.SpecialBand:
                    builder.Append("spc_");
                    break;
            }
            Regex rex = new Regex("[^0-9A-Za-z]");
            string cName = rex.Replace(ReferenceName, "");
            builder.Append(cName);
            _ScriptObjectName = builder.ToString();
        }

        public void SetPropertiesFromReferenceString(string refString)
        {
            try
            {
                string rString = refString.Trim().Replace("%", "");
                string[] components = rString.Split(new char[] { ':' }, StringSplitOptions.RemoveEmptyEntries);
                ReferenceType = CustomChannelReferenceTypeFromString(components[0]);
                ReferenceName = components[1];
                ScriptTextToReplace = refString;
            }
            catch (Exception ex)
            {
                ReferenceType = CustomChannelReferenceType.None;
                ReferenceName = _DefaultName;
                throw new ArgumentException("The reference string '" + refString + "' could not be parsed.");
            }
        }

        public static CustomChannelReferenceType CustomChannelReferenceTypeFromString(string refType)
        {
            string rType = refType.Trim().ToLower();

            switch (rType)
            {
                case "bandbyname":
                    return CustomChannelReferenceType.BandByName;
                    break;
                case "channelbyname":
                    return CustomChannelReferenceType.ChannelByName;
                    break;
                case "specialband":
                    return CustomChannelReferenceType.SpecialBand;
                    break;
            }
            throw new ArgumentException("Could not convert the string '" + refType + "' to a recognized custom channel reference type.");
        }

        public string GetChannelName(ImageCube cube)
        {
            switch (ReferenceType)
            {
                case CustomChannelReferenceType.BandByName:
                    if (cube.ContainsBand(ReferenceName))
                    {
                        return cube.GetSpectralBand(ReferenceName).BandName;
                    }
                    else
                    {
                        throw new ArgumentException("Could not find a spectral band named '" + ReferenceName + "' in the image cube.");
                    }
                    break;
                case CustomChannelReferenceType.ChannelByName:
                    if (cube.ContainsGreyscaleChannel(ReferenceName))
                    {
                        return cube.GetGreyscaleChannel(ReferenceName).ImageName;
                    }
                    else
                    {
                        throw new ArgumentException("Could not find a channel named '" + ReferenceName + "' in the image cube.");
                    }
                    break;
                case CustomChannelReferenceType.SpecialBand:
                    SpecialBandName sbName = SpectralBand.SpecialBandNameFromString(ReferenceName);
                    if (cube.InputConfig.ContainsSpecialSpectralBand(sbName))
                    {
                        return cube.InputConfig.GetSpecialSpectralBand(sbName).BandName;
                    }
                    break;
            }
            throw new ArgumentException("Unable to convert the reference to '" + ReferenceName + "' into an object that exists in the image cube.");
        }
    }
}
