// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace BenLincoln.DaVinci
{
    public enum DCSMode
    {
        Standard,
        Running
    }

    public enum DCSColourSpace
    {
        RGB,
        YUV,
        HSY,
        HSV,
        HSL,
        IHSL
    }

    public delegate void NameRelatedValueChangedHandler(object sender, NameRelatedValueChangedArgs e);
    public class DCSParameterSet
    {
        public const float DefaultRDCSSectionPercent = 1.0f;
        public const float DefaultIgnore = -32768.0f;
        public static float DefaultVariance
        {
            get
            {
                return 50.0f / 255.0f;
            }
        }

        public event NameRelatedValueChangedHandler NameRelatedValueChanged;
        protected DCSMode _Mode;
        protected DCSColourSpace _ColourSpace;
        protected float _RDCSSectionPercent;
        protected float _Ignore;
        protected float _Variance;

        protected bool _UseInStandardLumaColourImages;
        protected bool _UseInDCSEnhancedContrast;

        #region Properties

        public DCSMode Mode
        {
            get
            {
                return _Mode;
            }
            set
            {
                _Mode = value;
            }
        }

        public DCSColourSpace ColourSpace
        {
            get
            {
                return _ColourSpace;
            }
            set
            {
                _ColourSpace = value;
            }
        }

        public float RDCSSectionPercent
        {
            get
            {
                return _RDCSSectionPercent;
            }
            set
            {
                // prevent values outside the logical range from being used
                float newPercent = value;
                if (newPercent <= 0)
                {
                    newPercent = 0.01f;
                }
                if (newPercent > 1)
                {
                    newPercent = 1;
                }
                _RDCSSectionPercent = newPercent;
                OnNameRelatedValueChanged(new NameRelatedValueChangedArgs());
            }
        }

        public float Ignore
        {
            get
            {
                return _Ignore;
            }
            set
            {
                _Ignore = value;
                OnNameRelatedValueChanged(new NameRelatedValueChangedArgs());
            }
        }

        public float Variance
        {
            get
            {
                return _Variance;
            }
            set
            {
                _Variance = value;
                OnNameRelatedValueChanged(new NameRelatedValueChangedArgs());
            }
        }

        public bool UseInStandardLumaColourImages
        {
            get
            {
                return _UseInStandardLumaColourImages;
            }
            set
            {
                _UseInStandardLumaColourImages = value;
            }
        }

        public bool UseInDCSEnhancedContrast
        {
            get
            {
                return _UseInDCSEnhancedContrast;
            }
            set
            {
                _UseInDCSEnhancedContrast = value;
            }
        }

        #endregion

        public DCSParameterSet(DCSMode mode, DCSColourSpace colourSpace, float rDCSSectionPercent, float ignore, float variance)
        {
            _Mode = mode;
            _ColourSpace = colourSpace;
            _RDCSSectionPercent = rDCSSectionPercent;
            _Ignore = ignore;
            _Variance = variance;
            _UseInStandardLumaColourImages = false;
            _UseInDCSEnhancedContrast = false;
        }

        public DCSParameterSet()
            : this(DCSMode.Standard, DCSColourSpace.RGB, DefaultRDCSSectionPercent, DefaultIgnore, DefaultVariance)
        {
        }

        protected virtual void OnNameRelatedValueChanged(NameRelatedValueChangedArgs e)
        {
            if (NameRelatedValueChanged != null)
            {
                NameRelatedValueChanged(this, e);
            }
        }

        public static DCSMode DCSModeFromString(string modeName)
        {
            switch (modeName)
            {
                case "Standard":
                    return DCSMode.Standard;
                    break;
                case "Running":
                    return DCSMode.Running;
                    break;
            }
            throw new InvalidDataException("The string '" + modeName + "' could not be converted to a DCS mode.");
        }

        public static DCSColourSpace DCSColourSpaceFromString(string csName)
        {
            switch (csName)
            {
                case "RGB":
                    return DCSColourSpace.RGB;
                    break;
                case "YUV":
                    return DCSColourSpace.YUV;
                    break;
                case "HSY":
                    return DCSColourSpace.HSY;
                    break;
                case "HSV":
                    return DCSColourSpace.HSV;
                    break;
                case "HSL":
                    return DCSColourSpace.HSL;
                    break;
                case "IHSL":
                    return DCSColourSpace.IHSL;
                    break;
            }
            throw new InvalidDataException("The string '" + csName + "' could not be converted to a DCS colourspace.");
        }
    }
}
