// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.DaVinci
{
    /// <summary>
    /// Contains each band-combining-mode variation of a particular set of greyscale channels.
    /// This class is used so that when creating each of the variations of a particular set,
    /// the program only needs to tell DaVinci to calculate the statistical data once, instead
    /// of for each variation.
    /// </summary>
    public class GreyscaleChannelSet
    {
        protected ImageCube _ParentCube;
        protected string _ChannelSetName;
        protected SpectralBand[] _SpectralBands;
        protected GreyscaleChannel[] _Channels;
        protected BandCombineMethod _BandCombineMethods;

        protected int _FileCountWillBeWritten;
        protected int _FileCountWillBeSkipped;

        protected int _MomentCubeWindowSize;

        protected bool _UseAsThreeChannelChannels;

        protected TransformationProfileCollection _InputTransformations;
        protected TransformationProfileCollection _OutputTransformations;

        #region Properties

        public ImageCube ParentCube
        {
            get
            {
                return _ParentCube;
            }
            set
            {
                _ParentCube = value;
            }
        }

        public string ChannelSetName
        {
            get
            {
                return _ChannelSetName;
            }
            set
            {
                _ChannelSetName = value;
            }
        }

        public SpectralBand[] SpectralBands
        {
            get
            {
                return _SpectralBands;
            }
            set
            {
                _SpectralBands = value;
            }
        }

        public GreyscaleChannel[] Channels
        {
            get
            {
                return _Channels;
            }
            set
            {
                _Channels = value;
            }
        }

        public BandCombineMethod BandCombineMethods
        {
            get
            {
                return _BandCombineMethods;
            }
            set
            {
                _BandCombineMethods = value;
            }
        }

        public int FileCountWillBeWritten
        {
            get
            {
                return _FileCountWillBeWritten;
            }
        }

        public int FileCountWillBeSkipped
        {
            get
            {
                return _FileCountWillBeSkipped;
            }
        }

        public int MomentCubeWindowSize
        {
            get
            {
                return _MomentCubeWindowSize;
            }
            set
            {
                _MomentCubeWindowSize = value;
            }
        }

        public bool UseAsThreeChannelChannels
        {
            get
            {
                return _UseAsThreeChannelChannels;
            }
            set
            {
                _UseAsThreeChannelChannels = value;
            }
        }

        public TransformationProfileCollection InputTransformations
        {
            get
            {
                return _InputTransformations;
            }
            set
            {
                _InputTransformations = value;
            }
        }

        public TransformationProfileCollection OutputTransformations
        {
            get
            {
                return _OutputTransformations;
            }
            set
            {
                _OutputTransformations = value;
            }
        }


        #endregion

        public GreyscaleChannelSet(ImageCube parentCube)
        {
            if (parentCube == null)
            {
                throw new NullReferenceException("The parent cube cannot be null.");
            }
            ParentCube = parentCube;
            _ChannelSetName = "";
            _SpectralBands = new SpectralBand[0];
            _Channels = new GreyscaleChannel[0];
            _BandCombineMethods = BandCombineMethod.None;
            _FileCountWillBeWritten = 0;
            _FileCountWillBeSkipped = 0;
            _MomentCubeWindowSize = 1;
            UseAsThreeChannelChannels = true;
            InputTransformations = new TransformationProfileCollection();
            OutputTransformations = new TransformationProfileCollection();
        }

        public static bool BandCombineMethodEnabled(BandCombineMethod methods, BandCombineMethod method)
        {
            if ((method & methods) == method)
            {
                return true;
            }
            return false;
        }

        public bool UsesBandCombineMethod(BandCombineMethod method)
        {
            return BandCombineMethodEnabled(BandCombineMethods, method);
        }

        public void AddBandCombineMethod(BandCombineMethod method)
        {
            BandCombineMethods = BandCombineMethods | method;
        }

        public void RemoveBandCombineMethod(BandCombineMethod method)
        {
            BandCombineMethods = BandCombineMethods ^ method;
        }

        public bool IsEmpty()
        {
            if ((SpectralBands == null) || (SpectralBands.Length == 0))
            {
                return true;
            }
            return false;
        }

        public bool BandsMatchChannel(GreyscaleChannel channel)
        {
            if (SpectralBands.Length != channel.SpectralBands.Length)
            {
                return false;
            }
            // does the channel have all bands present in this set?
            foreach (SpectralBand band in SpectralBands)
            {
                bool channelHasBand = false;
                foreach (SpectralBand cBand in channel.SpectralBands)
                {
                    if (band.BandName == cBand.BandName)
                    {
                        channelHasBand = true;
                        break;
                    }
                }
                if (!channelHasBand)
                {
                    return false;
                }
            }

            return true;
        }

        public void AddChannel(GreyscaleChannel newChannel)
        {
            if ((Channels == null) || (Channels.Length == 0))
            {
                Channels = new GreyscaleChannel[1];
                SpectralBands = newChannel.SpectralBands;
            }
            else
            {
                if (!BandsMatchChannel(newChannel))
                {
                    throw new ArgumentException("The channel '" + newChannel + "' is based on a different set of spectral bands than the collection to which an attempt was made to add it.");
                }
                Array.Resize(ref _Channels, _Channels.Length + 1);
            }
            newChannel.ParentCube = ParentCube;
            Channels[Channels.Length - 1] = newChannel;
            AddBandCombineMethod(newChannel.BandCombineMode);
        }

        public bool ContainsChannel(string channelName)
        {
            for (int i = 0; i < _Channels.Length; i++)
            {
                if ((_Channels[i] != null) && (_Channels[i].ImageName == channelName))
                {
                    return true;
                }
            }
            return false;
        }

        public string[] GetChannelNames()
        {
            ArrayList names = new ArrayList();
            for (int i = 0; i < _Channels.Length; i++)
            {
                if (_Channels[i] != null)
                {
                    names.Add(_Channels[i].ImageName);
                }
            }
            string[] proto = new string[] { "1", "2" };
            return (string[])names.ToArray(proto.GetType());
        }

        public string GetSetCreationScriptCode(bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            string baseCubeScriptObjectName = "set_base_cube";
            string momentCubeScriptObjectName = "set_moment_cube";
            string scriptObjectName = "gscs_temp";

            // To do: prevent calculating statistics if all images based on those statistics already exist and shouldn't be overwritten

            _FileCountWillBeWritten = 0;
            _FileCountWillBeSkipped = 0;

            if (Channels.Length == 1)
            {
                builder.Append(Channels[0].GetCreationScriptCode(scriptObjectName, regenerateExistingFiles));
                _FileCountWillBeWritten = Channels[0].FileCountWillBeWritten;
                _FileCountWillBeSkipped = Channels[0].FileCountWillBeSkipped;
            }
            else
            {
                bool createdMomentCube = false;
                BandCombineMethod allSetBandCombineMethods = BandCombineMethod.None;

                for (int channelNumber = 0; channelNumber < Channels.Length; channelNumber++)
                {
                    Channels[channelNumber].GetCreationScriptCode("not_used", regenerateExistingFiles);
                    _FileCountWillBeWritten += Channels[channelNumber].FileCountWillBeWritten;
                    //_FileCountWillBeSkipped += Channels[channelNumber].FileCountWillBeSkipped;
                    _FileCountWillBeSkipped += (Channels[channelNumber].OutputFormats.Length - Channels[channelNumber].FileCountWillBeWritten);
                    if (Channels[channelNumber].FileCountWillBeWritten > 0)
                    {
                        allSetBandCombineMethods = allSetBandCombineMethods | Channels[channelNumber].BandCombineMode;
                    }
                }
                if (_FileCountWillBeWritten == 0)
                {
                    builder.Append("# All output files related to this set already exist, and the option to overwrite existing files is not enabled.");
                    builder.Append(Environment.NewLine);
                }
                else
                {
                    _FileCountWillBeWritten = 0;
                    _FileCountWillBeSkipped = 0;
                    for (int channelNumber = 0; channelNumber < Channels.Length; channelNumber++)
                    {
                        if (((Channels[channelNumber].BandCombineMode & BandCombineMethod.Custom) == BandCombineMethod.Custom) &&
                            (Channels[channelNumber].CustomBandCombinationScriptCode != ""))
                        {
                            builder.Append(Channels[channelNumber].GetCreationScriptCode(scriptObjectName, regenerateExistingFiles));
                            builder.Append(Channels[channelNumber].GetBandObjectDeallocationCode());
                        }
                        else
                        {
                            if (channelNumber == 0)
                            {
                                builder.Append(Channels[channelNumber].GetBandCubeScriptCode(baseCubeScriptObjectName));
                                if (GreyscaleChannel.RequiresMomentCubeCreation(BandCombineMethods))
                                {
                                    builder.Append(Channels[channelNumber].GetMomentCubeScriptCode(momentCubeScriptObjectName, baseCubeScriptObjectName, allSetBandCombineMethods, 
                                        MomentCubeWindowSize, InputTransformations, OutputTransformations));
                                    createdMomentCube = true;
                                }
                            }

                            builder.Append(Channels[channelNumber].GetMultiFileCreationScriptCode(scriptObjectName, regenerateExistingFiles, false, baseCubeScriptObjectName,
                                BandCombineMethods, false, momentCubeScriptObjectName, MomentCubeWindowSize, InputTransformations, OutputTransformations));
                        
                            if (channelNumber == (Channels.Length - 1))
                            {
                                builder.Append(OutputImage.GetDeallocationScriptCode(baseCubeScriptObjectName));
                                if (createdMomentCube)
                                {
                                    builder.Append(OutputImage.GetDeallocationScriptCode(momentCubeScriptObjectName));
                                }
                                builder.Append(Channels[channelNumber].GetBandObjectDeallocationCode());
                            }
                        }

                        _FileCountWillBeWritten += Channels[channelNumber].FileCountWillBeWritten;
                        //_FileCountWillBeSkipped += Channels[channelNumber].FileCountWillBeSkipped;
                        _FileCountWillBeSkipped += (Channels[channelNumber].OutputFormats.Length - Channels[channelNumber].FileCountWillBeWritten);
                    }
                }
                //for (int channelNumber = 0; channelNumber < Channels.Length; channelNumber++)
                //{
                    
                //}
            }

            builder.Append("# Done creating variations of this set.");
            builder.Append(Environment.NewLine);

            return OutputImage.ExecuteGlobalScriptPostProcessing(ParentCube, builder.ToString());
        }

        public bool ContainsChannelWithBandCombineMethod(BandCombineMethod method)
        {
            foreach (GreyscaleChannel channel in Channels)
            {
                if (channel.BandCombineMode == method)
                {
                    return true;
                }
            }
            return false;
        }

        public GreyscaleChannel GetChannelByBandCombineMethod(BandCombineMethod method)
        {
            foreach (GreyscaleChannel channel in Channels)
            {
                if (channel.BandCombineMode == method)
                {
                    return channel;
                }
            }
            throw new Exception("Could not find a channel in the set with the BandCombineMethod '" + method.ToString() + "'.");
        }
    }
}
