// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.DaVinci
{
    public class LumaColourImage : MultiChannelImage
    {
        protected string _LumaName;
        protected string _ColourName;
        protected string _LumaObjectName;
        protected string _ColourObjectName;
        //protected string _TintingColourName;
        protected string _LumaFilePath;
        protected string _ColourFilePath;
        protected OutputImage _LumaChannel;
        protected TintingColour _TintingColour;

        // the script code used to generate the colourized combination image
        //protected const string _GenerateScriptCode =
        //    " = hsy2rgb(blendcolour(rgb2hsy(tofloat(%LUMA%)),rgb2hsy(tofloat(%COLOUR%))))";
        protected const string _GenerateScriptCode =
            " = blendcolourrgb(%LUMA%,%COLOUR%)";

        #region Properties

        public string LumaName
        {
            get
            {
                return _LumaName;
            }
            set
            {
                _LumaName = value;
                GenerateName();
            }
        }

        public string ColourName
        {
            get
            {
                return _ColourName;
            }
            set
            {
                _ColourName = value;
                GenerateName();
            }
        }

        public string LumaObjectName
        {
            get
            {
                return _LumaObjectName;
            }
            set
            {
                _LumaObjectName = value;
                GenerateName();
            }
        }

        public string ColourObjectName
        {
            get
            {
                return _ColourObjectName;
            }
            set
            {
                _ColourObjectName = value;
                GenerateName();
            }
        }

        //public string TintingColourName
        //{
        //    get
        //    {
        //        return _TintingColourName;
        //    }
        //    set
        //    {
        //        _TintingColourName = value;
        //        _ColourName = _TintingColourName;
        //        GenerateName();
        //    }
        //}

        public TintingColour TintingColour
        {
            get
            {
                return _TintingColour;
            }
            set
            {
                _TintingColour = value;
                _ColourName = _TintingColour.TintingColourName;
                GenerateName();
            }
        }

        public string LumaFilePath
        {
            get
            {
                return _LumaFilePath;
            }
            set
            {
                _LumaFilePath = value;
                if (value != null)
                {
                    _LumaName = Path.GetFileNameWithoutExtension(_LumaFilePath);
                }
                else
                {
                    _LumaName = "Null";
                }
                GenerateName();
            }
        }

        public string ColourFilePath
        {
            get
            {
                return _ColourFilePath;
            }
            set
            {
                _ColourFilePath = value;
                if (value != null)
                {
                    _ColourName = Path.GetFileNameWithoutExtension(_ColourFilePath);
                }
                else
                {
                    _ColourName = "Null";
                }
                GenerateName();
            }
        }

        public OutputImage LumaChannel
        {
            get
            {
                return _LumaChannel;
            }
            set
            {
                _LumaChannel = value;
                if (value != null)
                {
                    if (_LumaChannel is GreyscaleChannel)
                    {
                        _Channels = new GreyscaleChannel[1];
                        _Channels[0] = (GreyscaleChannel)_LumaChannel;
                    }
                    else if (_LumaChannel is MultiChannelImage)
                    {
                        MultiChannelImage mci = (MultiChannelImage)_LumaChannel;
                        _Channels = mci.Channels;
                    }
                    _LumaName = _LumaChannel.ImageName;
                }
                else
                {
                    _LumaName = "Null";
                }
            }
        }

        #endregion

        public LumaColourImage(ImageCube parentCube)
            : base(parentCube)
        {
            _NamePrefix = "FC-LC-";
            //_NamePrefix = "FC-";
            _ImageName = "";
            _LumaName = "";
            _ColourName = "";
            _LumaObjectName = "";
            _ColourObjectName = "";
            //_TintingColourName = "";
            _LumaFilePath = "";
            _ColourFilePath = "";
        }

        public LumaColourImage(ImageCube parentCube, string lumaName, string colourName)
            : this(parentCube)
        {
            if ((lumaName == null) || (colourName == null) || (lumaName.Trim() == "") || (colourName.Trim() == ""))
            {
                throw new ArgumentException("The names passed to this constructor cannot be empty or null.");
            }
            _LumaName = lumaName;
            _ColourName = colourName;
            GenerateName();
        }

        protected override void _GenerateName()
        {
            StringBuilder nameBuilder = new StringBuilder(_NamePrefix);

            string lName = _LumaName.Trim().Replace(" ", "_");
            string cName = _ColourName.Trim().Replace(" ", "_");

            nameBuilder.Append(WrapNameComponentIfComplex(lName));
            nameBuilder.Append("_Luminance");
            nameBuilder.Append(_NameComponentSeparator);
            nameBuilder.Append(WrapNameComponentIfComplex(cName));
            nameBuilder.Append("_Colour");
            nameBuilder.Append(_NameSuffix);

            _ImageName = nameBuilder.ToString();
        }

        public override string GetCreationScriptCode(string scriptTargetObjectName, bool regenerateExistingFiles)
        {
            string filePath = GetFilePathWithoutExtension();
            StringBuilder builder = new StringBuilder();
            // if the option to recreate existing images is not enabled and the 
            // file already exists, do nothing.
            if ((!regenerateExistingFiles) && AllOutputFilesExist(filePath, OutputFormats))
            {
                builder.Append("# Information: all file-type variations of the luminance/colour false colour image '");
                builder.Append(_ImageName);
                builder.Append("' already appear to exist. The option to regenerate existing files is not enabled, so the script will not overwrite them.");
                builder.Append(Environment.NewLine);
                return builder.ToString();
            }
            _WillBeCreatedByScript = true;

            // You are in a maze of twisty passages, all different.
            // (This section - and the methods it calls - all need to be simplified/cleaned up
            // it works, but is inelegant and hard to maintain).

            if (_LumaFilePath != "")
            {
                if (_ColourFilePath != "")
                {
                    return ExecuteGlobalScriptPostProcessing(ParentCube, GetCreationScriptCodeFromLumaFileAndColourFile(scriptTargetObjectName, _LumaFilePath, _ColourFilePath, regenerateExistingFiles));
                }
                if (_ColourObjectName != "")
                {
                    return ExecuteGlobalScriptPostProcessing(ParentCube, GetCreationScriptCodeFromLumaFile(scriptTargetObjectName, _LumaFilePath, _ColourObjectName, regenerateExistingFiles));
                }
            }

            if (_LumaChannel != null)
            {
                if (_ColourFilePath != "")
                {
                    if (_LumaChannel is GreyscaleChannel)
                    {
                        return ExecuteGlobalScriptPostProcessing(ParentCube, GetCreationScriptCodeFromGreyscaleChannelAndColourFile(scriptTargetObjectName, (GreyscaleChannel)_LumaChannel, _ColourFilePath, regenerateExistingFiles));
                    }
                    else
                    {
                        return ExecuteGlobalScriptPostProcessing(ParentCube, GetCreationScriptCodeFromLumaChannelAndColourFile(scriptTargetObjectName, _LumaChannel, _ColourFilePath, regenerateExistingFiles));
                    }
                }
                //if (_TintingColourName != "")
                //{
                //    if (_LumaChannel is GreyscaleChannel)
                //    {
                //        return GetCreationScriptCodeFromChannelAndTint(scriptTargetObjectName, (GreyscaleChannel)_LumaChannel, _TintingColourName, regenerateExistingFiles);
                //    }
                //    else
                //    {
                //        throw new Exception("This creation method can only be used if the luminance object is of the GreyscaleChannel class.");
                //    }
                //}
                if (_TintingColour != null)
                {
                    if (_LumaChannel is GreyscaleChannel)
                    {
                        return ExecuteGlobalScriptPostProcessing(ParentCube, GetCreationScriptCodeFromChannelAndTint(scriptTargetObjectName, (GreyscaleChannel)_LumaChannel, _TintingColour, regenerateExistingFiles));
                    }
                    else
                    {
                        throw new Exception("This creation method can only be used if the luminance object is of the GreyscaleChannel class.");
                    }
                }
                if (_ColourObjectName != "")
                {
                    return ExecuteGlobalScriptPostProcessing(ParentCube, GetCreationScriptCodeFromLumaChannel(scriptTargetObjectName, _LumaChannel, _ColourObjectName, regenerateExistingFiles));
                }
            }

            if ((_LumaObjectName != "") && (_ColourObjectName != ""))
            {
                return ExecuteGlobalScriptPostProcessing(ParentCube, GetCreationScriptCodeFromObjectNames(scriptTargetObjectName, _LumaObjectName, _ColourObjectName, regenerateExistingFiles));
            }

            throw new ArgumentException("This instance of the LumaColourImage class ('" + ImageName + "') does not have enough variables defined to generate it.");
        }

        public string GetCreationScriptCodeFromLumaChannelAndColourFile(string scriptTargetObjectName, OutputImage lumaChannel, string colourFilePath, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' luminance/colour image");
            builder.Append(Environment.NewLine);

            builder.Append("# Apply the colour of the file '");
            builder.Append(colourFilePath);
            builder.Append("' to the luminance of the image '");
            builder.Append(lumaChannel.GetName());
            builder.Append("'");
            builder.Append(Environment.NewLine);

            string lumaObjectName = _TemporaryObjectNamePrefix + "luma_047";
            builder.Append(lumaChannel.GetReadScriptCode(lumaObjectName));
            string colourObjectName = _TemporaryObjectNamePrefix + "colour_047";
            builder.Append(_GetReadScriptCode("", colourObjectName, colourFilePath));

            // create a three-plane greyscale image out of the luma channel
            //builder.Append(lumaObjectName);
            //builder.Append(" = threeclone(");
            //builder.Append(lumaObjectName);
            //builder.Append(")");
            //builder.Append(Environment.NewLine);

            builder.Append(GetCreationScriptCodeFromObjectNames(scriptTargetObjectName, lumaObjectName, colourObjectName, regenerateExistingFiles));

            // clear out the temporary objects
            builder.Append(GetDeallocationScriptCode(lumaObjectName));
            builder.Append(GetDeallocationScriptCode(colourObjectName));

            return builder.ToString();
        }

        public string GetCreationScriptCodeFromLumaChannel(string scriptTargetObjectName, OutputImage lumaChannel, string colourObjectName, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' luminance/colour image");
            builder.Append(Environment.NewLine);

            builder.Append("# Apply the colour of the object '");
            builder.Append(colourObjectName);
            builder.Append("' to the luminance of the image '");
            builder.Append(lumaChannel.GetName());
            builder.Append("'");
            builder.Append(Environment.NewLine);

            string lumaObjectName = _TemporaryObjectNamePrefix + "luma_043";
            builder.Append(lumaChannel.GetReadScriptCode(lumaObjectName));

            // create a three-plane greyscale image out of the luma channel
            //builder.Append(lumaObjectName);
            //builder.Append(" = threeclone(");
            //builder.Append(lumaObjectName);
            //builder.Append(")");
            //builder.Append(Environment.NewLine);

            builder.Append(GetCreationScriptCodeFromObjectNames(scriptTargetObjectName, lumaObjectName, colourObjectName, regenerateExistingFiles));

            // clear out the temporary objects
            builder.Append(GreyscaleChannel.GetDeallocationScriptCode(lumaObjectName));

            return builder.ToString();
        }

        public string GetCreationScriptCodeFromGreyscaleChannelAndColourFile(string scriptTargetObjectName, GreyscaleChannel lumaChannel, string colourFilePath, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' luminance/colour image");
            builder.Append(Environment.NewLine);

            builder.Append("# Apply the colour of the file '");
            builder.Append(colourFilePath);
            builder.Append("' to the luminance of the image '");
            builder.Append(lumaChannel.GetName());
            builder.Append("'");
            builder.Append(Environment.NewLine);

            // since this is a greyscale channel luminance object, tell it to use its secondary image if one is present
            bool lcUSIPreviousState = lumaChannel.UseSecondaryImage;

            if (lumaChannel.IsSingleSpectralBand)
            {
                if (lumaChannel.HasSecondarySourceImage)
                {
                    lumaChannel.UseSecondaryImage = true;

                    builder.Append("# The luminance object is an individual spectral band for which a secondary source image is present.");
                    builder.Append(Environment.NewLine);
                    builder.Append("# The secondary source image will be used.");
                    builder.Append(Environment.NewLine);
                }
                else
                {
                    builder.Append("# The luminance object is an individual spectral band, but no secondary source image was found.");
                    builder.Append(Environment.NewLine);
                    builder.Append("# The primary source image will be used.");
                    builder.Append(Environment.NewLine);
                }
            }
            else
            {
                //builder.Append("# The luminance object was generated from multiple spectral bands.");
                //builder.Append(Environment.NewLine);
            }

            string lumaObjectName = _TemporaryObjectNamePrefix + "luma_047";
            builder.Append(lumaChannel.GetReadScriptCode(lumaObjectName));
            string colourObjectName = _TemporaryObjectNamePrefix + "colour_047";
            builder.Append(_GetReadScriptCode("", colourObjectName, colourFilePath));

            // create a three-plane greyscale image out of the luma channel
            //builder.Append(lumaObjectName);
            //builder.Append(" = threeclone(");
            //builder.Append(lumaObjectName);
            //builder.Append(")");
            //builder.Append(Environment.NewLine);

            builder.Append(GetCreationScriptCodeFromObjectNames(scriptTargetObjectName, lumaObjectName, colourObjectName, regenerateExistingFiles));
            
            // clear out the temporary objects
            builder.Append(GetDeallocationScriptCode(lumaObjectName));
            builder.Append(GetDeallocationScriptCode(colourObjectName));

            lumaChannel.UseSecondaryImage = lcUSIPreviousState;

            return builder.ToString();
        }

        //public string GetCreationScriptCodeFromChannelAndTint(string scriptTargetObjectName, GreyscaleChannel lumaChannel, string tintingColourName, bool regenerateExistingFiles)
        //{
        //    StringBuilder builder = new StringBuilder();

        //    if (!lumaChannel.ContainsColour(tintingColourName))
        //    {
        //        throw new ArgumentException("The channel '" + lumaChannel.ImageName + "' does not contain a colour named '" + tintingColourName + "'.");
        //    }

        //    builder.Append("# Apply the colour of the tint '");
        //    builder.Append(tintingColourName);
        //    builder.Append("' to the luminance of the image '");
        //    builder.Append(lumaChannel.ImageName);
        //    builder.Append("'");
        //    builder.Append(Environment.NewLine);

        //    // since this is a greyscale channel luminance object, tell it to use its secondary image if one is present
        //    bool lcUSIPreviousState = lumaChannel.UseSecondaryImage;

        //    if (lumaChannel.IsSingleSpectralBand)
        //    {
        //        if (lumaChannel.HasSecondarySourceImage)
        //        {
        //            lumaChannel.UseSecondaryImage = true;

        //            builder.Append("# The luminance object is an individual spectral band for which a secondary source image is present.");
        //            builder.Append(Environment.NewLine);
        //            builder.Append("# The secondary source image will be used.");
        //            builder.Append(Environment.NewLine);
        //        }
        //        else
        //        {
        //            builder.Append("# The luminance object is an individual spectral band, but no secondary source image was found.");
        //            builder.Append(Environment.NewLine);
        //            builder.Append("# The primary source image will be used.");
        //            builder.Append(Environment.NewLine);
        //        }
        //    }
        //    else
        //    {
        //        //builder.Append("# The luminance object was generated from multiple spectral bands.");
        //        //builder.Append(Environment.NewLine);
        //    }

        //    string lumaObjectName = _TemporaryObjectNamePrefix + "luma_041";
        //    builder.Append(lumaChannel.GetReadScriptCode(lumaObjectName));
        //    string colourObjectName = _TemporaryObjectNamePrefix + "colour_041";
        //    builder.Append(lumaChannel.GetTintingColour(tintingColourName).GetCreationScriptCode(colourObjectName, lumaObjectName));

        //    // create a three-plane greyscale image out of the luma channel
        //    //builder.Append(lumaObjectName);
        //    //builder.Append(" = threeclone(");
        //    //builder.Append(lumaObjectName);
        //    //builder.Append(")");
        //    //builder.Append(Environment.NewLine);

        //    builder.Append(GetCreationScriptCodeFromObjectNames(scriptTargetObjectName, lumaObjectName, colourObjectName, regenerateExistingFiles));

        //    // clear out the temporary objects
        //    builder.Append(GetDeallocationScriptCode(lumaObjectName));
        //    builder.Append(GetDeallocationScriptCode(colourObjectName));

        //    lumaChannel.UseSecondaryImage = lcUSIPreviousState;

        //    return builder.ToString();
        //}

        public string GetCreationScriptCodeFromChannelAndTint(string scriptTargetObjectName, GreyscaleChannel lumaChannel, TintingColour tColour, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' tinted greyscale image");
            builder.Append(Environment.NewLine);

            builder.Append("# Apply the colour of the tint '");
            builder.Append(tColour.TintingColourName);
            builder.Append("' to the luminance of the image '");
            builder.Append(lumaChannel.ImageName);
            builder.Append("'");
            builder.Append(Environment.NewLine);

            // since this is a greyscale channel luminance object, tell it to use its secondary image if one is present
            bool lcUSIPreviousState = lumaChannel.UseSecondaryImage;

            if (lumaChannel.IsSingleSpectralBand)
            {
                if (lumaChannel.HasSecondarySourceImage)
                {
                    lumaChannel.UseSecondaryImage = true;

                    builder.Append("# The luminance object is an individual spectral band for which a secondary source image is present.");
                    builder.Append(Environment.NewLine);
                    builder.Append("# The secondary source image will be used.");
                    builder.Append(Environment.NewLine);
                }
                else
                {
                    builder.Append("# The luminance object is an individual spectral band, but no secondary source image was found.");
                    builder.Append(Environment.NewLine);
                    builder.Append("# The primary source image will be used.");
                    builder.Append(Environment.NewLine);
                }
            }
            else
            {
                //builder.Append("# The luminance object was generated from multiple spectral bands.");
                //builder.Append(Environment.NewLine);
            }

            string lumaObjectName = _TemporaryObjectNamePrefix + "luma_041";
            builder.Append(lumaChannel.GetReadScriptCode(lumaObjectName));
            string colourObjectName = _TemporaryObjectNamePrefix + "colour_041";
            builder.Append(tColour.GetCreationScriptCode(colourObjectName, lumaObjectName));

            // create a three-plane greyscale image out of the luma channel
            //builder.Append(lumaObjectName);
            //builder.Append(" = threeclone(");
            //builder.Append(lumaObjectName);
            //builder.Append(")");
            //builder.Append(Environment.NewLine);

            builder.Append(GetCreationScriptCodeFromObjectNames(scriptTargetObjectName, lumaObjectName, colourObjectName, regenerateExistingFiles));

            // clear out the temporary objects
            builder.Append(GetDeallocationScriptCode(lumaObjectName));
            builder.Append(GetDeallocationScriptCode(colourObjectName));

            lumaChannel.UseSecondaryImage = lcUSIPreviousState;

            return builder.ToString();
        }

        public string GetCreationScriptCodeFromLumaFile(string scriptTargetObjectName, string lumaFilePath, string colourObjectName, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' luminance/colour image");
            builder.Append(Environment.NewLine);

            builder.Append("# Apply the colour of the file '");
            builder.Append(colourObjectName);
            builder.Append("' to the luminance of the file '");
            builder.Append(lumaFilePath);
            builder.Append("'");
            builder.Append(Environment.NewLine);

            string lumaObjectName = _TemporaryObjectNamePrefix + "luma_039";
            builder.Append(_GetReadScriptCode("", lumaObjectName, lumaFilePath));

            builder.Append(GetCreationScriptCodeFromObjectNames(scriptTargetObjectName, lumaObjectName, colourObjectName, regenerateExistingFiles));

            // clear out the temporary objects
            builder.Append(GetDeallocationScriptCode(lumaObjectName));

            return builder.ToString();
        }

        public string GetCreationScriptCodeFromLumaFileAndColourFile(string scriptTargetObjectName, string lumaFilePath, string colourFilePath, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' luminance/colour image");
            builder.Append(Environment.NewLine);

            builder.Append("# Apply the colour of the file '");
            builder.Append(colourFilePath);
            builder.Append("' to the luminance of the file '");
            builder.Append(lumaFilePath);
            builder.Append("'");
            builder.Append(Environment.NewLine);

            string lumaObjectName = _TemporaryObjectNamePrefix + "luma_037";
            builder.Append(_GetReadScriptCode("", lumaObjectName, lumaFilePath));
            string colourObjectName = _TemporaryObjectNamePrefix + "colour_037";
            builder.Append(_GetReadScriptCode("", colourObjectName, colourFilePath));

            builder.Append(GetCreationScriptCodeFromObjectNames(scriptTargetObjectName, lumaObjectName, colourObjectName, regenerateExistingFiles));

            // clear out the temporary objects
            builder.Append(GetDeallocationScriptCode(lumaObjectName));
            builder.Append(GetDeallocationScriptCode(colourObjectName));

            return builder.ToString();
        }

        public string GetCreationScriptCodeFromObjectNames(string scriptTargetObjectName, string lumaObjectName, string colourObjectName, bool regenerateExistingFiles)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append("# Create the '");
            builder.Append(ImageName);
            builder.Append("' luminance/colour image");
            builder.Append(Environment.NewLine);

            builder.Append("# Apply the colour of the '");
            builder.Append(lumaObjectName);
            builder.Append("' object to the luminance of the '");
            builder.Append(colourObjectName);
            builder.Append("' object");
            builder.Append(Environment.NewLine);

            string filePath = GetFilePathWithoutExtension();
            if ((regenerateExistingFiles) && (AnyOutputFilesExist(filePath, OutputFormats)))
            {
                builder.Append("# Warning: one or more variations of the target file already exist, but will be overwritten as the option to regenerate existing files is enabled.");
                builder.Append(Environment.NewLine);
                _WillOverwriteExistingData = true;
            }


            string blendedObjectName = _TemporaryObjectNamePrefix + "colourblend";

            string blendCommand = _GenerateScriptCode.Replace("%LUMA%", lumaObjectName);
            blendCommand = blendCommand.Replace("%COLOUR%", colourObjectName);

            builder.Append(blendedObjectName);
            builder.Append(blendCommand);
            builder.Append(Environment.NewLine);

            // normalize the image if that option is enabled
            //builder.Append(WrapScriptObjectInNormalizationCode(blendedObjectName, OutputNormalization));
            //builder.Append(GetFullNormalizationWrappedLine(blendedObjectName, OutputNormalization));
            //builder.Append(OutputTransformations.GetWrappedLine(blendedObjectName));

            // write the image to disk
            builder.Append(GetImageFileWriteCode(blendedObjectName, OutputFormats, filePath, regenerateExistingFiles));
 
            // clear out the temporary objects
            builder.Append(GetDeallocationScriptCode(blendedObjectName));
            builder.Append(GetDeallocationScriptCode(lumaObjectName));

            return builder.ToString();
        }

        public override bool ContainsDisallowedChannels()
        {
            if (LumaChannel is GreyscaleChannel)
            {
                GreyscaleChannel lc = (GreyscaleChannel)LumaChannel;
                if (!lc.UseAsLumaBand())
                {
                    return true;
                }
            }

            //foreach (GreyscaleChannel channel in Channels)
            //{
            //    if ((channel.ChannelName != LumaChannel.) && (!channel.UseAsThreeChannelChannel))
            //    {
            //        return true;
            //    }
            //}
            return false;
        }
    }
}
