// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.DaVinci
{
    [Flags]
    public enum BandCategory
    {
        None = 0,
        Radio = 1,
        Microwave = 2,
        FarInfrared = 4,
        MidInfrared = 8,
        NearInfrared = 16,
        HumanVisible = 32,
        Ultraviolet = 64,
        XRay = 128,
        GammaRay = 256,
        Derived = 65536
    }

    [Flags]
    public enum SpecialBandName
    {
        None,
        RadioELF,
        RadioSLF,
        RadioULF,
        RadioVLF,
        RadioLF,
        RadioMF,
        RadioHF,
        RadioVHF,
        RadioUHF,
        RadioSHF,
        RadioEHF,
        RadioNonSpecific,
        Radar,
        MicrowavePBand,
        MicrowaveLBand,
        MicrowaveSBand,
        MicrowaveCBand,
        MicrowaveXBand,
        MicrowaveKBand,
        MicrowaveQBand,
        MicrowaveVBand,
        MicrowaveWBand,
        MicrowaveNonspecific,
        Terahertz,
        ThermalInfrared,
        FarInfraredNonspecific,
        MidInfrared2220nm,
        MidInfrared1625nm,
        MidInfraredNonspecific,
        NearInfrared700nm,
        NearInfrared825nm,
        NearInfrared950nm,
        NearInfraredNonspecific,
        Red,
        Orange,
        Yellow,
        Green,
        Blue,
        Violet,
        UltravioletA,
        UltravioletB,
        UltravioletC,
        ExtremeUltraviolet,
        SoftXRay,
        HardXRay,
        GammaRay,
        Elevation,
        Sonar,
        Proton,
        Neutron,
        Electron,
        Neutrino,
        Quark,
        UserDefined1,
        UserDefined2,
        UserDefined3,
        UserDefined4,
        UserDefined5,
        UserDefined6,
        UserDefined7,
        UserDefined8,
        UserDefined9
    }

    public class SpectralBand : IComparable
    {
        protected int _BandNumber;
        protected string _BandName;
        protected BandCategory _BandCategories;
        protected SpecialBandName _SpecialBandNames;
        protected SpectralBandData[] _BandData;
        protected bool _UseAsLumaChannel;
        protected bool _UseAsThreeChannelChannel;
        //protected NormalizationMode _InputNormalization;
        protected TransformationProfileCollection _InputTransformations;
        protected ArrayList _MutuallyExclusiveBandNames;

        protected const string _PlaneValidationRegex = @"\[0,0,[0-9]\]";

        #region Properties

        public SpectralBandData[] BandData
        {
            get
            {
                return _BandData;
            }
            set
            {
                _BandData = value;
            }
        }

        /// <summary>
        /// An integer which represents the number of the band (for sorting). Lower numbers represent lower-frequency bands.
        /// </summary>
        public int BandNumber
        {
            get
            {
                return _BandNumber;
            }
            set
            {
                _BandNumber = value;
            }
        }

        /// <summary>
        /// A string which represents the frequency band. It will be used in all filenames which include the band in question.
        /// </summary>
        /// <example>
        /// NIR, G, 1000nm, etc.
        /// </example>
        public string BandName
        {
            get
            {
                return _BandName;
            }
            set
            {
                _BandName = value;
            }
        }

        public BandCategory BandCategories
        {
            get
            {
                return _BandCategories;
            }
            set
            {
                _BandCategories = value;
            }
        }

        public SpecialBandName SpecialBandNames
        {
            get
            {
                return _SpecialBandNames;
            }
            set
            {
                _SpecialBandNames = value;
            }
        }

        /// <summary>
        /// Controls whether or not this spectral band will be available for use as the luminance channel
        /// in colour/luma variations
        /// </summary>
        public bool UseAsLumaChannel
        {
            get
            {
                return _UseAsLumaChannel;
            }
            set
            {
                _UseAsLumaChannel = value;
            }
        }

        /// <summary>
        /// Controls whether or not this spectral band will be available for use in one of the three channels
        /// in three-channel variations
        /// </summary>
        public bool UseAsThreeChannelChannel
        {
            get
            {
                return _UseAsThreeChannelChannel;
            }
            set
            {
                _UseAsThreeChannelChannel = value;
            }
        }

        /// <summary>
        /// Controls whether or not this spectral band will be normalized before any other use
        /// </summary>
        //public NormalizationMode InputNormalization
        //{
        //    get
        //    {
        //        return _InputNormalization;
        //    }
        //    set
        //    {
        //        _InputNormalization = value;
        //    }
        //}

        public TransformationProfileCollection InputTransformations
        {
            get
            {
                return _InputTransformations;
            }
            set
            {
                _InputTransformations = value;
            }
        }

        /// <summary>
        /// A list of bands that should not be used in other channels in images that also contain this band
        /// For example, OnEarth image sets contain high- and low-gain thermal infrared data which is the same
        /// except for the amplitude scaling. Using both the high- and low-gain data in the same image would
        /// be redundant.
        /// Note that this restriction applies to greyscale channels made up of multiple spectral bands, 
        /// with the exception of greyscale channels made up of all bands (otherwise those channels wouldn't
        /// really contain _all_ bands).
        /// </summary>
        public ArrayList MutuallyExclusiveBandNames
        {
            get
            {
                return _MutuallyExclusiveBandNames;
            }
            set
            {
                _MutuallyExclusiveBandNames = value;
            }
        }

        #endregion

        /// <summary>
        /// Creates a new spectral band representation
        /// </summary>
        /// <param name="bandNumber">The band number to use for sorting. The higher the number, the higher the frequency.</param>
        /// <param name="bandName">The name to use when representing this spectral band - NIR, G, 240nm, etc.</param>
        public SpectralBand(int bandNumber, string bandName)
        {
            _BandNumber = bandNumber;
            _BandName = bandName;
            _BandCategories = BandCategory.None;
            _SpecialBandNames = SpecialBandName.None;
            _BandData = new SpectralBandData[2];
            _BandData[0] = new SpectralBandData();
            _BandData[1] = new SpectralBandData();
            _UseAsLumaChannel = true;
            _UseAsThreeChannelChannel = true;
            InputTransformations = new TransformationProfileCollection();
            _MutuallyExclusiveBandNames = new ArrayList();
        }

        public SpectralBand(SpectralBand templateBand, bool copyBandData)
            : this(templateBand.BandNumber, templateBand.BandName)
        {
            if (copyBandData)
            {
                _BandData = templateBand.BandData;
            }
            _BandCategories = templateBand.BandCategories;
            _SpecialBandNames = templateBand.SpecialBandNames;
            _UseAsLumaChannel = templateBand.UseAsLumaChannel;
            _UseAsThreeChannelChannel = templateBand.UseAsThreeChannelChannel;
            InputTransformations = templateBand.InputTransformations;
            _MutuallyExclusiveBandNames = templateBand.MutuallyExclusiveBandNames;
        }

        /// <summary>
        /// Creates a new spectral band representation and sets the number to the default of 0
        /// </summary>
        /// <param name="bandName">The name to use when representing this spectral band - NIR, G, 240nm, etc.</param>
        public SpectralBand(string bandName)
            : this(0, bandName)
        {
        }

        public SpectralBand()
            : this(0, "")
        {
        }

        public bool IsBandMember(BandCategory category)
        {
            if ((BandCategories & category) == category)
            {
                return true;
            }
            return false;
        }

        public void AddToBandCategory(BandCategory category)
        {
            BandCategories = BandCategories | category;
        }

        public void RemoveFromBandCategory(BandCategory category)
        {
            BandCategories = BandCategories ^ category;
        }

        /// <summary>
        /// Sets the source information for the primary version of the image for this spectral band.
        /// This version will always be used in traditional (R-G-B) false colour variations.
        /// If a secondary version is not present, the primary will also be used in tinted (and 
        /// tinted-derived, e.g. luma/colour) variations.
        /// </summary>
        /// <param name="fileName">The relative or absolute path to the source file.</param>
        /// <param name="plane">The plane in the file which contains the greyscale data for this channel (in DaVinci syntax, e.g. [0,0,1] for the first plane.</param>
        public void SetPrimaryImageData(string fileName, string plane)
        {
            SetImageData(0, fileName, plane);
        }

        public void SetPrimaryImageData(SpectralBandData band)
        {
            SetImageData(0, band);
        }

        /// <summary>
        /// Sets the source information for the secondary version of the image for this spectral band.
        /// If present, it will be used in tinted (and tinted-derived, e.g. luma/colour) variations
        /// in place of the primary version.
        /// </summary>
        /// <param name="fileName">The relative or absolute path to the source file.</param>
        /// <param name="plane">The plane in the file which contains the greyscale data for this channel (in DaVinci syntax, e.g. [0,0,1] for the first plane.</param>
        /// <param name="bitsPerChannel">The bit depth (per channel, **not** per pixel!) of the source image.</param>
        public void SetSecondaryImageData(string fileName, string plane)
        {
            SetImageData(1, fileName, plane);
        }

        public void SetSecondaryImageData(SpectralBandData band)
        {
            SetImageData(1, band);
        }

        protected void SetImageData(int imageNumber, string fileName, string plane)
        {
            SpectralBandData currentBand = _BandData[imageNumber];
            try
            {
                SpectralBandData band = CreateAndValidateSpectralBand(fileName, plane);
                SetImageData(imageNumber, band);
            }
            catch (Exception ex)
            {
                _BandData[imageNumber] = currentBand;
                if (currentBand.FileName == "")
                {
                    throw new ArgumentException("The specified image data was invalid. No previous values were present, so the values are still null.", ex);
                }
                else
                {
                    throw new ArgumentException("The specified image data was invalid. The previous values were retained.", ex);
                }
            }

        }

        protected void SetImageData(int imageNumber, SpectralBandData band)
        {
            if (!IsValidPlaneString(band.Plane))
            {
                throw new FormatException("The value specified for the image plane of the file '" + band.FileName + "' ('" + band.Plane + "') does not appear valid. The value should be in the format [0,0,x], where 'x' denotes the image plane number.");
            }
            _BandData[imageNumber] = band;
        }

        public bool HasPrimaryImage
        {
            get
            {
                if (_BandData[0].FileName != "")
                {
                    return true;
                }
                return false;
            }
        }

        public bool HasSecondaryImage
        {
            get
            {
                if (_BandData[1].FileName != "")
                {
                    return true;
                }
                return false;
            }
        }

        public SpectralBandData GetPrimaryImageData()
        {
            if (_BandData[0].FileName != "")
            {
                return _BandData[0];
            }
            if (_BandData[1].FileName != "")
            {
                return _BandData[1];
            }
            throw new Exception("The spectral band '" + _BandName + "' (band number " + _BandNumber.ToString() + ") does not contain a primary or secondary file reference.");
        }

        public SpectralBandData GetSecondaryImageData()
        {
            if (_BandData[1].FileName != "")
            {
                return _BandData[1];
            }
            if (_BandData[0].FileName != "")
            {
                return _BandData[0];
            }
            throw new Exception("The spectral band '" + _BandName + "' (band number " + _BandNumber.ToString() + ") does not contain a primary or secondary file reference.");
        }

        protected SpectralBandData CreateAndValidateSpectralBand(string fileName, string plane)
        {
            SpectralBandData band = new SpectralBandData();
            if (!File.Exists(fileName))
            {
                throw new FileNotFoundException("The file '" + fileName + "' was not found in the working directory (" + Directory.GetCurrentDirectory() + ").");
            }
            band.FileName = fileName;
            if (!IsValidPlaneString(plane))
            {
                throw new FormatException("The value specified for the image plane (" + plane + ") does not appear valid. The value should be in the format [0,0,x], where 'x' denotes the image plane number.");
            }
            band.Plane = plane;

            return band;
        }

        public static bool IsValidPlaneString(string plane)
        {
            Regex rEx = new Regex(_PlaneValidationRegex);
            if (!rEx.IsMatch(plane))
            {
                return false;
            }
            return true;
        }

        public int CompareTo(object obj)
        {
            if (obj is SpectralBand)
            {
                SpectralBand b2 = (SpectralBand)obj;
                if (b2.BandNumber != BandNumber)
                {
                    return _BandNumber.CompareTo(b2.BandNumber);
                }
                else
                {
                    return _BandName.CompareTo(b2.BandName);
                }
            }
            else
            {
                return _BandName.CompareTo(obj.ToString());
            }
        }

        public string[] GetMutuallyExclusiveBandList()
        {
            string proto = "";
            return (string[])MutuallyExclusiveBandNames.ToArray(proto.GetType());
        }

        public bool IsMutuallyExclusiveWithBand(string bandName)
        {
            if (MutuallyExclusiveBandNames.Contains(bandName))
            {
                return true;
            }
            return false;
        }

        public void AddMutuallyExclusiveBand(string newBandName)
        {
            if (!MutuallyExclusiveBandNames.Contains(newBandName))
            {
                MutuallyExclusiveBandNames.Add(newBandName);
            }
        }

        public static BandCategory BandCategoryFromString(string categoryName)
        {
            string cName = categoryName.Trim().ToLower();

            switch (cName)
            {
                case "radio":
                    return BandCategory.Radio;
                    break;
                case "microwave":
                    return BandCategory.Microwave;
                    break;
                case "farinfrared":
                    return BandCategory.FarInfrared;
                    break;
                case "midinfrared":
                    return BandCategory.MidInfrared;
                    break;
                case "nearinfrared":
                    return BandCategory.NearInfrared;
                    break;
                case "humanvisible":
                    return BandCategory.HumanVisible;
                    break;
                case "ultraviolet":
                    return BandCategory.Ultraviolet;
                    break;
                case "xray":
                    return BandCategory.XRay;
                    break;
                case "gammaray":
                    return BandCategory.GammaRay;
                    break;
                case "derived":
                    return BandCategory.Derived;
                    break;
            }
            throw new ArgumentException("The string '" + categoryName + "' could not be converted to a recognized spectral band category.");
        }

        public static SpecialBandName SpecialBandNameFromString(string name)
        {
            string sName = name.Trim().ToLower();

            switch (sName)
            {
                case "radioelf":
                    return SpecialBandName.RadioELF;
                    break;
                case "radioslf":
                    return SpecialBandName.RadioSLF;
                    break;
                case "radioulf":
                    return SpecialBandName.RadioULF;
                    break;
                case "radiovlf":
                    return SpecialBandName.RadioVLF;
                    break;
                case "radiolf":
                    return SpecialBandName.RadioLF;
                    break;
                case "radiomf":
                    return SpecialBandName.RadioMF;
                    break;
                case "radiohf":
                    return SpecialBandName.RadioHF;
                    break;
                case "radiovhf":
                    return SpecialBandName.RadioVHF;
                    break;
                case "radiouhf":
                    return SpecialBandName.RadioUHF;
                    break;
                case "radioshf":
                    return SpecialBandName.RadioSHF;
                    break;
                case "radioehf":
                    return SpecialBandName.RadioEHF;
                    break;
                case "radar":
                    return SpecialBandName.Radar;
                    break;
                case "radiononspecific":
                    return SpecialBandName.RadioNonSpecific;
                    break;
                case "microwavepband":
                    return SpecialBandName.MicrowavePBand;
                    break;
                case "microwavelband":
                    return SpecialBandName.MicrowaveLBand;
                    break;
                case "microwavesband":
                    return SpecialBandName.MicrowaveSBand;
                    break;
                case "microwavecband":
                    return SpecialBandName.MicrowaveCBand;
                    break;
                case "microwavexband":
                    return SpecialBandName.MicrowaveXBand;
                    break;
                case "microwavekband":
                    return SpecialBandName.MicrowaveKBand;
                    break;
                case "microwaveqband":
                    return SpecialBandName.MicrowaveQBand;
                    break;
                case "microwavevband":
                    return SpecialBandName.MicrowaveVBand;
                    break;
                case "microwavewband":
                    return SpecialBandName.MicrowaveWBand;
                    break;
                case "microwavenonspecific":
                    return SpecialBandName.MicrowaveNonspecific;
                    break;
                case "terahertz":
                    return SpecialBandName.Terahertz;
                    break;
                case "thermalinfrared":
                    return SpecialBandName.ThermalInfrared;
                    break;
                case "farinfrarednonspecific":
                    return SpecialBandName.FarInfraredNonspecific;
                    break;
                case "midinfrared2220nm":
                    return SpecialBandName.MidInfrared2220nm;
                    break;
                case "midinfrared1625nm":
                    return SpecialBandName.MidInfrared1625nm;
                    break;
                case "midinfrarednonspecific":
                    return SpecialBandName.MidInfraredNonspecific;
                    break;
                case "nearinfrared700nm":
                    return SpecialBandName.NearInfrared700nm;
                    break;
                case "nearinfrared825nm":
                    return SpecialBandName.NearInfrared825nm;
                    break;
                case "nearinfrared950nm":
                    return SpecialBandName.NearInfrared950nm;
                    break;
                case "nearinfrarednonspecific":
                    return SpecialBandName.NearInfraredNonspecific;
                    break;
                case "red":
                    return SpecialBandName.Red;
                    break;
                case "orange":
                    return SpecialBandName.Orange;
                    break;
                case "yellow":
                    return SpecialBandName.Yellow;
                    break;
                case "green":
                    return SpecialBandName.Green;
                    break;
                case "blue":
                    return SpecialBandName.Blue;
                    break;
                case "violet":
                    return SpecialBandName.Violet;
                    break;
                case "ultravioleta":
                    return SpecialBandName.UltravioletA;
                    break;
                case "ultravioletb":
                    return SpecialBandName.UltravioletB;
                    break;
                case "ultravioletc":
                    return SpecialBandName.UltravioletC;
                    break;
                case "softxray":
                    return SpecialBandName.SoftXRay;
                    break;
                case "hardxray":
                    return SpecialBandName.HardXRay;
                    break;
                case "gammaray":
                    return SpecialBandName.GammaRay;
                    break;
                case "elevation":
                    return SpecialBandName.Elevation;
                    break;
                case "sonar":
                    return SpecialBandName.Sonar;
                    break;
                case "proton":
                    return SpecialBandName.Proton;
                    break;
                case "neutron":
                    return SpecialBandName.Neutron;
                    break;
                case "electron":
                    return SpecialBandName.Electron;
                    break;
                case "neutrino":
                    return SpecialBandName.Neutrino;
                    break;
                case "quark":
                    return SpecialBandName.Quark;
                    break;
                case "userdefined1":
                    return SpecialBandName.UserDefined1;
                    break;
                case "userdefined2":
                    return SpecialBandName.UserDefined2;
                    break;
                case "userdefined3":
                    return SpecialBandName.UserDefined3;
                    break;
                case "userdefined4":
                    return SpecialBandName.UserDefined4;
                    break;
                case "userdefined5":
                    return SpecialBandName.UserDefined5;
                    break;
                case "userdefined6":
                    return SpecialBandName.UserDefined6;
                    break;
                case "userdefined7":
                    return SpecialBandName.UserDefined7;
                    break;
                case "userdefined8":
                    return SpecialBandName.UserDefined8;
                    break;
                case "userdefined9":
                    return SpecialBandName.UserDefined9;
                    break;
            }
            throw new ArgumentException("The string '" + name + "' could not be converted to a recognized name for a special spectral band.");
        }
    }
}
