// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;

namespace BenLincoln.DaVinci
{
    public class TMSBColourConfigurationFile : TMSBConfigurationFile
    {
        protected TMSBColourConfiguration _ColourConfig;

        protected enum ConfigProcessingState
        {
            Normal,
            TintingColourSection,
            TintingColour,
            GradientMain,
            Gradient,
            GradientColour
        }

        #region Properties

        public TMSBColourConfiguration ColourConfig
        {
            get
            {
                return _ColourConfig;
            }
            set
            {
                _ColourConfig = value;
            }
        }

        #endregion

        public TMSBColourConfigurationFile(string configFileName)
        {
            ColourConfig = new TMSBColourConfiguration();
            ReadData(configFileName);
        }

        protected override void ReadTypeSpecificData(Stream sourceStream)
        {
            XmlTextReader xReader = new XmlTextReader(sourceStream);
            //xReader.Settings.Schemas.Add(_FileSchema);

            string currentElementName = "";
            ConfigProcessingState currentState = ConfigProcessingState.Normal;

            TintingColourXMLParameters newColour = new TintingColourXMLParameters();
            ColourGradient newGradient = new ColourGradient();
            GradientColourXMLParameters newGradientColourParameters = new GradientColourXMLParameters();

            while (xReader.Read())
            {
                switch (xReader.NodeType)
                {
                    case XmlNodeType.Element:
                        currentElementName = xReader.Name;
                        //Console.WriteLine("XML DEBUG: Begin Element - Current Processing State '" + currentState.ToString() + "' - Element name: '" + currentElementName + "'.");
                        switch (currentState)
                        {
                            case ConfigProcessingState.Normal:
                                switch (currentElementName)
                                {
                                    case "TintingColours":
                                        currentState = ConfigProcessingState.TintingColourSection;
                                        break;
                                    case "Gradients":
                                        currentState = ConfigProcessingState.GradientMain;
                                        break;
                                }
                                break;
                            // beginning of a tinting colour definition
                            case ConfigProcessingState.TintingColourSection:
                                if (currentElementName == "TintingColour")
                                {
                                    currentState = ConfigProcessingState.TintingColour;
                                    newColour = new TintingColourXMLParameters();
                                }
                                break;

                            // beginning of a colour gradient definition
                            case ConfigProcessingState.GradientMain:
                                if (currentElementName == "Gradient")
                                {
                                    currentState = ConfigProcessingState.Gradient;
                                    newGradient = new ColourGradient();
                                }
                                break;

                            // beginning of the definition of an individual colour within a gradient
                            case ConfigProcessingState.Gradient:
                                if (currentElementName == "GradientColour")
                                {
                                    currentState = ConfigProcessingState.GradientColour;
                                    newGradientColourParameters = new GradientColourXMLParameters();
                                }
                                break;
                        }
                        break;
                    case XmlNodeType.EndElement:
                        currentElementName = xReader.Name;
                        //Console.WriteLine("XML DEBUG: End Element - Current Processing State '" + currentState.ToString() + "' - Element name: '" + currentElementName + "'.");
                        switch (currentState)
                        {
                            case ConfigProcessingState.TintingColourSection:
                                if (currentElementName == "TintingColours")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;

                            // end of a tinting colour definition
                            case ConfigProcessingState.TintingColour:
                                if (currentElementName == "TintingColour")
                                {
                                    currentState = ConfigProcessingState.TintingColourSection;
                                    ColourConfig.AddTintingColour(TintingColour.FromXMLParameters(newColour));
                                }
                                break;

                            // end of all gradient definitions
                            case ConfigProcessingState.GradientMain:
                                if (currentElementName == "Gradients")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;

                            // end of a gradient definition
                            case ConfigProcessingState.Gradient:
                                if (currentElementName == "Gradient")
                                {
                                    currentState = ConfigProcessingState.GradientMain;
                                    ColourConfig.AddColourGradient(newGradient);
                                }
                                break;

                            // end of a colour definition within a gradient
                            case ConfigProcessingState.GradientColour:
                                if (currentElementName == "GradientColour")
                                {
                                    currentState = ConfigProcessingState.Gradient;
                                    newGradient.AddColour(GradientColour.FromXMLParameters(newGradientColourParameters));
                                }
                                break;
                        }
                        break;
                    case XmlNodeType.Text:
                        string currentText = xReader.Value.Trim();
                        switch (currentState)
                        {
                            // tinting colour definition-related data
                            case ConfigProcessingState.TintingColour:
                                switch (currentElementName)
                                {
                                    case "Name":
                                        newColour.TintingColourName = currentText;
                                        break;
                                    case "Format":
                                        newColour.ColourParameters.Format = currentText;
                                        break;
                                    case "Red":
                                        //newColour.ColourParameters.Red = float.Parse(currentText);
                                        newColour.ColourParameters.Red = BenLincoln.Utility.FormatConverter.StringToFloat(currentText);
                                        break;
                                    case "Green":
                                        //newColour.ColourParameters.Green = float.Parse(currentText);
                                        newColour.ColourParameters.Green = BenLincoln.Utility.FormatConverter.StringToFloat(currentText);
                                        break;
                                    case "Blue":
                                        //newColour.ColourParameters.Blue = float.Parse(currentText);
                                        newColour.ColourParameters.Blue = BenLincoln.Utility.FormatConverter.StringToFloat(currentText);
                                        break;
                                }
                                break;

                            // gradient definitions
                            case ConfigProcessingState.Gradient:
                                switch (currentElementName)
                                {
                                    case "Name":
                                        newGradient.GradientName = currentText;
                                        break;
                                }
                                break;
                            case ConfigProcessingState.GradientColour:
                                switch (currentElementName)
                                {
                                    case "Position":
                                        newGradientColourParameters.Position = BenLincoln.Utility.FormatConverter.StringToInt(currentText);
                                        break;
                                    case "Format":
                                        newGradientColourParameters.ColourParameters.Format = currentText;
                                        break;
                                    case "Red":
                                        //newGradientColourParameters.ColourParameters.Red = float.Parse(currentText);
                                        newGradientColourParameters.ColourParameters.Red = BenLincoln.Utility.FormatConverter.StringToFloat(currentText);
                                        break;
                                    case "Green":
                                        //newGradientColourParameters.ColourParameters.Green = float.Parse(currentText);
                                        newGradientColourParameters.ColourParameters.Green = BenLincoln.Utility.FormatConverter.StringToFloat(currentText);
                                        break;
                                    case "Blue":
                                        //newGradientColourParameters.ColourParameters.Blue = float.Parse(currentText);
                                        newGradientColourParameters.ColourParameters.Blue = BenLincoln.Utility.FormatConverter.StringToFloat(currentText);
                                        break;
                                }
                                break;
                        }
                        break;
                }
            }
        }
    }
}
