// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Schema;

namespace BenLincoln.DaVinci
{
    public enum TMSBConfigurationFileType
    {
        Input,
        Output,
        Colour,
        Processing,
        TransformationProfileList,
        ExecutionOptions
    }

    public class TMSBConfigurationFile : IComparable
    {
        protected const string _DefaultApplicationVersion = "1.0";
        protected const string _ConfigDirectoryName = "TMSB_Config";
        protected const string _XMLFragmentImportRegex = @"<import>[^<]{0,2048}</import>";
        protected const string _XMLFragmentImportCommentRegex = @"<!--(?:.(?<!--))*-->";

        // the maximum recursion depth for XML fragment importing
        protected const int _MaxImportIterations = 128;

        protected string _FilePath;
        protected string _OriginalFilePath;
        protected bool _ReadOnly;
        protected TMSBConfigurationFileType _FileType;
        protected string _ApplicationVersion;
        protected string _Name;
        protected string _Description;
        protected string _Warnings;
        protected string _Author;
        protected DateTime _Created;
        protected DateTime _LastModified;

        protected ArrayList _TempFileNames;

        protected XmlSchema _FileSchema;

        #region Properties

        //public string FilePath
        //{
        //    get
        //    {
        //        return _FilePath;
        //    }
        //    set
        //    {
        //        _FilePath = value;
        //    }
        //}
        public string FilePath
        {
            get
            {
                if (_OriginalFilePath != "")
                {
                    return _OriginalFilePath;
                }
                return _FilePath;
            }
            set
            {
                if (_OriginalFilePath == "")
                {
                    _OriginalFilePath = value;
                }
                _FilePath = value;
            }
        }

        public bool ReadOnly
        {
            get
            {
                return _ReadOnly;
            }
        }

        public TMSBConfigurationFileType FileType
        {
            get
            {
                return _FileType;
            }
            set
            {
                _FileType = value;
            }
        }

        public string ApplicationVersion
        {
            get
            {
                return _ApplicationVersion;
            }
            set
            {
                _ApplicationVersion = value;
            }
        }

        public string Name
        {
            get
            {
                return _Name;
            }
            set
            {
                _Name = value;
            }
        }

        public string Description
        {
            get
            {
                return _Description;
            }
            set
            {
                _Description = value;
            }
        }

        public string Warnings
        {
            get
            {
                return _Warnings;
            }
            set
            {
                _Warnings = value;
            }
        }

        public string Author
        {
            get
            {
                return _Author;
            }
            set
            {
                _Author = value;
            }
        }

        public DateTime Created
        {
            get
            {
                return _Created;
            }
            set
            {
                _Created = value;
            }
        }

        public DateTime LastModified
        {
            get
            {
                return _LastModified;
            }
            set
            {
                _LastModified = value;
            }
        }


        #endregion

        public TMSBConfigurationFile()
        {
            _FilePath = "";
            _OriginalFilePath = "";
            _ReadOnly = true;
            _ApplicationVersion = _DefaultApplicationVersion;
            _Name = "";
            _Description = "";
            _Warnings = "";
            _Author = "";
            _Created = DateTime.MinValue;
            _LastModified = DateTime.MinValue;
            _TempFileNames = new ArrayList();
        }

        ~TMSBConfigurationFile()
        {
            if (_TempFileNames.Count > 0)
            {
                string[] tempFiles = (string[])_TempFileNames.ToArray("".GetType());
                for (int i = 0; i < tempFiles.Length; i++)
                {
                    try
                    {
                        File.Delete(tempFiles[i]);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Error deleting temporary file '" + tempFiles[i] +
                            "': " + ex.Message);
                    }
                }
            }
        }

        protected static string GetDescriptionString(string rawText)
        {
            string result = rawText.Replace("\t", "");

            while (result.Contains("  "))
            {
                result = result.Replace("  ", " ");
            }

            return result;
        }

        public void ReadData(string configurationFileName)
        {
            Stream inStream = GetConfigurationStream(configurationFileName);
            if (inStream is FileStream)
            {
                _ReadOnly = false;
                FileStream fInStream = (FileStream)inStream;
                FilePath = fInStream.Name;
            }
            else
            {
                //_FilePath = "Embedded resource: '" + configurationFileName + "'";
                FilePath = configurationFileName;
            }
            //Console.WriteLine("Validating '" + configurationFileName + "'.");
            _FileSchema = null;
            try
            {
                ValidateXML(inStream);
                //_FileSchema = GetSchema(inStream);
            }
            catch (Exception ex)
            {
                throw new Exception("Failed to validate the file '" + configurationFileName + "'. " +
                    "The error message was '" + ex.Message + "'.", ex);
            }

            // get the metadata properties
            XmlTextReader xReader = new XmlTextReader(inStream);
            //XmlReaderSettings xSettings = new XmlReaderSettings();
            //xSettings.ValidationType = ValidationType.Schema;
            //xReader.

            //xReader.Settings.Schemas.Add(_FileSchema);

            string currentElementName = "";
            bool inHeader = false;
            bool doneReading = false;
            while (xReader.Read())
            {
                switch (xReader.NodeType)
                {
                    case XmlNodeType.Element:
                        currentElementName = xReader.Name;
                        if (currentElementName == "MetaData")
                        {
                            inHeader = true;
                        }
                        break;
                    case XmlNodeType.EndElement:
                        currentElementName = xReader.Name;
                        if (currentElementName == "MetaData")
                        {
                            inHeader = false;
                            doneReading = true;
                        }
                        break;
                    case XmlNodeType.Text:
                        string currentText = xReader.Value.Trim();
                        if (inHeader)
                        {
                            switch (currentElementName)
                            {
                                case "FileType":
                                    _FileType = TypeNameToType(currentText);
                                    break;
                                case "GeneratingVersion":
                                    _ApplicationVersion = currentText;
                                    break;
                                case "Name":
                                    _Name = currentText;
                                    break;
                                case "Description":
                                    _Description = GetDescriptionString(currentText);
                                    break;
                                case "Warnings":
                                    _Warnings = GetDescriptionString(currentText);
                                    break;
                                case "Author":
                                    _Author = currentText;
                                    break;
                                case "Created":
                                    _Created = DateTime.Parse(currentText);
                                    break;
                                case "LastModified":
                                    _LastModified = DateTime.Parse(currentText);
                                    break;
                            }
                        }
                        break;
                }
                if (doneReading)
                {
                    break;
                }
            }

            if (inStream.CanSeek)
            {
                inStream.Seek(0, SeekOrigin.Begin);
            }
            ReadTypeSpecificData(inStream);
            
            inStream.Close();
        }

        public void WriteData(string configurationFileName)
        {
            try
            {
                FileStream oStream = new FileStream(configurationFileName, FileMode.Create, FileAccess.Write);

                // write the config file header
                XmlTextWriter oWriter = new XmlTextWriter(oStream, Encoding.Unicode);
                oWriter.Formatting = Formatting.Indented;
                oWriter.WriteStartDocument();
                oWriter.WriteStartElement("TMSB");
                oWriter.WriteStartElement("MetaData");
                
                    oWriter.WriteStartElement("ApplicationData");
                        oWriter.WriteStartElement("FileType");
                        oWriter.WriteValue(TypeToTypeName(FileType));
                        oWriter.WriteEndElement();
                        oWriter.WriteStartElement("GeneratingVersion");
                        oWriter.WriteValue(_DefaultApplicationVersion);
                        oWriter.WriteEndElement();
                    oWriter.WriteEndElement();
                    
                    oWriter.WriteStartElement("Name");
                    oWriter.WriteValue(Name);
                    oWriter.WriteEndElement();
                    
                    oWriter.WriteStartElement("Description");
                    oWriter.WriteValue(Description);
                    oWriter.WriteEndElement();
                    
                    oWriter.WriteStartElement("Author");
                    oWriter.WriteValue(Author);
                    oWriter.WriteEndElement();
                    
                    oWriter.WriteStartElement("Created");
                    //oWriter.WriteValue(Created.ToLongDateString());
                    oWriter.WriteValue(Created);
                    oWriter.WriteEndElement();
                    
                    oWriter.WriteStartElement("LastModified");
                    //oWriter.WriteValue(DateTime.Now.ToLongDateString());
                    oWriter.WriteValue(DateTime.Now.ToUniversalTime());
                    oWriter.WriteEndElement();
                    
                oWriter.WriteEndElement();

                WriteTypeSpecificData(oWriter);
                oWriter.WriteEndElement();
                
                oWriter.WriteEndDocument();
                oWriter.Close();
                oStream.Close();
                try
                {
                    //oWriter.Dispose();
                    oStream.Dispose();
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            catch (Exception ex)
            {
                throw new IOException("Unable to write configuration data to the file '" + configurationFileName +
                    "'. The error description was '" + ex.Message + "'.", ex);
            }
        }

        protected virtual void ReadTypeSpecificData(Stream sourceStream)
        {
            //throw new NotImplementedException("This class does not implement the ReadTypeSpecificData method.");
        }

        protected virtual void WriteTypeSpecificData(XmlTextWriter oWriter)
        {
            //throw new NotImplementedException("This class does not implement the WriteTypeSpecificData method.");
        }

        public static TMSBConfigurationFileType TypeNameToType(string typeName)
        {
            switch (typeName)
            {
                case "Input Configuration":
                    return TMSBConfigurationFileType.Input;
                    break;
                case "Output Configuration":
                    return TMSBConfigurationFileType.Output;
                    break;
                case "Colour Configuration":
                    return TMSBConfigurationFileType.Colour;
                    break;
                case "Processing Configuration":
                    return TMSBConfigurationFileType.Processing;
                    break;
                case "Transformation Profile List":
                    return TMSBConfigurationFileType.TransformationProfileList;
                    break;
                case "Execution Options":
                    return TMSBConfigurationFileType.ExecutionOptions;
                    break;
                default:
                    throw new InvalidDataException("The string '" + typeName + "' does not identify a filetype known to this application.");
            }
        }

        public static string TypeToTypeName(TMSBConfigurationFileType configType)
        {
            switch (configType)
            {
                case TMSBConfigurationFileType.Input:
                    return "Input Configuration";
                    break;
                case TMSBConfigurationFileType.Output:
                    return "Output Configuration";
                    break;
                case TMSBConfigurationFileType.Colour:
                    return "Colour Configuration";
                    break;
                case TMSBConfigurationFileType.Processing:
                    return "Processing Configuration";
                    break;
                case TMSBConfigurationFileType.TransformationProfileList:
                    return "Transformation Profile List";
                    break;
                case TMSBConfigurationFileType.ExecutionOptions:
                    return "Execution Options";
                    break;
                default:
                    throw new InvalidDataException("The configuration file type '" + configType.ToString() + 
                        "' could not be matched to a filetype known to this application.");
            }
        }

        protected static string GetFileContents(string filePath)
        {
            FileStream iStream = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            StreamReader iReader = new StreamReader(iStream);
            string contents = iReader.ReadToEnd();
            iReader.Close();
            iStream.Close();
            return contents;
        }

        protected static Stream GetConfigurationFileMemoryStream(string filePath)
        {
            try
            {
                string contents = GetFileContents(filePath);

                // replace "import" statements with their referenced contents
                Regex importRex = new Regex(_XMLFragmentImportRegex, RegexOptions.IgnoreCase | RegexOptions.Multiline);
                Regex commentRex = new Regex(_XMLFragmentImportCommentRegex, RegexOptions.IgnoreCase | RegexOptions.Multiline);
                Regex rRex1 = new Regex("<import>", RegexOptions.IgnoreCase | RegexOptions.Multiline);
                Regex rRex2 = new Regex("</import>", RegexOptions.IgnoreCase | RegexOptions.Multiline);
                //bool foundMatch = false;
                int importMatchCount = 1911;
                int iterationCount = 0;
                while (importMatchCount > 0)
                {
                    // strip comments
                    contents = commentRex.Replace(contents, "");

                    MatchCollection matches = importRex.Matches(contents);
                    importMatchCount = matches.Count;

                    foreach (Match m in matches)
                    {
                        //foundMatch = true;
                        string importFileName = m.Value;
                        importFileName = rRex1.Replace(importFileName, "");
                        importFileName = rRex2.Replace(importFileName, "");
                        importFileName = importFileName.Trim();

                        string importContents = "";
                        try
                        {
                            importContents = GetFileContents(GetAbsoluteConfigurationFilePath(importFileName));
                        }
                        catch (Exception ex)
                        {
                            importContents = "<!-- Unable to import the file '" + importFileName + "' - " + ex.Message + " -->";
                        }

                        contents = contents.Replace(m.Value, importContents);
                    }
                    iterationCount++;
                    if (iterationCount > _MaxImportIterations)
                    {
                        throw new IOException("TMSB reached the limit of " + _MaxImportIterations.ToString() +
                            " import statement iterations while processing the file '" +
                            filePath + "'. It is likely that one or more referenced XML fragments contain " +
                            "a circular reference. Please correct the configuration files and then start " +
                            "TMSB again.");
                    }
                }

                MemoryStream ms = new MemoryStream((contents.Length * 4));
                UnicodeEncoding ue = new UnicodeEncoding();
                
                // for debugging purposes
                //FileStream fs = new FileStream("XML_Debug_Output.xml", FileMode.Append, FileAccess.Write);

                byte[] cb = ue.GetBytes(contents);
                for (int i = 0; i < cb.Length; i++)
                {
                    ms.WriteByte(cb[i]);
                }

                //Console.Write(contents);

                //if (foundMatch)
                //{
                //    Console.WriteLine("x");
                //}

                ms.Seek(0, SeekOrigin.Begin);

                //fs.Close();

                return ms;
            }
            catch (Exception ex)
            {
                throw new IOException("Unable to load and/or parse the file '" + filePath + "' - the exact error message was '" +
                    ex.Message + "'.", ex);
            }
        }

        public Stream GetConfigurationStream(string configurationFileName)
        {
            bool foundSource = false;

            string filePath = "";

            try
            {
                bool foundPath = false;
                if (Path.IsPathRooted(configurationFileName))
                {
                    if (File.Exists(configurationFileName))
                    {
                        foundPath = true;
                        filePath = configurationFileName;
                    }
                }
                if (!foundPath)
                {
                    filePath = GetAbsoluteConfigurationFilePath(configurationFileName);
                }

                FilePath = filePath;

                //FileStream fStream = new FileStream(filePath, FileMode.Open, FileAccess.Read);
                //return fStream;

                // quick hack attempt to work around Mono's inability to handle the 0xFFFF character at 
                // the beginning of UTF-16-encoded XML memory streams
                bool useUTFHack = true;

                if (!useUTFHack)
                {
                    return GetConfigurationFileMemoryStream(filePath);
                }
                else
                {
                    string tempFileName = Path.GetTempFileName();
                    try
                    {
                        FileStream oStream = new FileStream(tempFileName, FileMode.Create, FileAccess.Write);
                        StreamWriter oWriter = new StreamWriter(oStream, new UnicodeEncoding());
                        _TempFileNames.Add(tempFileName);
                        Stream ms = GetConfigurationFileMemoryStream(filePath);
                        StreamReader iReader = new StreamReader(ms, new UnicodeEncoding());
                        char[] chars = iReader.ReadToEnd().ToCharArray();
                        for (int i = 0; i < chars.Length; i++)
                        {
                            oWriter.Write(chars[i]);
                        }
                        oWriter.Close();
                        oStream.Close();
                        iReader.Close();
                    }
                    catch (Exception ex)
                    {
                        throw new IOException("Unable to create the temporary XML file '" + tempFileName + "': " + ex.Message, ex);
                    }
                    try
                    {
                        FileStream oStream = new FileStream(tempFileName, FileMode.Open, FileAccess.Read);
                        return oStream;
                    }
                    catch (Exception ex)
                    {
                        throw new IOException("Unable to re-read the temporary XML file '" + tempFileName + "': " + ex.Message, ex);
                    }
                }
            }
            catch (FileNotFoundException fEx)
            {
                // do nothing, proceed to next code block
            }
            catch (Exception eX)
            {
                throw new IOException("An error occurred when accessing the file '" + filePath + "'. The exception message was '" + eX.Message +
                    "'.", eX);
            }

            //try
            //{
            //    Stream eStream = GetEmbeddedFileStream(configurationFileName);
            //    return eStream;
            //}
            //catch (FileNotFoundException fEx)
            //{
            //    // do nothing, proceed to next code block
            //}
            //catch (Exception eX)
            //{
            //    throw new IOException("An error occurred when accessing the embedded resource '" + configurationFileName + "'. The exception message was '" + eX.Message +
            //        "'.", eX);
            //}

            throw new FileNotFoundException("Unable to locate the file '" + configurationFileName + "' in any known directory or embedded within the assembly.");
        }

        public static string[] GetEmbeddedFileNamesByRegex(string pattern)
        {
            return GetEmbeddedFileNamesByRegex(Assembly.GetExecutingAssembly(), pattern, false);
        }

        public static string[] GetEmbeddedFileNamesByRegex(Assembly asm, string pattern)
        {
            return GetEmbeddedFileNamesByRegex(asm, pattern, false);
        }

        public static string[] GetEmbeddedFileNamesByRegex(Assembly asm, string pattern, bool caseSensitive)
        {
            ArrayList results = new ArrayList();
            Regex rEx = new Regex(pattern, RegexOptions.IgnoreCase | RegexOptions.Singleline);
            if (caseSensitive)
            {
                rEx = new Regex(pattern, RegexOptions.Singleline);
            }
            foreach (string currentPath in asm.GetManifestResourceNames())
            {
                if (rEx.IsMatch(currentPath))
                {
                    results.Add(currentPath);
                }
            }
            return (string[])results.ToArray(pattern.GetType());
        }

        public static Stream GetEmbeddedFileStream(string fileName)
        {
            return GetEmbeddedFileStream(Assembly.GetExecutingAssembly(), fileName);
        }

        public static Stream GetEmbeddedFileStream(Assembly asm, string fileName)
        {
            int matchCount = 0;
            string resourcePath = "";
            foreach (string currentPath in asm.GetManifestResourceNames())
            {
                //if (currentPath.EndsWith("." + fileName))
                if (currentPath.EndsWith(fileName))
                {
                    resourcePath = currentPath;
                    matchCount++;
                }
            }
            if (matchCount == 0)
            {
                throw new FileNotFoundException("Unable to locate an embedded resource ending with '." + fileName + "'.");
            }
            if (matchCount > 1)
            {
                throw new Exception("Error: there are multiple files embedded in the assembly whose names end with '." +
                    fileName + "'. The embedded resources must be renamed and the assembly rebuilt in order for them to be referenced.");
            }
            return asm.GetManifestResourceStream(resourcePath);
        }

        public static string GetAbsoluteConfigurationFilePath(string configurationFileName)
        {
            bool foundFile = false;
            string result = "";
            string filePath = "";

            // try the TMSB-specific locations first
            foreach (string searchDir in _GetSearchPathList())
            {
                filePath = searchDir + Path.DirectorySeparatorChar + configurationFileName;
                if (File.Exists(filePath))
                {
                    foundFile = true;
                    result = filePath;
                    break;
                }
            }

            //// fall back to trying everything in the PATH environment variable
            //if (!foundFile)
            //{
            //    filePath = _ConfigDirectoryName + Path.DirectorySeparatorChar + configurationFileName;
            //    if (File.Exists(filePath))
            //    {
            //        foundFile = true;
            //        result = filePath;
            //    }
            //}

            if (!foundFile)
            {
                throw new FileNotFoundException("The file '" + configurationFileName + "' was not found.");
            }
            return result;
        }

        public static TMSBConfigurationFile GetConfigurationFile(string path, bool returnFullFile)
        {
            try
            {
                TMSBConfigurationFile cf = new TMSBConfigurationFile();
                cf.ReadData(path);
                switch (cf.FileType)
                {
                    case TMSBConfigurationFileType.Input:
                        if (returnFullFile)
                        {
                            return new TMSBInputConfigurationFile(path);
                        }
                        else
                        {
                            return cf;
                        }
                        break;
                    case TMSBConfigurationFileType.Processing:
                        if (returnFullFile)
                        {
                            return new TMSBProcessingConfigurationFile(path);
                        }
                        else
                        {
                            return cf;
                        }
                        break;
                    case TMSBConfigurationFileType.Output:
                        if (returnFullFile)
                        {
                            return new TMSBOutputConfigurationFile(path);
                        }
                        else
                        {
                            return cf;
                        }
                        break;
                    case TMSBConfigurationFileType.Colour:
                        if (returnFullFile)
                        {
                            return new TMSBColourConfigurationFile(path);
                        }
                        else
                        {
                            return cf;
                        }
                        break;
                    case TMSBConfigurationFileType.TransformationProfileList:
                        if (returnFullFile)
                        {
                            return new TransformationProfileConfigurationFile(path);
                        }
                        else
                        {
                            return cf;
                        }
                        break;
                    case TMSBConfigurationFileType.ExecutionOptions:
                        if (returnFullFile)
                        {
                            return new TMSBEngineControllerConfigurationFile(path);
                        }
                        else
                        {
                            return cf;
                        }
                        break;
                    default:
                        throw new IOException("The file '" + path + "' is not of a type recognized by this application.");
                        break;
                }
            }
            catch (FileNotFoundException fEx)
            {
                throw new FileNotFoundException("The path '" + path + "' does not appear to represent a file that exists on this system.", fEx);
            }
            catch (IOException iEx)
            {
                throw new IOException("An IO error occurred while reading the file '" + path + 
                    "'. The specific error message was '" + iEx.Message + "'.", iEx);
            }
            catch (Exception ex)
            {
                throw new IOException("Unable to convert the file '" + path + "' into a TMSB configuration file. " +
                    "The specific error message was '" + ex.Message + "'.", ex);
            }
            throw new IOException("Unable to convert the file '" + path + "' into a TMSB configuration file.");
        }

        public static TMSBConfigurationFile[] GetConfigurationFiles(bool returnAllTypes, TMSBConfigurationFileType configFileType)
        {
            ArrayList results = new ArrayList();
            TMSBConfigurationFile proto = null;

            string[] searchPaths = _GetSearchPathList();

            foreach (string sPath in searchPaths)
            {
                try
                {
                    foreach (string file in Directory.GetFiles(sPath))
                    {
                        try
                        {
                            bool shouldOpen = true;
                            if (Path.GetExtension(file).ToLower() != ".xml")
                            {
                                shouldOpen = false;
                            }
                            string fN = Path.GetFileName(file);
                            if (fN.ToLower().StartsWith("fragment-"))
                            {
                                shouldOpen = false;
                            }

                            if (shouldOpen)
                            {
                                TMSBConfigurationFile cf = GetConfigurationFile(file, false);
                                if (returnAllTypes)
                                {
                                    results.Add(cf);
                                }
                                else
                                {
                                    if (cf.FileType == configFileType)
                                    {
                                        results.Add(cf);
                                        if (proto == null)
                                        {
                                            proto = cf;
                                        }
                                    }
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            // do nothing (for now)
                        }
                    }
                }
                catch (Exception ex)
                {
                    // do nothing (for now)
                }
            }

            if (proto == null)
            {
                return new TMSBConfigurationFile[0];
            }
            else
            {
                return (TMSBConfigurationFile[])results.ToArray(proto.GetType());
            }
        }

        public static string GetApplicationConfigDirectoryName()
        {
            return Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + Path.DirectorySeparatorChar + _ConfigDirectoryName;
        }

        public static string GetPersonalConfigDirectoryName()
        {
            return Environment.GetFolderPath(Environment.SpecialFolder.Personal) + Path.DirectorySeparatorChar + _ConfigDirectoryName;
        }

        public static TMSBConfigurationFile GetConfigurationFile(string path)
        {
            return GetConfigurationFile(path, true);
        }

        protected static string[] _GetSearchPathList()
        {
            ArrayList pathList = new ArrayList();
            
            // current working directory
            //AddToArrayListIfNotPresent(pathList, Directory.GetCurrentDirectory() + Path.DirectorySeparatorChar + _ConfigDirectoryName);

            // a TMSB config directory underneath the user's home directory
            AddToArrayListIfNotPresent(pathList, GetPersonalConfigDirectoryName());

            // the location of the TMSB application
            //AddToArrayListIfNotPresent(pathList, Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + Path.DirectorySeparatorChar + _ConfigDirectoryName);
            //AddToArrayListIfNotPresent(pathList, Path.GetDirectoryName(Assembly.GetCallingAssembly().Location) + Path.DirectorySeparatorChar + _ConfigDirectoryName);
            AddToArrayListIfNotPresent(pathList, GetApplicationConfigDirectoryName());

            string proto = "";

            return (string[])pathList.ToArray(proto.GetType());
        }

        protected static ArrayList AddToArrayListIfNotPresent(ArrayList list, string newEntry)
        {
            if (!list.Contains(newEntry))
            {
                list.Add(newEntry);
            }
            return list;
        }

        protected void ValidateXML(Stream inStream)
        {
            return;

            if (inStream.CanSeek)
            {
                inStream.Seek(0, SeekOrigin.Begin);
            }

            XmlTextReader xReader = new XmlTextReader(inStream);
            XmlReaderSettings xSettings = new XmlReaderSettings();
            xSettings.ValidationType = ValidationType.Schema;
            xSettings.ValidationEventHandler += new ValidationEventHandler(ValidationFailed);

            //string[] embeddedSchemaPaths = GetEmbeddedFileNamesByRegex(@"\" + Path.DirectorySeparatorChar + @"schema-.*\.xsd");
            string[] embeddedSchemaPaths = GetEmbeddedFileNamesByRegex(@"\.schema-.*\.xsd");

            bool foundSchema = false;

            foreach (string embeddedSchemaPath in embeddedSchemaPaths)
            {
                try
                {
                    Stream sStream = GetEmbeddedFileStream(embeddedSchemaPath);
                    XmlSchema schema = XmlSchema.Read(sStream, EmbeddedFileValidationFailed);
                    xSettings.Schemas.Add(schema);
                    foundSchema = true;
                }
                catch (Exception ex)
                {
                    // do nothing - ignore the file
                }
            }

            if (!foundSchema)
            {
                throw new Exception("Could not find any embedded schemas to match against the file.");
            }

            XmlReader xVReader = XmlReader.Create(xReader, xSettings);

            // have the reader parse the file
            while (xVReader.Read())
            {
                    // do nothing explicit
            }

            if (inStream.CanSeek)
            {
                inStream.Seek(0, SeekOrigin.Begin);
            }
        }

        protected XmlSchema GetSchema(Stream inStream)
        {
            string[] embeddedSchemaPaths = GetEmbeddedFileNamesByRegex(@"\.schema-.*\.xsd");

            foreach (string embeddedSchemaPath in embeddedSchemaPaths)
            {
                try
                {
                    if (inStream.CanSeek)
                    {
                        inStream.Seek(0, SeekOrigin.Begin);
                    }

                    XmlTextReader xReader = new XmlTextReader(inStream);
                    XmlReaderSettings xSettings = new XmlReaderSettings();
                    xSettings.ValidationType = ValidationType.Schema;
                    xSettings.ValidationEventHandler += new ValidationEventHandler(ValidationFailed);

                    Stream sStream = GetEmbeddedFileStream(embeddedSchemaPath);
                    XmlSchema schema = XmlSchema.Read(sStream, EmbeddedFileValidationFailed);
                    xSettings.Schemas.Add(schema);

                    XmlReader xVReader = XmlReader.Create(xReader, xSettings);

                    // have the reader parse the file
                    while (xVReader.Read())
                    {
                        // do nothing explicit
                    }

                    if (inStream.CanSeek)
                    {
                        inStream.Seek(0, SeekOrigin.Begin);
                    }

                    return schema;
                }
                catch (Exception ex)
                {
                    // do nothing - ignore the file
                }
            }

            throw new Exception("The file '" + FilePath + "' failed validation against all XML schemas known to this application.");

        }

        protected void ValidationFailed(object sender, ValidationEventArgs args)
        {
            throw new InvalidDataException("The file '" + FilePath + "' failed validation against the current XML schema.");
        }

        private void EmbeddedFileValidationFailed(object sender, ValidationEventArgs args)
        {
            throw new InvalidDataException("An embedded file failed XML validation. The failure message was '" + args.Message + "'");
        }

        protected bool StringToBool(string stringValue)
        {
            string val = stringValue.ToLower().Trim();

            switch (val)
            {
                case "true":
                    return true;
                    break;
                case "1":
                    return true;
                    break;
                case "yes":
                    return true;
                    break;
                case "y":
                    return true;
                    break;
                case "t":
                    return true;
                    break;
                case "false":
                    return false;
                    break;
                case "0":
                    return true;
                    break;
                case "no":
                    return true;
                    break;
                case "n":
                    return true;
                    break;
                case "f":
                    return true;
                    break;
                default:
                    if ((val.Contains("file")) && (val.Contains("not")) && (val.Contains("found")))
                    {
                        throw new InvalidDataException("The FILE_NOT_FOUND extended boolean state is not implemented in this release.");
                    }
                    break;
            }

            throw new InvalidDataException("The string '" + stringValue + "' could not be converted to a boolean (true/false) value.");
        }

        public int CompareTo(object obj)
        {
            if (obj is TMSBConfigurationFile)
            {
                TMSBConfigurationFile b2 = (TMSBConfigurationFile)obj;
                if (Name != b2.Name)
                {
                    return Name.CompareTo(b2.Name);
                }
                return FilePath.CompareTo(b2.FilePath);
            }
            else
            {
                return Name.CompareTo(obj.ToString());
            }
        }

        public override string ToString()
        {
            return Name;
        }
    }
}
