// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Threading;

using BenLincoln.Utility;

namespace BenLincoln.DaVinci
{
    public enum TMSBEngineState
    {
        Initializing,
        Running,
        Pausing,
        Paused,
        Resuming,
        Stopping,
        Stopped,
        Finished
    }

    public class TMSBEngine
    {
        protected TMSBEngineController _ParentController;
        protected string _BaseDirectory;
        //protected string _DaVinciCommand;
        protected const string _DVScriptFileName = "tmsb_script.dv";
        protected const string _DVScriptLogFileName = "tmsb_script_log.txt";
        protected const string _TemporaryDVScriptNamePrefix = "tmsb_script-temp-";
        protected const string _GreyscaleSetDefaultName = "Greyscale channel images";
        public const int SleepLengthPaused = 2000;
        public const int SleepLengthWaitingForThreadAvailability = 3000;
        public const int SleepLengthWaitingForScript = 1000;
        public const int SleepLengthWaitingForScriptToFinish = 5000;
        protected const int _MaximumLogLinesToParse = 65536;
        protected const int _SectionHeaderWidth = 80;
        // if the total elapsed time for execution of a particular script quantum
        // is less than this value, that elapsed time will *not* be added to the 
        // timetracker. This is mainly to avoid skewing the results when many of the 
        // quanta represent existing images which will not be reprocessed.
        public const int MinimumElapsedSecondsForTrackerAdd = 1;

        protected const int _MaximumIterationsWithoutOutput = 10;
        //protected int _MinimumSourceBandsInInputDirectory;
        protected TMSBEngineControllerConfiguration _EngineControllerConfig;
        //protected ImageCube _Cube;
        protected int _TotalImageCount;
        protected int _CurrentImageNumber;

        protected int _EngineNumber;
        protected TMSBEngineState _CurrentState;

        protected bool _ContinueScriptExecution;

        protected Thread _EngineThread;

        protected ArrayList _EngineCores;

        protected TimeTracker _Tracker;
        protected NumericValueTracker _MemoryUseTracker;
        protected NumericValueTracker _ProcessorTimeTracker;

        public static readonly object TemporaryFilePathLockObject = new object();
        public static readonly object ScriptGenerationLockObject = new object();
        public static readonly object LogWriteLockObject = new object();
        public readonly object EngineCoreArrayLockObject = new object();
        public readonly object EngineCoreStatusLockObject = new object();

        public delegate void DirectoryProcessingStartedHandler(object sender, DirectoryProcessingArgs d);
        public delegate void CubeScriptBeingGeneratedHandler(object sender, ThreadedEventArgs c);
        public delegate void CubeScriptGeneratedHandler(object sender, CubeScriptGeneratedEventArgs c);
        public delegate void ScriptExecutionCompleteHandler(object sender, ScriptExecutionEventArgs s);
        public delegate void ScriptExecutionAbortedHandler(object sender, ScriptExecutionEventArgs s);
        public delegate void ScriptExecutionInformationalMessageHandler(object sender, ScriptExecutionEventArgs s);
        public delegate void ScriptExecutionDebugMessageHandler(object sender, ScriptExecutionEventArgs s);
        public delegate void ImageStartedHandler(object sender, ImageEventArgs a);
        public delegate void ImageFinishedHandler(object sender, ImageFinishedEventArgs a);
        public delegate void ScriptExecutionErrorHandler(object sender, ScriptExecutionErrorArgs e);
        public delegate void DirectoryProcessingFinishedHandler(object sender, DirectoryProcessingArgs d);
        public delegate void EngineExecutionCompleteHandler(object sender, DirectoryProcessingArgs d);
        public delegate void WaitingForProcessingThreadHandler(object sender, EngineManagementEventArgs e);
        public delegate void ObtainedProcessingThreadHandler(object sender, EngineManagementEventArgs e);
        public delegate void UnhandledExceptionHandler(object sender, UnhandledExceptionArgs a);

        public event DirectoryProcessingStartedHandler DirectoryProcessingStarted;
        public event CubeScriptBeingGeneratedHandler CubeScriptBeingGenerated;
        public event CubeScriptGeneratedHandler CubeScriptGenerated;
        public event ScriptExecutionCompleteHandler ScriptExecutionComplete;
        public event ScriptExecutionInformationalMessageHandler ScriptExecutionInformationalMessageGenerated;
        public event ScriptExecutionDebugMessageHandler ScriptExecutionDebugMessageGenerated;
        public event ScriptExecutionAbortedHandler ScriptExecutionAborted;
        public event ImageStartedHandler ImageStarted;
        public event ImageFinishedHandler ImageFinished;
        public event ScriptExecutionErrorHandler ScriptExecutionError;
        public event DirectoryProcessingFinishedHandler DirectoryProcessingFinished;
        public event EngineExecutionCompleteHandler EngineExecutionComplete;
        public event WaitingForProcessingThreadHandler WaitingForProcessingThread;
        public event ObtainedProcessingThreadHandler ObtainedProcessingThread;
        public event UnhandledExceptionHandler UnhandledExceptionThrown;

        #region Properties

        public string BaseDirectory
        {
            get
            {
                return _BaseDirectory;
            }
            set
            {
                _BaseDirectory = value;
            }
        }

        public TimeTracker Tracker
        {
            get
            {
                return _Tracker;
            }
            set
            {
                _Tracker = value;
            }
        }

        //public string DaVinciCommand
        //{
        //    get
        //    {
        //        return _DaVinciCommand;
        //    }
        //    set
        //    {
        //        _DaVinciCommand = value;
        //    }
        //}

        //public int MinimumSourceBandsInInputDirectory
        //{
        //    get
        //    {
        //        return _MinimumSourceBandsInInputDirectory;
        //    }
        //    set
        //    {
        //        _MinimumSourceBandsInInputDirectory = value;
        //    }
        //}

        public TMSBEngineControllerConfiguration EngineControllerConfig
        {
            get
            {
                return _EngineControllerConfig;
            }
            set
            {
                _EngineControllerConfig = value;
            }
        }

        public int TotalImageCount
        {
            get
            {
                return _TotalImageCount;
            }
        }

        public int CurrentImageNumber
        {
            get
            {
                return _CurrentImageNumber;
            }
        }

        //public string CurrentStateDescription
        //{
        //    get
        //    {
        //        return _CurrentStateDescription;
        //    }
        //}

        public int EngineNumber
        {
            get
            {
                return _EngineNumber;
            }
            set
            {
                _EngineNumber = value;
            }
        }

        public int ActiveEngineCoreCount
        {
            get
            {
                UpdateEngineCoreStatus();
                return _EngineCores.Count;
            }
        }

        public TMSBEngineCore[] EngineCores
        {
            get
            {
                UpdateEngineCoreStatus();
                TMSBEngineCore protoCore = new TMSBEngineCore();
                lock (EngineCoreArrayLockObject)
                {
                    TMSBEngineCore[] cores = (TMSBEngineCore[])_EngineCores.ToArray(protoCore.GetType());
                    return cores;
                }
            }
        }

        public Thread EngineThread
        {
            get
            {
                return _EngineThread;
            }
            set
            {
                _EngineThread = value;
            }
        }

        public TMSBEngineState CurrentState
        {
            get
            {
                return _CurrentState;
            }
        }

        public NumericValueTracker MemoryUseTracker
        {
            get
            {
                return _MemoryUseTracker;
            }
        }

        public NumericValueTracker ProcessorTimeTracker
        {
            get
            {
                return _ProcessorTimeTracker;
            }
        }

        #endregion

        public TMSBEngine(TMSBEngineController parentController)
        {
            _ParentController = parentController;
            //BaseDirectory = Environment.CurrentDirectory;
            BaseDirectory = "ZZZ_BOGUS";
            _Tracker = new TimeTracker();
            _TotalImageCount = 0;
            _CurrentImageNumber = 0;
            //_CurrentStateDescription = "Initializing";
            _CurrentState = TMSBEngineState.Initializing;
            _EngineNumber = 0;
            _EngineCores = new ArrayList();
            _ContinueScriptExecution = true;
            _MemoryUseTracker = new NumericValueTracker();
            _ProcessorTimeTracker = new NumericValueTracker();
        }

        public TMSBEngine(TMSBEngineController parentController, int threadNumber, string baseDirectory, TMSBEngineControllerConfiguration ecConfig)
            : this(parentController)
        {
            _EngineNumber = threadNumber;
            BaseDirectory = baseDirectory;
            EngineControllerConfig = ecConfig;
        }

        ~TMSBEngine()
        {
            //KillThreads();
            if ((_EngineThread != null) && (_EngineThread.IsAlive))
            {
                _EngineThread.Abort();
            }
            _ParentController = null;
        }

        public void Pause()
        {
            if (CurrentState != TMSBEngineState.Running)
            {
                return;
            }
            _CurrentState = TMSBEngineState.Pausing;
        }

        public void Resume()
        {
            if (CurrentState != TMSBEngineState.Paused)
            {
                return;
            }
            _CurrentState = TMSBEngineState.Resuming;
        }

        public void Stop()
        {
            if (CurrentState == TMSBEngineState.Finished)
            {
                return;
            }
            _CurrentState = TMSBEngineState.Stopping;
        }

        public void ForceStop()
        {
            //lock (EngineCores)
            //{
                Stop();
                for (int i = 0; i < EngineCores.Length; i++)
                {
                    try
                    {
                        if (EngineCores[i] != null)
                        {
                            EngineCores[i].KillProcessingThreads();
                            EngineCores[i].AbortImmediately();
                        }
                    }
                    catch (Exception ex)
                    {
                        // do nothing - this only happens when the engine is closing down anyway
                    }
                }
            //}
        }

        public void ExecuteProcessingThreaded()
        {
            _EngineThread = new Thread(new ThreadStart(ExecuteProcessing));
            _EngineThread.Start();
        }

        public void ExecuteProcessing()
        {
            _Tracker.Start();
            _MemoryUseTracker = new NumericValueTracker();
            _CurrentState = TMSBEngineState.Running;
            try
            {
                TransformationProfileConfigurationFile tcFile = new TransformationProfileConfigurationFile(EngineControllerConfig.TransformationProfileConfigFilePath);
                TMSBColourConfigurationFile ccFile = new TMSBColourConfigurationFile(EngineControllerConfig.ColourConfigFilePath);
                TMSBInputConfigurationFile icFile = new TMSBInputConfigurationFile(EngineControllerConfig.InputConfigFilePath, tcFile.TransformationProfiles);
                TMSBProcessingConfigurationFile pcFile = new TMSBProcessingConfigurationFile(EngineControllerConfig.ProcessingConfigFilePath, tcFile.TransformationProfiles);
                TMSBOutputConfigurationFile ocFile = new TMSBOutputConfigurationFile(EngineControllerConfig.OutputConfigFilePath, tcFile.TransformationProfiles);

                bool scriptWasGeneratedByTMSB = false;
                //_TotalImageCount = 0;
                Tracker.Reset();
                string outputScriptFilePath = GetOutputScriptFilePath();
                if (DirectoryProcessingStarted != null)
                {
                    DirectoryProcessingArgs ds = new DirectoryProcessingArgs(EngineNumber, -1, BaseDirectory, Tracker.GetTotalElapsedTime());
                    DirectoryProcessingStarted(this, ds);
                }
                switch (_EngineControllerConfig.GenerateScripts)
                {
                    case ScriptGenerationMode.Always:
                        GenerateScript(tcFile, icFile, ccFile, pcFile, ocFile);
                        scriptWasGeneratedByTMSB = true;
                        break;
                    case ScriptGenerationMode.IfNotAlreadyPresent:
                        if (File.Exists(outputScriptFilePath))
                        {
                            GenerateScriptInformationalMessage(-1, "An existing script was found in the directory '" +
                                BaseDirectory + "'. The existing script will be used.");
                        }
                        else
                        {
                            GenerateScriptInformationalMessage(-1, "No existing script was found in the directory '" +
                                BaseDirectory + "'. A script will be generated.");
                            GenerateScript(tcFile, icFile, ccFile, pcFile, ocFile);
                            scriptWasGeneratedByTMSB = true;
                        }
                        break;
                    case ScriptGenerationMode.Never:
                        if (!File.Exists(outputScriptFilePath))
                        {
                            GenerateScriptInformationalMessage(-1, "No existing script was found in the directory '" +
                                BaseDirectory + "'. This directory will not be processed.");
                        }
                        break;
                }

                if ((_ContinueScriptExecution) && (EngineControllerConfig.ExecuteScripts) && (File.Exists(outputScriptFilePath)))
                {
                    ExecuteScript(scriptWasGeneratedByTMSB);
                }
                Tracker.Stop();
                if (DirectoryProcessingFinished != null)
                {
                    DirectoryProcessingArgs ds = new DirectoryProcessingArgs(EngineNumber, -1, BaseDirectory, Tracker.GetTotalElapsedTime());
                    DirectoryProcessingFinished(this, ds);
                }
            }
            catch (NotEnoughSourceSpectralBandsException nEx)
            {
                GenerateScriptAbortedEvent(-1, nEx.Message);
            }
            catch (Exception ex)
            {
                if ((ex is IOException) && (IOExceptionIsOutOfDiskSpaceError((IOException)ex)))
                {
                    GenerateScriptAbortedEvent(-1, GetInsufficientDiskSpaceErrorMessage(BaseDirectory, "TMSBEngine: ExecuteProcessing()", (IOException)ex));
                }
                else
                {
                    if (UnhandledExceptionThrown != null)
                    {
                        UnhandledExceptionArgs u = new UnhandledExceptionArgs(EngineNumber, -1, ex);
                        UnhandledExceptionThrown(this, u);
                    }
                }
            }
            finally
            {
                if (EngineExecutionComplete != null)
                {
                    DirectoryProcessingArgs ds = new DirectoryProcessingArgs(EngineNumber, -1, BaseDirectory, Tracker.GetTotalElapsedTime());
                    EngineExecutionComplete(this, ds);
                }
                _CurrentState = TMSBEngineState.Finished;
            }
        }

        protected void ExecuteScript(bool scriptWasGeneratedByTMSB)
        {
            string fullScriptPath = GetOutputScriptFilePath();
            if (!File.Exists(fullScriptPath))
            {
                throw new IOException("The script file '" + fullScriptPath + "' could not be re-read for execution as it does not appear to exist.");
            }

            // create a new log file
            string mLogPath = GetMasterLogFilePath();
            if (File.Exists(mLogPath))
            {
                try
                {
                    File.Delete(mLogPath);
                }
                catch (Exception ex)
                {
                    throw new IOException("Unable to delete the existing log file '" + mLogPath + "'. The error description was '" + ex.Message + "'.");
                }
            }

            ScriptQuantum[] scriptQuanta = new ScriptQuantum[0];

            try
            {
                int maximumImageNumberInScript = 0;
                TMSBDVScriptReader dvsr = new TMSBDVScriptReader(this, scriptWasGeneratedByTMSB, fullScriptPath);
                scriptQuanta = dvsr.GetScriptQuanta(ref maximumImageNumberInScript);
                // if the script was not generated by this instance of TMSB (or the count of images is zero 
                // for some other reason), use the total image count parsed out of the script
                if ((_TotalImageCount == 0) || (!scriptWasGeneratedByTMSB))
                {
                    _TotalImageCount = Math.Max(maximumImageNumberInScript, _TotalImageCount);
                }
            }
            catch (Exception ex)
            {
                throw new IOException("An error occurred while splitting the generated script. The error description was '" + ex.Message + "'.");
            }

            if (scriptQuanta.Length == 0)
            {
                throw new IOException("The script file '" + fullScriptPath + "' could not be split. No errors were reported, but processing cannot continue.");
            }

            // process quanta with no input or output first (these will usually be images that were already 
            // created during a previous run against the same directory)
            for (int i = 0; i < scriptQuanta.Length; i++)
            {
                while (CurrentState == TMSBEngineState.Paused)
                {
                    Thread.Sleep(SleepLengthPaused);
                    GenerateScriptDebugMessage(-1, "Engine is paused (loop: script quanta with no input or output).");
                }
                if ((scriptQuanta[i].RequiredImages.Length == 0) && (scriptQuanta[i].OutputImages.Length == 0))
                {
                    ExecuteQuantum(scriptWasGeneratedByTMSB, i, ref scriptQuanta[i]);
                }
                if (!_ContinueScriptExecution)
                {
                    break;
                }
            }

            if (_ContinueScriptExecution)
            {
                bool allQuantaProcessed = false;
                int iterationsWithNoProcessableImages = 0;
                Random rand = new Random();
                while (!allQuantaProcessed)
                {
                    bool foundUnprocessedQuanta = false;
                    bool processedAtLeastOneImage = false;
                    for (int i = 0; i < scriptQuanta.Length; i++)
                    {
                        while (CurrentState == TMSBEngineState.Paused)
                        {
                            Thread.Sleep(SleepLengthPaused);
                            GenerateScriptDebugMessage(-1, "Engine is paused (loop: script quanta).");
                        }
                        if (!scriptQuanta[i].Executed)
                        {
                            foundUnprocessedQuanta = true;
                            if (QuantumPrerequisitesExist(scriptQuanta[i]))
                            {
                                ExecuteQuantum(scriptWasGeneratedByTMSB, i, ref scriptQuanta[i]);
                                processedAtLeastOneImage = true;
                                // wait a random number of milliseconds to keep things from locking up too closely
                                Thread.Sleep(rand.Next(100));
                            }
                        }
                        if (!_ContinueScriptExecution)
                        {
                            break;
                        }
                    }
                    if (processedAtLeastOneImage)
                    {
                        iterationsWithNoProcessableImages = 0;
                    }
                    else
                    {
                        if (ActiveEngineCoreCount == 0)
                        {
                            iterationsWithNoProcessableImages++;
                        }
                        if (iterationsWithNoProcessableImages > _MaximumIterationsWithoutOutput)
                        {
                            StringBuilder aBuilder = new StringBuilder();
                            aBuilder.Append("An attempt was made to process all remaining images in the script '");
                            aBuilder.Append(fullScriptPath);
                            aBuilder.Append("' ");
                            aBuilder.Append(_MaximumIterationsWithoutOutput.ToString());
                            aBuilder.Append(" times, without finding any for which all prerequisite files were present. ");
                            if (scriptWasGeneratedByTMSB)
                            {
                                aBuilder.Append("It is likely that this is due to a bug in the automatic script generation process. ");
                                aBuilder.Append("Please consult the FAQ, and if necessary, contact the developer.");
                            }
                            else
                            {
                                aBuilder.Append("This script was not generated by TMSB during its current run. ");
                                aBuilder.Append("If it has been customized, it is likely that there is a problem with the customizations. ");
                                aBuilder.Append("If no customizations were made, consider restarting the processing of this directory, ");
                                aBuilder.Append("but enable the automatic generation of the script.");
                            }
                            GenerateScriptAbortedEvent(-1, aBuilder.ToString());
                        }
                    }
                    if (!foundUnprocessedQuanta)
                    {
                        allQuantaProcessed = true;
                    }
                    if (!_ContinueScriptExecution)
                    {
                        break;
                    }
                    Thread.Sleep(SleepLengthWaitingForThreadAvailability);
                    UpdateEngineCoreStatus();
                }
            }
            DateTime waitStart = DateTime.Now;
            bool waited = false;
            while ((_ContinueScriptExecution) && (_EngineCores.Count > 0))
            {
                string waitMessage = "Waiting for " + _EngineCores.Count.ToString() +
                    " sub-script";
                if (_EngineCores.Count > 1)
                {
                    waitMessage += "s";
                }
                waitMessage += " to finish executing.";
                if (waited)
                {
                    waitMessage += " Elapsed time: " + TimeTracker.GetLongFormattedTimeSpanString(DateTime.Now - waitStart);
                }
                else
                {
                    waited = true;
                }
                GenerateScriptInformationalMessage(-1, waitMessage);
                Thread.Sleep(SleepLengthWaitingForScriptToFinish);
                UpdateEngineCoreStatus();
            }
            Thread.Sleep(SleepLengthWaitingForScriptToFinish);
            if (ScriptExecutionComplete != null)
            {
                ScriptExecutionEventArgs sa = new ScriptExecutionEventArgs(EngineNumber, -1, BaseDirectory, fullScriptPath);
                ScriptExecutionComplete(this, sa);
            }
        }

        public void UpdateEngineCoreStatus()
        {
            //lock (EngineCoreStatusLockObject)
            //{
                for (int i = 0; i < _EngineCores.Count; i++)
                {
                    TMSBEngineCore core = (TMSBEngineCore)_EngineCores[i];
                    if (!core.ContinueEngineExecution)
                    {
                        _ContinueScriptExecution = false;
                    }
                    if (core.FinishedExecuting)
                    {
                        _EngineCores.RemoveAt(i);
                        i--;
                    }
                }
            //}

            Tracker.TimeEstimationScalingFactor = TimeTracker.GetTimeEstimationScalingFactor((double)_EngineCores.Count);

            UpdateEngineCorePerformanceData();
        }

        protected void UpdateEngineCorePerformanceData()
        {
            if (!_ParentController.EngineControllerConfiguration.CollectPerformanceData)
            {
                return;
            }
            //lock (EngineCoreStatusLockObject)
            //{
                double totalMemoryUse = 0.0d;
                double totalPPT = 0.0d;
                for (int i = 0; i < _EngineCores.Count; i++)
                {
                    TMSBEngineCore core = (TMSBEngineCore)_EngineCores[i];
                    try
                    {
                        double coreMem = core.MemoryUseTracker.CurrentValue;
                        totalMemoryUse += coreMem;
                        _MemoryUseTracker.UpdateMaximum(core.MemoryUseTracker.Maximum);
                        _MemoryUseTracker.UpdateMinimum(core.MemoryUseTracker.Minimum);
                    }
                    catch (Exception ex)
                    {
                        GenerateScriptDebugMessage(-1, "Unable to obtain the current memory in use by processing thread number "
                            + i.ToString() + ". The error description was '" + ex.Message + "'.");
                    }
                    try
                    {
                        if (!core.FinishedExecuting)
                        {
                            double corePPT = core.GetCurrentPercentProcessorTime();
                            totalPPT += corePPT;
                        }
                        _ProcessorTimeTracker.UpdateMaximum(core.ProcessorTimeTracker.Maximum);
                        _ProcessorTimeTracker.UpdateMinimum(core.ProcessorTimeTracker.Minimum);
                    }
                    catch (Exception ex)
                    {
                        GenerateScriptDebugMessage(-1, "Unable to obtain the current percent processor time in use by processing thread number "
                            + i.ToString() + ". The error description was '" + ex.Message + "'.");
                    }
                }
                _MemoryUseTracker.AddValue(totalMemoryUse);
                _ProcessorTimeTracker.AddValue(totalPPT);
            //}
        }

        protected void ExecuteQuantum(bool scriptWasGeneratedByTMSB, int scriptQuantumNumber, ref ScriptQuantum scriptQuantum)
        {
            scriptQuantum.Executed = true;

            if (CurrentState == TMSBEngineState.Stopping)
            {
                _CurrentState = TMSBEngineState.Stopped;
                _ContinueScriptExecution = false;
                GenerateScriptAbortedEvent(-1, "A stop command was received by the processing engine.");
                return;
            }

            if (CurrentState == TMSBEngineState.Pausing)
            {
                _CurrentState = TMSBEngineState.Paused;
            }

            if (CurrentState == TMSBEngineState.Resuming)
            {
                _CurrentState = TMSBEngineState.Running;
            }

            UpdateEngineCoreStatus();

            int nextThreadNumber = -1;
            DateTime waitStart = DateTime.Now;
            bool hasWaited = false;
            string quantumDescription = "";
            bool isRegistered = false;

            while (!isRegistered)
            {
                while (nextThreadNumber == -1)
                {
                    while (CurrentState == TMSBEngineState.Paused)
                    {
                        Thread.Sleep(SleepLengthPaused);
                        GenerateScriptDebugMessage(-1, "Engine is paused (loop: waiting for available processing thread).");
                    }

                    if ((CurrentState == TMSBEngineState.Stopping) || (CurrentState == TMSBEngineState.Stopped))
                    {
                        _ContinueScriptExecution = false;
                        GenerateScriptAbortedEvent(-1, "A stop command was received by the processing engine.");
                        return;
                    }

                    nextThreadNumber = _ParentController.GetNextAvailableProcessingThreadNumber();
                    TimeSpan waitTime = DateTime.Now - waitStart;
                    if (!hasWaited)
                    {
                        quantumDescription = GetExtendedQuantumInformationString(scriptQuantum);
                    }
                    if (nextThreadNumber == -1)
                    {
                        string waitMessage = "Waiting for an available processing thread to assign to the script quantum described as [" +
                            quantumDescription + "].";
                        if (hasWaited)
                        {
                            waitMessage += " Elapsed wait time: " + TimeTracker.GetLongFormattedTimeSpanString(waitTime);
                        }
                        GenerateWaitingForProcessingThreadEvent(waitMessage);
                        Thread.Sleep(SleepLengthWaitingForThreadAvailability);
                        hasWaited = true;
                    }
                    else
                    {
                        string waitMessage = "Assigned processing thread number " + nextThreadNumber.ToString() + " to the script quantum described as [" +
                            quantumDescription + "].";
                        if (hasWaited)
                        {
                            waitMessage += " Elapsed wait time: " + TimeTracker.GetLongFormattedTimeSpanString(waitTime);
                        }
                        GenerateObtainedProcessingThreadEvent(waitMessage);
                        hasWaited = false;
                    }
                }

                // this extra safety mechanism is here because during testing, the "thread-safe" lock code 
                // in GetNextAvailableProcessingThreadNumber() managed to return the same number to two 
                // Engines simultaneously. Maybe lock() doesn't work if two threads reach it at literally 
                // the exact same time?
                isRegistered = true;
                TMSBEngineCore newCore = new TMSBEngineCore(this, nextThreadNumber, scriptWasGeneratedByTMSB, scriptQuantumNumber, scriptQuantum);
                try
                {
                    _ParentController.RegisterEngineCore(nextThreadNumber, ref newCore);
                }
                catch (Exception ex)
                {
                    isRegistered = false;
                    newCore = null;
                }

                if (isRegistered)
                {
                    _EngineCores.Add(newCore);
                    newCore.ExecuteQuantumThreaded();
                }
            }

        }

        protected bool QuantumPrerequisitesExist(ScriptQuantum quantum)
        {
            ArrayList missingList = new ArrayList();
            ArrayList foundList = new ArrayList();
            bool result = true;
            foreach (string prereq in quantum.RequiredImages)
            {
                string pPath = BaseDirectory + Path.DirectorySeparatorChar + prereq;

                if (File.Exists(pPath))
                {
                    if (!foundList.Contains(pPath))
                    {
                        foundList.Add(pPath);
                    }
                }
                else
                {
                    result = false;
                    if (!missingList.Contains(pPath))
                    {
                        missingList.Add(pPath);
                    }
                }
            }
            // this is a fairly involved string, so only generate it if it will actually be used
            if (ScriptExecutionDebugMessageGenerated != null)
            {
                StringBuilder builder = new StringBuilder();
                builder.Append("Found ");
                builder.Append(foundList.Count);
                builder.Append(" of ");
                builder.Append(quantum.RequiredImages.Length);
                builder.Append(" required images.");
                builder.Append(GetQuantumImageListString(missingList, "Missing images"));
                builder.Append(GetQuantumImageListString(foundList, "Found images"));
                if (missingList.Count > 0)
                {
                    builder.Append(" This quantum will not be processed during this iteration.");
                }
                else
                {
                    builder.Append(" This quantum will be queued for processing.");
                }
                GenerateScriptDebugMessage(-1, builder.ToString());
                //GenerateScriptInformationalMessage(-1, builder.ToString());
            }
            return result;
        }

        protected string GetQuantumImageListString(ArrayList imageList, string listName)
        {
            StringBuilder builder = new StringBuilder();

            if (imageList.Count > 0)
            {
                builder.Append(" ");
                builder.Append(listName);
                builder.Append(" (");
                builder.Append(imageList.Count);
                builder.Append("): [");
                foreach (string imageName in imageList)
                {
                    builder.Append("'");
                    builder.Append(imageName);
                    builder.Append("', ");
                }
                builder.Remove(builder.Length - 2, 1);
                builder.Append("]. ");
            }

            return builder.ToString();
        }

        public void WriteRelevantLogTextToMasterLog(int globalThreadNumber, string logText, out string relevantText)
        {
            StringBuilder builder = new StringBuilder(Environment.NewLine);
            StringBuilder rtBuilder = new StringBuilder();
            bool foundMarker = false;

            string[] lines = logText.Split(new string[] { Environment.NewLine }, StringSplitOptions.None);

            foreach (string line in lines)
            {
                if (line.Contains(TMSBDVScriptReader.LogMarkerText))
                {
                    if (foundMarker)
                    {
                        break;
                    }
                    foundMarker = true;
                }
                else
                {
                    if (foundMarker)
                    {
                        rtBuilder.AppendLine(line);
                        if (EngineControllerConfig.GlobalThreadCount > 1)
                        {
                            builder.Append(String.Format("Thread {0:00}\t", globalThreadNumber));
                            builder.Append(line);
                            builder.Append(Environment.NewLine);
                        }
                        else
                        {
                            builder.AppendLine(line);
                        }
                    }
                }
            }

            relevantText = rtBuilder.ToString();
            if (relevantText.Trim() != "")
            {
                // prevent multiple threads from writing to the log file simultaneously
                lock (LogWriteLockObject)
                {
                    WriteTextToFile(GetMasterLogFilePath(), builder.ToString(), true);
                }
            }
        }

        protected string GetLogText(StreamReader iReader)
        {
            StringBuilder builder = new StringBuilder(Environment.NewLine);

            int currentLineNumber = 0;

            while (!iReader.EndOfStream)
            {
                string line = iReader.ReadLine();
                builder.AppendLine(line);

                if (line.Contains(TMSBDVScriptReader.LogEndMarkerText))
                {
                    break;
                }

                currentLineNumber++;
                if (currentLineNumber >= _MaximumLogLinesToParse)
                {
                    break;
                }
            }

            string result = builder.ToString();
            return result;
        }

        public void WriteTextToFile(string filePath, string text, bool append)
        {
            try
            {
                FileMode fMode = FileMode.Create;
                if (append)
                {
                    fMode = FileMode.Append;
                }
                FileStream oStream = new FileStream(filePath, fMode, FileAccess.Write);
                StreamWriter oWriter = new StreamWriter(oStream);
                WriteToStream(oWriter, text);
                oWriter.Close();
                oStream.Close();
                try
                {
                    oWriter.Dispose();
                    oStream.Dispose();
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            catch (Exception ex)
            {
                throw new IOException("An error occurred while writing to the file '" + filePath +
                    "'. The error description was '" + ex.Message + "'.");
            }
        }

        public void DeleteFile(string filePath)
        {
            if (File.Exists(filePath))
            {
                try
                {
                    File.Delete(filePath);
                }
                catch (Exception ex)
                {
                    // this is a non-critical error, so it really should just log a message
                    //throw new IOException("Unable to delete the file '" + filePath + "'. The error description was '" +
                    //    ex.Message + "'.");
                    GenerateScriptInformationalMessage(-1, "Unable to delete the file '" + filePath + "'. The error description was '" +
                        ex.Message + "'. This is a non-critical error.");
                }
            }
        }

        public string GetTemporaryScriptFilePath(string baseDirectory)
        {
            return GetTemporaryFilePath(baseDirectory, _TemporaryDVScriptNamePrefix, "dv");
        }

        protected string GetTemporaryFilePath(string baseDirectory, string namePrefix, string fileExtension)
        {
            lock (TemporaryFilePathLockObject)
            {
                string baseString = baseDirectory + Path.DirectorySeparatorChar +
                    namePrefix + "{0:000000}." + fileExtension;
                for (int i = 0; i < 1000000; i++)
                {
                    string result = String.Format(baseString, i);
                    if (!File.Exists(result))
                    {
                        CreateEmptyFile(result);
                        return result;
                    }
                }
                throw new IOException("Could not obtain a name for the temporary file with the format '" + baseString + "'.");
            }
        }

        protected void CreateEmptyFile(string filePath)
        {
            try
            {
                FileStream tStream = new FileStream(filePath, FileMode.Create, FileAccess.Write);
                StreamWriter tWriter = new StreamWriter(tStream);
                tWriter.Write("");
                tWriter.Close();
                tStream.Close();
            }
            catch (Exception ex)
            {
                throw new IOException("Unable to create the file '" + filePath + "'. The error description was '" + ex.Message + "'.", ex);
            }
        }

        public string GetOutputScriptFilePath()
        {
            return BaseDirectory + Path.DirectorySeparatorChar + _DVScriptFileName;
        }

        public string GetMasterLogFilePath()
        {
            return BaseDirectory + Path.DirectorySeparatorChar + _DVScriptLogFileName;
        }

        public void GenerateScript(TransformationProfileConfigurationFile tcFile, TMSBInputConfigurationFile icFile, TMSBColourConfigurationFile ccFile, 
            TMSBProcessingConfigurationFile pcFile, TMSBOutputConfigurationFile ocFile)
        {
            // this section is locked to avoid stomping on other threads. I'm not sure why this section isn't thread-safe
            // ...but it really isn't.
            lock (ScriptGenerationLockObject)
            {
                string stepName = "";

                if (CubeScriptBeingGenerated != null)
                {
                    CubeScriptBeingGenerated(this, new ThreadedEventArgs(EngineNumber, -1));
                }

                GenerateScriptDebugMessage(-1, "Generating image cube script.");

                if (!Directory.Exists(BaseDirectory))
                {
                    throw new DirectoryNotFoundException("Error: the input directory '" + BaseDirectory + "' could not be found.");
                }

                Directory.SetCurrentDirectory(BaseDirectory);

                GenerateScriptDebugMessage(-1, "Instantiating image cube.");

                ImageCube cube = new ImageCube(this, BaseDirectory, BaseDirectory, tcFile.TransformationProfiles, icFile.InputConfig,
                    ccFile.ColourConfig, pcFile.ProcessingConfig, ocFile.OutputConfig);

                if (cube.InputConfig.SpectralBands.Length < EngineControllerConfig.MinimumSourceBandsInInputDirectory)
                {
                    //throw new Exception("Not enough source spectral bands were found in the directory '" + BaseDirectory +
                    //    "' to allow processing. " + cube.InputConfig.SpectralBands.Length.ToString() +
                    //    "bands were found in source files, but " + EngineControllerConfig.MinimumSourceBandsInInputDirectory.ToString() +
                    //    "bands are required using the current configuration. If this directory should be processed, " +
                    //    "verify that the correct input configuration is in use and the source files are named correctly.");
                    throw new NotEnoughSourceSpectralBandsException(BaseDirectory, cube.InputConfig.SpectralBands.Length,
                        EngineControllerConfig.MinimumSourceBandsInInputDirectory);
                }

                GenerateScriptDebugMessage(-1, "Obtaining script output file path.");

                string dvScriptFilePath = GetOutputScriptFilePath();

                try
                {
                    GenerateScriptDebugMessage(-1, "Preparing to write the script to the file '" + dvScriptFilePath + "'.");

                    FileStream oStream = new FileStream(dvScriptFilePath, FileMode.Create, FileAccess.Write);
                    StreamWriter oWriter = new StreamWriter(oStream);

                    stepName = "generating the channels for the image cube";
                    if (_ContinueScriptExecution)
                    {
                        try
                        {
                            cube.GenerateChannels();
                        }
                        catch (Exception ex)
                        {
                            throw new Exception("An error occurred while " + stepName + ". The error message was '" +
                                ex.Message + "'.");
                        }
                    }
                    else
                    {
                        GenerateScriptGenerationStepBypassedDebugMessage(stepName);
                    }


                    stepName = "generating the permutations for the image cube";
                    if (_ContinueScriptExecution)
                    {
                        try
                        {
                            cube.GeneratePermutations();
                        }
                        catch (Exception ex)
                        {
                            throw new Exception("An error occurred while " + stepName + ". The error message was '" +
                                ex.Message + "'.");
                        }
                    }
                    else
                    {
                        GenerateScriptGenerationStepBypassedDebugMessage(stepName);
                    }


                    stepName = "obtaining the sorted greyscale channels from the image cube";
                    if (_ContinueScriptExecution)
                    {
                        try
                        {
                            cube.GetSortedGreyscaleChannels();
                        }
                        catch (Exception ex)
                        {
                            throw new Exception("An error occurred while " + stepName + ". The error message was '" +
                                ex.Message + "'.");
                        }
                    }
                    else
                    {
                        GenerateScriptGenerationStepBypassedDebugMessage(stepName);
                    }


                    int imageNumber = 1;
                    int filesToCreate = 0;
                    int filesAlreadyExisting = 0;


                    if (_ContinueScriptExecution)
                    {
                        GenerateScriptDebugMessage(-1, "Beginning main script-writing sequence.");

                        // write the script processor path for Unix/Linux-style operating systems
                        if (EngineControllerConfig.DaVinciCommand != "")
                        {
                            WriteLineToStream(oWriter, GetDaVinciScriptHeader());
                        }

                        WriteSectionHeader(oWriter, _GreyscaleSetDefaultName);
                    }
                    else
                    {
                        GenerateScriptGenerationStepBypassedDebugMessage("beginning main script-writing sequence");
                    }

                    int suppressedPermCount = 0;
                    int firstSuppressedImageNumber = 0;
                    int lastSuppressedImageNumber = 0;
                    bool lastImageWasSuppressed = false;
                    int lastWrittenSuppressedImageNumber = -1;
                    StringBuilder builder = new StringBuilder();


                    if (_ContinueScriptExecution)
                    {
                        GenerateScriptDebugMessage(-1, "Writing greyscale channel sets to the script.");

                        foreach (GreyscaleChannelSet channelSet in cube.GreyscaleChannelSets)
                        {
                            GenerateScriptDebugMessage(-1, "Writing greyscale channel set '" + channelSet.ChannelSetName + "' to the script.");

                            builder = new StringBuilder();
                            int firstImageNumber = imageNumber;
                            int lastImageNumber = imageNumber + (channelSet.Channels.Length - 1);

                            string gsSetName = _GreyscaleSetDefaultName;
                            if (channelSet.ChannelSetName.Trim() == "")
                            {
                                if (channelSet.Channels.Length == 1)
                                {
                                    gsSetName = channelSet.Channels[0].ImageName;
                                }
                                else
                                {
                                    gsSetName = "Multiple Images";
                                }
                            }
                            else
                            {
                                gsSetName = channelSet.ChannelSetName;
                            }

                            builder.Append(GetImageSetBegin(imageNumber, gsSetName, firstImageNumber, lastImageNumber));

                            builder.Append(channelSet.GetSetCreationScriptCode(cube.OutputConfig.RegenerateExistingFiles));
                            filesToCreate += channelSet.FileCountWillBeWritten;
                            filesAlreadyExisting += channelSet.FileCountWillBeSkipped;

                            builder.Append(GetImageSetEnd(imageNumber, gsSetName, firstImageNumber, lastImageNumber));

                            WriteCurrentCodeAndSuppressedBuffer(ref oWriter, ref builder, channelSet.FileCountWillBeWritten, ref imageNumber,
                                ref firstImageNumber, ref lastImageNumber, ref lastImageWasSuppressed, ref firstSuppressedImageNumber,
                                ref lastSuppressedImageNumber, ref suppressedPermCount, ref lastWrittenSuppressedImageNumber);

                            imageNumber += channelSet.Channels.Length;
                            GenerateScriptDebugMessage(-1, "Done writing greyscale channel set '" + channelSet.ChannelSetName + "' to the script.");
                        }

                        // if there's anything in the buffer, write it now
                        if (suppressedPermCount > 0)
                        {
                            int notUsed = 1;
                            bool forceTrue = true;
                            builder = new StringBuilder();

                            WriteCurrentCodeAndSuppressedBuffer(ref oWriter, ref builder, 1, ref notUsed,
                                ref notUsed, ref notUsed, ref forceTrue, ref firstSuppressedImageNumber,
                                ref lastSuppressedImageNumber, ref suppressedPermCount, ref lastWrittenSuppressedImageNumber);
                        }

                        WriteSuppressedImageCount(ref oWriter, suppressedPermCount);

                        GenerateScriptDebugMessage(-1, "Done writing greyscale channel sets to the script.");
                    }

                    WriteImageSetScriptCode(oWriter, cube, "Three-Channel False-Colour Images", cube.ThreeChannelImages, ref imageNumber, ref filesToCreate, ref filesAlreadyExisting);

                    WriteImageSetScriptCode(oWriter, cube, "Tinted Channels", cube.TintedChannels, ref imageNumber, ref filesToCreate, ref filesAlreadyExisting);

                    GenerateScriptGenerationStepBypassedDebugMessage("generating tinted image script code");

                    WriteImageSetScriptCode(oWriter, cube, "Gradient-Mapped Images", cube.ColourGradientImages, ref imageNumber, ref filesToCreate, ref filesAlreadyExisting);

                    WriteImageSetScriptCode(oWriter, cube, "Decorrelation-Stretch Images", cube.DCSImages, ref imageNumber, ref filesToCreate, ref filesAlreadyExisting);

                    WriteImageSetScriptCode(oWriter, cube, "Luma/Colour Images", cube.LumaColourImages, ref imageNumber, ref filesToCreate, ref filesAlreadyExisting);

                    _TotalImageCount = (imageNumber - 1);

                    if (_ContinueScriptExecution)
                    {
                        GenerateScriptDebugMessage(-1, "Writing summary information to the script.");

                        imageNumber--;
                        WriteLineToStream(oWriter, String.Format("# Total image variations: {0}.", imageNumber));
                        //_TotalImageCount = imageNumber;
                        WriteLineToStream(oWriter, String.Format("# {0} files will be created.", filesToCreate));
                        WriteLineToStream(oWriter, String.Format("# {0} files already exist and will not be overwritten.", Math.Max(filesAlreadyExisting, 0)));

                        WriteToStream(oWriter, Environment.NewLine);
                        WriteToStream(oWriter, Environment.NewLine);
                        WriteToStream(oWriter, Environment.NewLine);

                        WriteLineToStream(oWriter, "# Execution Complete");
                        WriteToStream(oWriter, Environment.NewLine);
                    }
                    else
                    {
                        GenerateScriptGenerationStepBypassedDebugMessage("writing summary information");
                    }

                    oWriter.Close();
                    oStream.Close();
                    try
                    {
                        oWriter.Dispose();
                        oStream.Dispose();
                    }
                    catch (Exception ex)
                    {
                        // do nothing
                    }
                }
                catch (Exception ex)
                {
                    bool foundReason = false;
                    if (ex is IOException)
                    {
                        if (IOExceptionIsOutOfDiskSpaceError((IOException)ex))
                        {
                            foundReason = true;
                            GenerateScriptAbortedEvent(-1, GetInsufficientDiskSpaceErrorMessage(BaseDirectory, "TMSBEngine: GenerateScript()", (IOException)ex));
                        }
                    }
                    if (!foundReason)
                    {
                        throw new Exception("An error occurred while generating the file '" + dvScriptFilePath + "'. The error message was '" + ex.Message + "'.");
                    }
                }
                finally
                {
                    Tracker.Reset();
                    Tracker.TotalElementCount = _TotalImageCount;
                    if (CubeScriptGenerated != null)
                    {
                        CubeScriptGeneratedEventArgs csa = new CubeScriptGeneratedEventArgs(EngineNumber, -1, _TotalImageCount);
                        CubeScriptGenerated(this, csa);
                    }
                }
                if (_ContinueScriptExecution)
                {
                    GenerateScriptDebugMessage(-1, "Done generating image cube script.");
                }
                else
                {
                    GenerateScriptDebugMessage(-1, "Generation of the image cube script was aborted. See the preceeding messages for the exact location.");
                }
            }
        }

        protected void GenerateScriptGenerationStepInitiatedDebugMessage(string stepName)
        {
            GenerateScriptDebugMessage(-1, "Script generation step '" + stepName + "' initiated.");
        }

        protected void GenerateScriptGenerationStepBypassedDebugMessage(string stepName)
        {
            GenerateScriptDebugMessage(-1, "Script generation step '" + stepName + "' was bypassed.");
        }

        public void GenerateImageStartedEvent(int globalThreadNumber, ScriptQuantum scriptQuantum)
        {
            if (ImageStarted != null)
            {
                ImageEventArgs isa = new ImageEventArgs(EngineNumber, globalThreadNumber, scriptQuantum.SectionName, 
                    scriptQuantum.ImageName, scriptQuantum.FirstImageNumber, scriptQuantum.TotalImages, TotalImageCount);
                ImageStarted(this, isa);
            }
        }

        public void GenerateImageFinishedEvent(int globalThreadNumber, ScriptQuantum scriptQuantum, TimeSpan elapsed, NumericValueTracker memoryUseTracker,
            NumericValueTracker processorTimeTracker, int globalImageNumber)
        {
            if (ImageFinished != null)
            {
                ImageFinishedEventArgs ifa = new ImageFinishedEventArgs(EngineNumber, globalThreadNumber, scriptQuantum.SectionName,
                    scriptQuantum.ImageName, scriptQuantum.LastImageNumber, scriptQuantum.TotalImages, TotalImageCount, elapsed, memoryUseTracker, processorTimeTracker,
                    globalImageNumber, BaseDirectory, scriptQuantum);
                ImageFinished(this, ifa);
            }
        }

        public void GenerateScriptExecutionErrorEvent(int globalThreadNumber, string directoryPath, string scriptPath, string[] errorMessages)
        {
            if (ScriptExecutionError != null)
            {
                ScriptExecutionErrorArgs s = new ScriptExecutionErrorArgs(EngineNumber, globalThreadNumber, directoryPath, scriptPath, errorMessages);
                ScriptExecutionError(this, s);
            }
        }

        public void GenerateScriptExecutionErrorEvent(int globalThreadNumber, string directoryPath, string scriptPath, string errorMessage)
        {
            GenerateScriptExecutionErrorEvent(globalThreadNumber, directoryPath, scriptPath, new string[] { errorMessage });
        }

        public void GenerateWaitingForProcessingThreadEvent(string information)
        {
            if (WaitingForProcessingThread != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(EngineNumber, -1, information);
                WaitingForProcessingThread(this, e);
            }
        }

        public void GenerateObtainedProcessingThreadEvent(string information)
        {
            if (ObtainedProcessingThread != null)
            {
                EngineManagementEventArgs e = new EngineManagementEventArgs(EngineNumber, -1, information);
                ObtainedProcessingThread(this, e);
            }
        }

        public void GenerateScriptAbortedEvent(int globalThreadNumber, string reasonMessage)
        {
            if (ScriptExecutionAborted != null)
            {
                ScriptExecutionEventArgs s = new ScriptExecutionEventArgs(EngineNumber, globalThreadNumber, BaseDirectory, reasonMessage);
                ScriptExecutionAborted(this, s);
            }
            _ContinueScriptExecution = false;
            Stop();
        }

        public void GenerateScriptInformationalMessage(int globalThreadNumber, string message)
        {
            if (ScriptExecutionInformationalMessageGenerated != null)
            {
                ScriptExecutionEventArgs s = new ScriptExecutionEventArgs(EngineNumber, globalThreadNumber, BaseDirectory, message);
                ScriptExecutionInformationalMessageGenerated(this, s);
            }
        }

        public void GenerateScriptDebugMessage(int globalThreadNumber, string message)
        {
            if (ScriptExecutionDebugMessageGenerated != null)
            {
                ScriptExecutionEventArgs s = new ScriptExecutionEventArgs(EngineNumber, globalThreadNumber, BaseDirectory, message);
                ScriptExecutionDebugMessageGenerated(this, s);
            }
        }

        protected string GetPrePostFix(int imageNumber, string imageName, int firstImageNumber, int lastImageNumber)
        {
            string prePostFix = "";
            if (lastImageNumber > firstImageNumber)
            {
                prePostFix = String.Format("image numbers {0} - ", firstImageNumber) +
                    String.Format("{0}", lastImageNumber);
            }
            else
            {
                prePostFix = String.Format("image number {0}", imageNumber);
            }

            prePostFix += " - '" + imageName + "'";

            return prePostFix;
        }

        protected string GetImageSetBegin(int imageNumber, string imageName, int firstImageNumber, int lastImageNumber)
        {
            StringBuilder builder = new StringBuilder();
            string prePostFix = GetPrePostFix(imageNumber, imageName, firstImageNumber, lastImageNumber);
            builder.Append("# Begin ");
            builder.Append(prePostFix);
            builder.Append(Environment.NewLine);
            builder.Append("printf(\"Begin " + prePostFix + "\\n\")");
            builder.Append(Environment.NewLine);
            builder.Append("echodatetime()");
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        protected string GetImageSetEnd(int imageNumber, string imageName, int firstImageNumber, int lastImageNumber)
        {
            StringBuilder builder = new StringBuilder();
            string prePostFix = GetPrePostFix(imageNumber, imageName, firstImageNumber, lastImageNumber);
            builder.Append("# End ");
            builder.Append(prePostFix);
            builder.Append(Environment.NewLine);
            builder.Append("printf(\"End " + prePostFix + "\\n\")");
            builder.Append(Environment.NewLine);
            builder.Append("echodatetime()");
            builder.Append(Environment.NewLine);
            builder.Append("printf(\"\\n\")");
            builder.Append(Environment.NewLine);
            builder.Append(Environment.NewLine);
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        protected void WriteSectionHeader(StreamWriter oWriter, string sectionName)
        {
            StringBuilder headerDividerBuilder = new StringBuilder();
            for (int i = 0; i < _SectionHeaderWidth; i++)
            {
                headerDividerBuilder.Append('#');
            }
            StringBuilder headerNameBuilder = new StringBuilder();
            int spacingWidth = ((_SectionHeaderWidth - 2) - sectionName.Length);
            int spacingWidthLeft = spacingWidth / 2;
            int spacingWidthRight = spacingWidthLeft;
            if ((spacingWidth % 2) == 1)
            {
                spacingWidth--;
                spacingWidthLeft = spacingWidth / 2;
                spacingWidthRight = spacingWidthLeft;
                spacingWidthLeft++;
            }
            headerNameBuilder.Append('#');
            for (int i = 0; i < spacingWidthLeft; i++)
            {
                headerNameBuilder.Append(' ');
            }
            headerNameBuilder.Append(sectionName);
            for (int i = 0; i < spacingWidthRight; i++)
            {
                headerNameBuilder.Append(' ');
            }
            headerNameBuilder.Append('#');

            WriteLineToStream(oWriter, headerDividerBuilder.ToString());
            WriteLineToStream(oWriter, headerNameBuilder.ToString());
            WriteLineToStream(oWriter, headerDividerBuilder.ToString());

            string completeSectionName = "Section '" + sectionName + "'";

            WriteLineToStream(oWriter, "# " + completeSectionName);
            WriteLineToStream(oWriter, "printf(\"" + completeSectionName + "\\n\")");

            WriteToStream(oWriter, Environment.NewLine);
            WriteToStream(oWriter, Environment.NewLine);
            WriteToStream(oWriter, Environment.NewLine);
        }

        protected void WriteImageSetScriptCode(StreamWriter oWriter, ImageCube cube, string imageSetName, OutputImage[] permutations, ref int imageNumber, ref int filesToCreate, ref int filesAlreadyExisting)
        {
            if (!_ContinueScriptExecution)
            {
                GenerateScriptGenerationStepBypassedDebugMessage("imageSetName");
                return;
            }
            GenerateScriptDebugMessage(-1, "Writing the image set '" + imageSetName + "' to the script.");

            WriteToStream(oWriter, Environment.NewLine);
            WriteToStream(oWriter, Environment.NewLine);
            WriteToStream(oWriter, Environment.NewLine);

            WriteSectionHeader(oWriter, imageSetName);

            int suppressedPermCount = 0;
            int firstSuppressedImageNumber = 0;
            int lastSuppressedImageNumber = 0;
            int lastWrittenSuppressedImageNumber = -1;
            bool lastImageWasSuppressed = false;
            StringBuilder builder = new StringBuilder();

            foreach (OutputImage perm in permutations)
            {
                GenerateScriptDebugMessage(-1, "Writing the image permutation '" + perm.ImageName + "' to the script.");
                builder = new StringBuilder();

                builder.Append(GetImageSetBegin(imageNumber, perm.ImageName, imageNumber, imageNumber));

                builder.Append(perm.GetCreationScriptCode("imageObject", cube.OutputConfig.RegenerateExistingFiles));
                filesToCreate += perm.FileCountWillBeWritten;
                filesAlreadyExisting += (perm.OutputFormats.Length - perm.FileCountWillBeWritten);

                builder.Append(GetImageSetEnd(imageNumber, perm.ImageName, imageNumber, imageNumber));

                WriteCurrentCodeAndSuppressedBuffer(ref oWriter, ref builder, perm.FileCountWillBeWritten, ref imageNumber,
                    ref imageNumber, ref imageNumber, ref lastImageWasSuppressed, ref firstSuppressedImageNumber,
                    ref lastSuppressedImageNumber, ref suppressedPermCount, ref lastWrittenSuppressedImageNumber);

                imageNumber++;
                GenerateScriptDebugMessage(-1, "Done writing the image permutation '" + perm.ImageName + "' to the script.");
            }

            // if there's anything in the buffer, write it now
            if (suppressedPermCount > 0)
            {
                int notUsed = 1;
                bool forceTrue = true;
                builder = new StringBuilder();

                WriteCurrentCodeAndSuppressedBuffer(ref oWriter, ref builder, 1, ref notUsed,
                    ref notUsed, ref notUsed, ref forceTrue, ref firstSuppressedImageNumber,
                    ref lastSuppressedImageNumber, ref suppressedPermCount, ref lastWrittenSuppressedImageNumber);
            }

            WriteSuppressedImageCount(ref oWriter, suppressedPermCount);

            GenerateScriptDebugMessage(-1, "Done writing the image set '" + imageSetName + "' to the script.");
        }

        protected void WriteSuppressedImageCount(ref StreamWriter oWriter, int suppressedPermCount)
        {
            if (suppressedPermCount > 0)
            {
                WriteLineToStream(oWriter, Environment.NewLine);
                WriteLineToStream(oWriter, String.Format("# Information: {0} images were suppressed because all of their output files already exist.", suppressedPermCount));
                WriteLineToStream(oWriter, Environment.NewLine);
            }
        }

        protected void WriteCurrentCodeAndSuppressedBuffer(ref StreamWriter oWriter, ref StringBuilder builder, int fileCountWillBeWritten,
            ref int imageNumber, ref int firstImageNumber, ref int lastImageNumber, ref bool lastImageWasSuppressed,
            ref int firstSuppressedImageNumber, ref int lastSuppressedImageNumber, ref int suppressedPermCount, 
            ref int lastWrittenSuppressedImageNumber)
        {
            bool writeToScript = true;

            if ((!EngineControllerConfig.WriteScriptCodeWhenNoImagesWillBeGenerated) && (fileCountWillBeWritten == 0))
            {
                writeToScript = false;
                if (lastImageWasSuppressed)
                {
                    lastSuppressedImageNumber += Math.Max(1, (lastImageNumber - firstImageNumber));
                }
                else
                {
                    firstSuppressedImageNumber = firstImageNumber;
                    lastSuppressedImageNumber = lastImageNumber;
                }
                lastImageWasSuppressed = true;
                suppressedPermCount++;
            }
            else
            {
                // if there are any queued suppressed images, write their code now
                if ((lastImageWasSuppressed) && (firstSuppressedImageNumber > lastWrittenSuppressedImageNumber))
                {
                    StringBuilder sB = new StringBuilder();
                    string sImageName = "1 pre-existing image";
                    int sc = (lastSuppressedImageNumber - firstSuppressedImageNumber) + 1;
                    if (sc > 1)
                    {
                        sImageName = String.Format("{0} pre-existing images", sc);
                    }

                    sB.Append(GetImageSetBegin(firstSuppressedImageNumber, sImageName, firstSuppressedImageNumber, lastSuppressedImageNumber));
                    sB.Append("# Create the '");
                    sB.Append(sImageName);
                    if (sc == 1)
                    {
                        sB.Append("' image");
                    }
                    else
                    {
                        sB.Append("' images");
                    }
                    sB.Append(Environment.NewLine);
                    sB.AppendLine("# All output files for these image variations already exist in the output directory.");
                    sB.AppendLine("# To force regeneration of the existing files, enable the 'Overwrite Existing Images' option in TMSB.");
                    sB.Append(GetImageSetEnd(firstSuppressedImageNumber, sImageName, firstSuppressedImageNumber, lastSuppressedImageNumber));
                    WriteToStream(oWriter, sB.ToString());
                    lastWrittenSuppressedImageNumber = lastSuppressedImageNumber;
                }
                lastImageWasSuppressed = false;
            }

            if (writeToScript)
            {
                WriteToStream(oWriter, builder.ToString());
            }
        }

        public void WriteLineToStream(StreamWriter oWriter, string text)
        {
            WriteToStream(oWriter, text + Environment.NewLine);
        }

        public void WriteToStream(StreamWriter oWriter, string text)
        {
            bool fatalErrorOccurred = false;
            string fatalErrorMessage = "";
            Exception fatalErrorException = new Exception();

            try
            {
                oWriter.Write(text);
            }
            catch (IOException ioEx)
            {
                if (IOExceptionIsOutOfDiskSpaceError(ioEx))
                {
                    fatalErrorMessage = GetInsufficientDiskSpaceErrorMessage(BaseDirectory, "TMSBEngine: WriteToStream()", ioEx);
                }
                else
                {
                    //fatalErrorMessage = "";
                }
                fatalErrorException = ioEx;
                fatalErrorOccurred = true;
            }
            catch (Exception ex)
            {
                //fatalErrorMessage = "An error occurred while writing to a stream. This specific ";
                fatalErrorException = ex;
                fatalErrorOccurred = true;
            }

            if (fatalErrorOccurred)
            {
                if (fatalErrorMessage == "")
                {
                    fatalErrorMessage = "An error occurred while writing to a stream. The specific error description was '" +
                        fatalErrorException.Message.Trim() +
                        "'. This specific error is not currently recognized, but is fatal, and execution will be halted.";
                }
                GenerateScriptAbortedEvent(-1, fatalErrorMessage);
            }
        }

        public static bool IOExceptionIsOutOfDiskSpaceError(IOException ioEx)
        {
            // nothing fancy for now
            // may need to become fancy for internationalized version
            if (ioEx.Message.ToLower().Contains("enough space"))
            {
                return true;
            }
            return false;
        }

        public string GetDaVinciScriptHeader()
        {
            StringBuilder builder = new StringBuilder();
            builder.Append("#!");
            builder.Append(Argument.QuoteArgumentIfNecessary(EngineControllerConfig.DaVinciCommand));
            builder.Append(" ");
            builder.Append(EngineControllerConfig.DaVinciCommandArgs);
            builder.Append(Environment.NewLine);
            string baseDirectory = Path.GetDirectoryName(EngineControllerConfig.DaVinciCommand);
            baseDirectory = baseDirectory.Replace('\\', '/');
            switch (EngineControllerConfig.ExecutionPlatform)
            {
                case ExecutionPlatform.Windows:
                    //builder.AppendLine("source($DV_LIB+\"/library_win.dvrc\")");
                    builder.AppendLine("source(\"" + baseDirectory + "/library/library_win.dvrc\")");
                    break;
                case ExecutionPlatform.UnixLike:
                    //builder.AppendLine("source($DV_LIB+\"/library_linux.dvrc\")");
                    builder.AppendLine("source(\"" + baseDirectory + "/library/library_linux.dvrc\")");
                    break;
                default:
                    throw new Exception(TMSBEngineController.GetUnsupportedOSMessage(EngineControllerConfig.ExecutionPlatform.ToString()));
                    break;
            }
            return builder.ToString();
        }

        public string GetExtendedQuantumInformationString(ScriptQuantum scriptQuantum)
        {
            StringBuilder builder = new StringBuilder();
            builder.Append("Engine Number: ");
            builder.Append(EngineNumber.ToString());
            builder.Append(", ");
            builder.Append(TMSBDVScriptReader.GetQuantumDescriptionString(scriptQuantum));
            builder.Append(", Base Directory: '");
            builder.Append(BaseDirectory);
            builder.Append("'");
            return builder.ToString();
        }

        public static string GetInsufficientDiskSpaceErrorMessage(string baseDirectory, string programLocation, IOException sourceException)
        {
            return "The directory '" + baseDirectory +
                "' does not appear to have enough free space to continue executing the current script." +
                " The exact error description returned was '" + sourceException.Message.Trim() + 
                "'. This message was triggered at program location '" + programLocation + "'.";
        }
    }
}
