// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;

using BenLincoln.Utility;

namespace BenLincoln.DaVinci
{
    public enum ExecutionPlatform
    {
        Windows,
        UnixLike
    }

    public enum ExecutionPlatformMode
    {
        Automatic,
        ForceWindows,
        ForceUnixLike
    }

    public enum OutputLevel
    {
        None = 0,
        Minimal = 1,
        Standard = 2,
        Verbose = 3,
        Debug = 4
    }

    public enum ScriptGenerationMode
    {
        Never,
        Always,
        IfNotAlreadyPresent
    }

    public enum LogMode
    {
        Overwrite,
        Append
    }

    public enum MemoryRepresentation
    {
        PowerOfTwo,
        PowerOfTen
    }

    public class TMSBEngineControllerConfiguration
    {
        protected string _InputConfigFilePath;
        protected string _ColourConfigFilePath;
        protected string _ProcessingConfigFilePath;
        protected string _OutputConfigFilePath;
        protected string _TransformationProfileConfigFilePath;
        protected string[] _InputDirectories;
        protected int _MinimumSourceBandsInInputDirectory;
        protected bool _ScanInputDirectoriesRecursively;
        protected ScriptGenerationMode _GenerateScripts;
        protected bool _ExecuteScripts;
        protected bool _HaltOnErrors;
        protected bool _DeleteTemporaryFiles;
        protected bool _WriteScriptCodeWhenNoImagesWillBeGenerated;
        protected OutputLevel _OutputLevel;
        protected string _DaVinciCommand;
        protected string _DaVinciCommandArgs;
        protected ExecutionPlatformMode _ExecutionPlatformMode;
        //protected ExecutionPlatform _ExecutionPlatform;
        protected ProcessPriorityClass _DaVinciProcessPriority;
        protected int _EngineThreadCount;
        protected int _GlobalThreadCount;
        protected int _ScriptExecutionTimeLimit;
        protected bool _WriteOutputToConsole;
        protected bool _WriteOutputToLogFile;
        protected string _LogFilePath;
        //protected bool _OverwriteExistingLogFile;
        protected LogMode _LogFileMode;
        //protected bool _RepresentMemoryUsageInPowerOfTwoFormat;
        protected MemoryRepresentation _MemoryUnits;
        protected bool _CollectPerformanceData;
        protected bool _DisplayImageThumbnails;

        #region Properties

        public string InputConfigFilePath
        {
            get
            {
                return _InputConfigFilePath;
            }
            set
            {
                _InputConfigFilePath = value;
            }
        }

        public string ColourConfigFilePath
        {
            get
            {
                return _ColourConfigFilePath;
            }
            set
            {
                _ColourConfigFilePath = value;
            }
        }

        public string ProcessingConfigFilePath
        {
            get
            {
                return _ProcessingConfigFilePath;
            }
            set
            {
                _ProcessingConfigFilePath = value;
            }
        }

        public string OutputConfigFilePath
        {
            get
            {
                return _OutputConfigFilePath;
            }
            set
            {
                _OutputConfigFilePath = value;
            }
        }

        public string TransformationProfileConfigFilePath
        {
            get
            {
                return _TransformationProfileConfigFilePath;
            }
            set
            {
                _TransformationProfileConfigFilePath = value;
            }
        }

        public string[] InputDirectories
        {
            get
            {
                return _InputDirectories;
            }
            set
            {
                _InputDirectories = value;
            }
        }

        public int MinimumSourceBandsInInputDirectory
        {
            get
            {
                return _MinimumSourceBandsInInputDirectory;
            }
            set
            {
                _MinimumSourceBandsInInputDirectory = value;
            }
        }

        public bool ScanInputDirectoriesRecursively
        {
            get
            {
                return _ScanInputDirectoriesRecursively;
            }
            set
            {
                _ScanInputDirectoriesRecursively = value;
            }
        }

        public ScriptGenerationMode GenerateScripts
        {
            get
            {
                return _GenerateScripts;
            }
            set
            {
                _GenerateScripts = value;
            }
        }

        public bool ExecuteScripts
        {
            get
            {
                return _ExecuteScripts;
            }
            set
            {
                _ExecuteScripts = value;
            }
        }

        public bool HaltOnErrors
        {
            get
            {
                return _HaltOnErrors;
            }
            set
            {
                _HaltOnErrors = value;
            }
        }

        public bool DeleteTemporaryFiles
        {
            get
            {
                return _DeleteTemporaryFiles;
            }
            set
            {
                _DeleteTemporaryFiles = value;
            }
        }

        public bool WriteScriptCodeWhenNoImagesWillBeGenerated
        {
            get
            {
                return _WriteScriptCodeWhenNoImagesWillBeGenerated;
            }
            set
            {
                _WriteScriptCodeWhenNoImagesWillBeGenerated = value;
            }
        }

        public OutputLevel OutputLevel
        {
            get
            {
                return _OutputLevel;
            }
            set
            {
                _OutputLevel = value;
            }
        }

        public string DaVinciCommand
        {
            get
            {
                return _DaVinciCommand;
            }
            set
            {
                _DaVinciCommand = value;
            }
        }

        public string DaVinciCommandArgs
        {
            get
            {
                return _DaVinciCommandArgs;
            }
            set
            {
                _DaVinciCommandArgs = value;
            }
        }

        public ExecutionPlatform ExecutionPlatform
        {
            //get
            //{
            //    return _ExecutionPlatform;
            //}
            //set
            //{
            //    _ExecutionPlatform = value;
            //}
            get
            {
                return GetCurrentExecutionPlatform();
            }
        }

        public ExecutionPlatformMode ExecutionPlatformMode
        {
            get
            {
                return _ExecutionPlatformMode;
            }
            set
            {
                _ExecutionPlatformMode = value;
            }
        }

        public ProcessPriorityClass DaVinciProcessPriority
        {
            get
            {
                return _DaVinciProcessPriority;
            }
            set
            {
                _DaVinciProcessPriority = value;
            }
        }

        public int EngineThreadCount
        {
            get
            {
                return _EngineThreadCount;
            }
            set
            {
                _EngineThreadCount = value;
                if (_GlobalThreadCount < _EngineThreadCount)
                {
                    _GlobalThreadCount = _EngineThreadCount;
                }
            }
        }

        public int GlobalThreadCount
        {
            get
            {
                return _GlobalThreadCount;
            }
            set
            {
                if (value >= _EngineThreadCount)
                {
                    _GlobalThreadCount = value;
                }
                else
                {
                    _GlobalThreadCount = _EngineThreadCount;
                }
            }
        }

        public int ScriptExecutionTimeLimit
        {
            get
            {
                return _ScriptExecutionTimeLimit;
            }
            set
            {
                _ScriptExecutionTimeLimit = value;
            }
        }

        public bool WriteOutputToConsole
        {
            get
            {
                return _WriteOutputToConsole;
            }
            set
            {
                _WriteOutputToConsole = value;
            }
        }

        public bool WriteOutputToLogFile
        {
            get
            {
                return _WriteOutputToLogFile;
            }
            set
            {
                _WriteOutputToLogFile = value;
            }
        }

        public string LogFilePath
        {
            get
            {
                return _LogFilePath;
            }
            set
            {
                _LogFilePath = value;
            }
        }

        //public bool OverwriteExistingLogFile
        //{
        //    get
        //    {
        //        return _OverwriteExistingLogFile;
        //    }
        //    set
        //    {
        //        _OverwriteExistingLogFile = value;
        //    }
        //}

        public LogMode LogFileMode
        {
            get
            {
                return _LogFileMode;
            }
            set
            {
                _LogFileMode = value;
            }
        }

        //public bool RepresentMemoryUsageInPowerOfTwoFormat
        //{
        //    get
        //    {
        //        return _RepresentMemoryUsageInPowerOfTwoFormat;
        //    }
        //    set
        //    {
        //        _RepresentMemoryUsageInPowerOfTwoFormat = value;
        //    }
        //}

        public MemoryRepresentation MemoryUnits
        {
            get
            {
                return _MemoryUnits;
            }
            set
            {
                _MemoryUnits = value;
            }
        }

        public bool CollectPerformanceData
        {
            get
            {
                return _CollectPerformanceData;
            }
            set
            {
                _CollectPerformanceData = value;
            }
        }

        public bool DisplayImageThumbnails
        {
            get
            {
                return _DisplayImageThumbnails;
            }
            set
            {
                _DisplayImageThumbnails = value;
            }
        }

        #endregion

        public TMSBEngineControllerConfiguration()
        {
            ExecutionPlatformMode = ExecutionPlatformMode.Automatic;
            //ExecutionPlatform = GetCurrentExecutionPlatform();
            _InputConfigFilePath = "";
            ColourConfigFilePath = GetColourConfigFilePath();
            _ProcessingConfigFilePath = "";
            _OutputConfigFilePath = "";
            TransformationProfileConfigFilePath = GetTransformationProfileConfigFilePath();
            //_OverwriteExistingLogFile = true;
            LogFileMode = LogMode.Overwrite;
            _InputDirectories = new string[0];
            _MinimumSourceBandsInInputDirectory = 3;
            _ScanInputDirectoriesRecursively = false;
            GenerateScripts = ScriptGenerationMode.Always;
            _ExecuteScripts = false;
            HaltOnErrors = true;
            DeleteTemporaryFiles = true;
            WriteScriptCodeWhenNoImagesWillBeGenerated = false;
            OutputLevel = OutputLevel.Standard;
            DaVinciCommand = "";
            //DaVinciCommandArgs = "-f";
            DaVinciCommandArgs = "";
            DaVinciProcessPriority = ProcessPriorityClass.Normal;
            EngineThreadCount = 1;
            GlobalThreadCount = EngineThreadCount;
            _ScriptExecutionTimeLimit = 172800000;  // 48 hours
            _WriteOutputToConsole = true;
            _WriteOutputToLogFile = false;
            LogFilePath = "";
            //RepresentMemoryUsageInPowerOfTwoFormat = true;
            MemoryUnits = MemoryRepresentation.PowerOfTwo;
            CollectPerformanceData = true;
            DisplayImageThumbnails = true;
        }

        public static string GetColourConfigFilePath()
        {
            return "Global-Colours.xml";
        }

        public static string GetTransformationProfileConfigFilePath()
        {
            return "Global-TransformationProfiles.xml";
        }

        public static ExecutionPlatform AutoDetectCurrentExecutionPlatform()
        {
            // this should probably still be a little more robust
            if (Path.DirectorySeparatorChar == '\\')
            {
                return ExecutionPlatform.Windows;
            }
            if (Path.DirectorySeparatorChar == '/')
            {
                return ExecutionPlatform.UnixLike;
            }
            throw new Exception("TMSB was unable to automatically determine the executing OS. " +
                "Please use one of the manual/force OS options to select a supported OS.");
        }

        public ExecutionPlatform GetCurrentExecutionPlatform()
        {
            switch (this.ExecutionPlatformMode)
            {
                case ExecutionPlatformMode.ForceWindows:
                    return ExecutionPlatform.Windows;
                    break;
                case ExecutionPlatformMode.ForceUnixLike:
                    return ExecutionPlatform.UnixLike;
                    break;
                case ExecutionPlatformMode.Automatic:
                    //// this needs to be a little more robust
                    //return ExecutionPlatform.Windows;
                    return AutoDetectCurrentExecutionPlatform();
                    break;
                default:
                    throw new NotImplementedException("The selected execution platform mode (" + this.ExecutionPlatformMode.ToString() +
                        ") is not currently supported.");
                    break;
            }
        }

        public static void AppendConfigurationInformationLine(ref StringBuilder builder, string settingName, string settingValue)
        {
            builder.Append("\t");
            builder.Append(settingName);
            builder.Append(": ");
            builder.Append(settingValue);
            builder.Append(Environment.NewLine);
        }

        public string GetConfigurationInformationString()
        {
            StringBuilder builder = new StringBuilder();

            builder.AppendLine("- Engine Controller Configuration -");

            AppendConfigurationInformationLine(ref builder, "InputConfigFilePath", InputConfigFilePath);
            AppendConfigurationInformationLine(ref builder, "ColourConfigFilePath", ColourConfigFilePath);
            AppendConfigurationInformationLine(ref builder, "ProcessingConfigFilePath", ProcessingConfigFilePath);
            AppendConfigurationInformationLine(ref builder, "OutputConfigFilePath", OutputConfigFilePath);
            AppendConfigurationInformationLine(ref builder, "TransformationProfileConfigFilePath", TransformationProfileConfigFilePath);
            builder.AppendLine("Input Directories:");
            foreach (string dir in InputDirectories)
            {
                builder.Append("\t\t");
                builder.Append(dir);
                builder.Append(Environment.NewLine);
            }
            AppendConfigurationInformationLine(ref builder, "MinimumSourceBandsInInputDirectory", BenLincoln.Utility.FormatConverter.IntToString(MinimumSourceBandsInInputDirectory));
            AppendConfigurationInformationLine(ref builder, "ScanInputDirectoriesRecursively", ScanInputDirectoriesRecursively.ToString());
            AppendConfigurationInformationLine(ref builder, "GenerateScripts", GenerateScripts.ToString());
            AppendConfigurationInformationLine(ref builder, "ExecuteScripts", ExecuteScripts.ToString());
            AppendConfigurationInformationLine(ref builder, "HaltOnErrors", HaltOnErrors.ToString());
            AppendConfigurationInformationLine(ref builder, "DeleteTemporaryFiles", DeleteTemporaryFiles.ToString());
            AppendConfigurationInformationLine(ref builder, "WriteScriptCodeWhenNoImagesWillBeGenerated", WriteScriptCodeWhenNoImagesWillBeGenerated.ToString());
            AppendConfigurationInformationLine(ref builder, "OutputLevel", OutputLevel.ToString());
            AppendConfigurationInformationLine(ref builder, "DaVinciCommand", DaVinciCommand);
            AppendConfigurationInformationLine(ref builder, "DaVinciCommandArgs", DaVinciCommandArgs);
            AppendConfigurationInformationLine(ref builder, "ExecutionPlatform", ExecutionPlatform.ToString());
            AppendConfigurationInformationLine(ref builder, "DaVinciProcessPriority", DaVinciProcessPriority.ToString());
            AppendConfigurationInformationLine(ref builder, "EngineThreadCount", BenLincoln.Utility.FormatConverter.IntToString(EngineThreadCount));
            AppendConfigurationInformationLine(ref builder, "GlobalThreadCount", BenLincoln.Utility.FormatConverter.IntToString(GlobalThreadCount));
            AppendConfigurationInformationLine(ref builder, "ScriptExecutionTimeLimit", BenLincoln.Utility.FormatConverter.IntToString(ScriptExecutionTimeLimit));
            AppendConfigurationInformationLine(ref builder, "WriteOutputToConsole", WriteOutputToConsole.ToString());
            AppendConfigurationInformationLine(ref builder, "WriteOutputToLogFile", WriteOutputToLogFile.ToString());
            AppendConfigurationInformationLine(ref builder, "LogFilePath", LogFilePath);
            //AppendConfigurationInformationLine(ref builder, "OverwriteExistingLogFile", OverwriteExistingLogFile.ToString());
            //AppendConfigurationInformationLine(ref builder, "RepresentMemoryUsageInPowerOfTwoFormat", RepresentMemoryUsageInPowerOfTwoFormat.ToString());
            AppendConfigurationInformationLine(ref builder, "LogFileMode", LogFileMode.ToString());
            AppendConfigurationInformationLine(ref builder, "MemoryUnits", MemoryUnits.ToString());
            AppendConfigurationInformationLine(ref builder, "CollectPerformanceData", CollectPerformanceData.ToString());

            return builder.ToString();
        }

        public void CheckRequiredValues()
        {
            if (InputConfigFilePath == "")
            {
                throw new ArgumentException("A valid path to an input configuration file was not specified.");
            }

            if (ProcessingConfigFilePath == "")
            {
                throw new ArgumentException("A valid path to a processing configuration file was not specified.");
            }

            if (OutputConfigFilePath == "")
            {
                throw new ArgumentException("A valid path to an output configuration file was not specified.");
            }

            if (InputDirectories.Length < 1)
            {
                throw new ArgumentException("No input directories were specified.");
            }

            if (MinimumSourceBandsInInputDirectory < 1)
            {
                throw new ArgumentOutOfRangeException("The value specified for the minimum number of input files (" + _MinimumSourceBandsInInputDirectory.ToString() +
                    ") is invalid. It must be at least 1, and values less than 3 are unlikely to produce useful output.");
            }

            if (ExecuteScripts)
            {
                if (DaVinciCommand == "")
                {
                    throw new ArgumentException(@"The option to execute generated DaVinci scripts was specified, but no DaVinci command (for example, 'C:\Program Files\Davinci\Davinci.bat' on Windows, or '/usr/local/bin/davinci -f' on Unix/Linux) was specified.");
                }
                if (!File.Exists(DaVinciCommand))
                {
                    throw new ArgumentException("The path to the DaVinci command processor does not appear to represent a file that exists on this system. " +
                        "Please verify that '" + DaVinciCommand + "' is the correct location of the executable, and that it is the correct executable for " +
                        "the operating system this system is running.");
                }
            }
        }

        public static ProcessPriorityClass ProcessPriorityClassFromString(string priorityName)
        {
            string pName = priorityName.Trim().ToLower();

            switch (pName)
            {
                case "idle":
                    return ProcessPriorityClass.Idle;
                    break;
                case "low":     // undocumented feature!
                    return ProcessPriorityClass.Idle;
                    break;
                case "belownormal":
                    return ProcessPriorityClass.BelowNormal;
                    break;
                case "below normal":
                    return ProcessPriorityClass.BelowNormal;
                    break;
                case "normal":
                    return ProcessPriorityClass.Normal;
                    break;
                case "abovenormal":
                    return ProcessPriorityClass.AboveNormal;
                    break;
                case "above normal":
                    return ProcessPriorityClass.AboveNormal;
                    break;
                case "high":
                    return ProcessPriorityClass.High;
                    break;
                case "realtime":
                    return ProcessPriorityClass.RealTime;
                    break;
            }

            throw new ArgumentException("Unable to convert the string '" + priorityName + "' into a recognized ProcessPriorityClass.");
        }

        public static string ProcessPriorityClassToString(ProcessPriorityClass priority)
        {
            switch (priority)
            {
                case ProcessPriorityClass.Idle:
                    return "Idle";
                    break;
                case ProcessPriorityClass.BelowNormal:
                    return "BelowNormal";
                    break;
                case ProcessPriorityClass.Normal:
                    return "Normal";
                    break;
                case ProcessPriorityClass.AboveNormal:
                    return "AboveNormal";
                    break;
                case ProcessPriorityClass.High:
                    return "High";
                    break;
                case ProcessPriorityClass.RealTime:
                    return "RealTime";
                    break;
            }

            throw new ArgumentException("Unable to convert the string '" + priority + "' into a recognized ProcessPriorityClass.");
        }

        public static OutputLevel OutputLevelFromString(string outputLevel)
        {
            string name = outputLevel.Trim().ToLower();

            switch (name)
            {
                case "none":
                    return OutputLevel.None;
                    break;
                case "minimal":
                    return OutputLevel.Minimal;
                    break;
                case "standard":
                    return OutputLevel.Standard;
                    break;
                case "verbose":
                    return OutputLevel.Verbose;
                    break;
                case "debug":
                    return OutputLevel.Debug;
                    break;
            }

            throw new ArgumentException("Unable to convert the string '" + outputLevel + "' into a recognized OutputLevel.");
        }

        public static string OutputLevelToString(OutputLevel outputLevel)
        {
            switch (outputLevel)
            {
                case OutputLevel.None:
                    return "None";
                    break;
                case OutputLevel.Minimal:
                    return "Minimal";
                    break;
                case OutputLevel.Standard:
                    return "Standard";
                    break;
                case OutputLevel.Verbose:
                    return "Verbose";
                    break;
                case OutputLevel.Debug:
                    return "Debug";
                    break;
            }

            throw new ArgumentException("Unable to convert the OutputLevel '" + outputLevel.ToString() + "' into a string.");
        }

        public static OutputLevel OutputLevelFromInt(int outputLevel)
        {
            switch (outputLevel)
            {
                case 0:
                    return OutputLevel.None;
                    break;
                case 1:
                    return OutputLevel.Minimal;
                    break;
                case 2:
                    return OutputLevel.Standard;
                    break;
                case 3:
                    return OutputLevel.Verbose;
                    break;
                case 4:
                    return OutputLevel.Debug;
                    break;
            }
            throw new ArgumentOutOfRangeException("The integer " + outputLevel.ToString() + " could not be converted to a recognized OutputLevel.");
        }


        public static ScriptGenerationMode ScriptGenerationModeFromString(string generationMode)
        {
            string name = generationMode.Trim().ToLower();

            switch (name)
            {
                case "never":
                    return ScriptGenerationMode.Never;
                    break;
                case "always":
                    return ScriptGenerationMode.Always;
                    break;
                case "ifnotalreadypresent":
                    return ScriptGenerationMode.IfNotAlreadyPresent;
                    break;
                case "if not already present":
                    return ScriptGenerationMode.IfNotAlreadyPresent;
                    break;
            }

            throw new ArgumentException("Unable to convert the string '" + generationMode + "' into a recognized ScriptGenerationMode.");
        }

        public static string ScriptGenerationModeToString(ScriptGenerationMode generationMode)
        {
            return ScriptGenerationModeToString(generationMode, false);
        }

        public static string ScriptGenerationModeToString(ScriptGenerationMode generationMode, bool returnUIFriendlyString)
        {
            switch (generationMode)
            {
                case ScriptGenerationMode.Never:
                    return "Never";
                    break;
                case ScriptGenerationMode.Always:
                    return "Always";
                    break;
                case ScriptGenerationMode.IfNotAlreadyPresent:
                    if (returnUIFriendlyString)
                    {
                        return "If Not Already Present";
                    }
                    else
                    {
                        return "IfNotAlreadyPresent";
                    }
                    break;
            }

            throw new ArgumentException("Unable to convert the ScriptGenerationMode '" + generationMode.ToString() + "' into a string.");
        }


        public static LogMode LogModeFromString(string logMode)
        {
            string name = logMode.Trim().ToLower();

            switch (name)
            {
                case "overwrite":
                    return LogMode.Overwrite;
                    break;
                case "append":
                    return LogMode.Append;
                    break;
            }

            throw new ArgumentException("Unable to convert the string '" + logMode + "' into a recognized LogMode.");
        }


        public static string LogModeToString(LogMode logMode)
        {
            switch (logMode)
            {
                case LogMode.Overwrite:
                    return "Overwrite";
                    break;
                case LogMode.Append:
                    return "Append";
                    break;
            }

            throw new ArgumentException("Unable to convert the LogMode '" + logMode.ToString() + "' into a string.");
        }


        public static MemoryRepresentation MemoryRepresentationFromString(string memRep)
        {
            string name = memRep.Trim().ToLower();

            switch (name)
            {
                case "poweroftwo":
                    return MemoryRepresentation.PowerOfTwo;
                    break;
                case "power of two":
                    return MemoryRepresentation.PowerOfTwo;
                    break;
                case "poweroften":
                    return MemoryRepresentation.PowerOfTen;
                    break;
                case "power of ten":
                    return MemoryRepresentation.PowerOfTen;
                    break;
            }

            throw new ArgumentException("Unable to convert the string '" + memRep + "' into a recognized MemoryRepresentation.");
        }

        public static string MemoryRepresentationToString(MemoryRepresentation memRep)
        {
            return MemoryRepresentationToString(memRep, false);
        }

        public static string MemoryRepresentationToString(MemoryRepresentation memRep, bool returnUIFriendlyString)
        {
            switch (memRep)
            {
                case MemoryRepresentation.PowerOfTwo:
                    if (returnUIFriendlyString)
                    {
                        return "Power Of Two";
                    }
                    else
                    {
                        return "PowerOfTwo";
                    }
                    break;
                case MemoryRepresentation.PowerOfTen:
                    if (returnUIFriendlyString)
                    {
                        return "Power Of Ten";
                    }
                    else
                    {
                        return "PowerOfTen";
                    }
                    break;
            }

            throw new ArgumentException("Unable to convert the MemoryRepresentation '" + memRep.ToString() + "' into a string.");
        }

        public static ExecutionPlatformMode ExecutionPlatformModeFromString(string executionPlatformMode)
        {
            string name = executionPlatformMode.Trim().ToLower();

            switch (name)
            {
                case "automatic":
                    return ExecutionPlatformMode.Automatic;
                    break;
                case "forcewindows":
                    return ExecutionPlatformMode.ForceWindows;
                    break;
                case "forceunixlike":
                    return ExecutionPlatformMode.ForceUnixLike;
                    break;
            }

            throw new ArgumentException("Unable to convert the string '" + executionPlatformMode + "' into a recognized ExecutionPlatformMode.");
        }


        public static string ExecutionPlatformModeToString(ExecutionPlatformMode executionPlatformMode)
        {
            switch (executionPlatformMode)
            {
                case ExecutionPlatformMode.Automatic:
                    return "Automatic";
                    break;
                case ExecutionPlatformMode.ForceWindows:
                    return "ForceWindows";
                    break;
                case ExecutionPlatformMode.ForceUnixLike:
                    return "ForceUnixLike";
                    break;
            }

            throw new ArgumentException("Unable to convert the ExecutionPlatformMode '" + executionPlatformMode.ToString() + "' into a string.");
        }

        public static BenLincoln.DaVinci.ExecutionPlatform BLDVExecutionPlatformFromBLUExecutionPlatform(BenLincoln.Utility.ExecutionPlatform platform)
        {
            switch (platform)
            {
                case BenLincoln.Utility.ExecutionPlatform.Windows:
                    return BenLincoln.DaVinci.ExecutionPlatform.Windows;
                    break;
                case BenLincoln.Utility.ExecutionPlatform.UnixLike:
                    return BenLincoln.DaVinci.ExecutionPlatform.UnixLike;
                    break;
            }
            throw new Exception("Unsupported platform: " + platform.ToString());
        }

        public static BenLincoln.Utility.ExecutionPlatform BLUExecutionPlatformFromBLDVExecutionPlatform(BenLincoln.DaVinci.ExecutionPlatform platform)
        {
            switch (platform)
            {
                case BenLincoln.DaVinci.ExecutionPlatform.Windows:
                    return BenLincoln.Utility.ExecutionPlatform.Windows;
                    break;
                case BenLincoln.DaVinci.ExecutionPlatform.UnixLike:
                    return BenLincoln.Utility.ExecutionPlatform.UnixLike;
                    break;
            }
            throw new Exception("Unsupported platform: " + platform.ToString());
        }

    }
}
