// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;

using BenLincoln.Utility;

namespace BenLincoln.DaVinci
{
    public class TMSBEngineCore
    {
        protected TMSBEngine _ParentEngine;

        // so these threads can be killed on exit
        protected Thread _CoreThread;
        protected Process _ScriptProcess;
        protected int _ScriptProcessID;
        protected Thread _StandardOutputThread;
        protected Thread _StandardErrorThread;

        protected int _GlobalThreadNumber;
        protected bool _ScriptWasGeneratedByTMSB;
        protected int _ScriptQuantumNumber;
        protected ScriptQuantum _ScriptQuantum;

        protected bool _FinishedExecuting;
        protected bool _ContinueEngineExecution;

        protected NumericValueTracker _MemoryUseTracker;
        protected NumericValueTracker _ProcessorTimeTracker;
        protected PerformanceCounter[] _ProcessorTimeCounters;

        //public const string DaVinciProcessNameRegex = "^davinci[#0-9]{0,12}$";
        // need to base this on the user-selected script processor name
        public const string DaVinciProcessNameRegex = "dshadow";
        //public const string DaVinciProcessNameRegex = "";
        protected const int _MaximumScriptQuantumRetries = 3;
        protected const int _SleepLengthScriptQuantumRetry = 5000;

        protected int _IterationsWithZeroPercentProcessorTime;
        protected const int _MaxIterationsWithZeroPercentProcessorTime = 5;

        protected int _GlobalImageNumber;

        protected bool _AbortImmediately;

        #region Properties

        public TMSBEngine ParentEngine
        {
            get
            {
                return _ParentEngine;
            }
        }

        public int GlobalThreadNumber
        {
            get
            {
                return _GlobalThreadNumber;
            }
        }

        public int ScriptQuantumNumber
        {
            get
            {
                return _ScriptQuantumNumber;
            }
        }

        public ScriptQuantum ScriptQuantum
        {
            get
            {
                return _ScriptQuantum;
            }
        }

        public bool FinishedExecuting
        {
            get
            {
                return _FinishedExecuting;
            }
        }

        public bool ContinueEngineExecution
        {
            get
            {
                return _ContinueEngineExecution;
            }
        }

        public NumericValueTracker MemoryUseTracker
        {
            get
            {
                return _MemoryUseTracker;
            }
        }

        public NumericValueTracker ProcessorTimeTracker
        {
            get
            {
                return _ProcessorTimeTracker;
            }
        }

        #endregion

        public TMSBEngineCore()
        {
            _AbortImmediately = false;
            _GlobalThreadNumber = -1;
            _ScriptProcessID = 0;
            _ScriptWasGeneratedByTMSB = false;
            _ScriptQuantumNumber = -1;
            _FinishedExecuting = false;
            _ContinueEngineExecution = true;
            _MemoryUseTracker = new NumericValueTracker();
            _ProcessorTimeTracker = new NumericValueTracker();
            _ProcessorTimeCounters = new PerformanceCounter[0];
            _IterationsWithZeroPercentProcessorTime = 0;
            _GlobalImageNumber = 0;
        }

        public TMSBEngineCore(TMSBEngine parentEngine, int globalThreadNumber, bool scriptWasGeneratedByTMSB, int scriptQuantumNumber, ScriptQuantum scriptQuantum)
            : this()
        {
            _ParentEngine = parentEngine;
            _GlobalThreadNumber = globalThreadNumber;
            _ScriptWasGeneratedByTMSB = scriptWasGeneratedByTMSB;
            _ScriptQuantumNumber = scriptQuantumNumber;
            _ScriptQuantum = scriptQuantum;
        }

        ~TMSBEngineCore()
        {
            KillThreads();
            _ParentEngine = null;
            try
            {
                _ScriptProcess.Dispose();
            }
            catch (Exception ex)
            {
                // do nothing
            }
            for (int i = 0; i < _ProcessorTimeCounters.Length; i++)
            {
                if (_ProcessorTimeCounters[i] != null)
                {
                    try
                    {
                        _ProcessorTimeCounters[i].Dispose();
                    }
                    catch (Exception ex)
                    {
                        // do nothing
                    }
                }
            }
        }

        public void KillThreads()
        {
            KillProcessingThreads();
            if ((_CoreThread != null) && (_CoreThread.IsAlive))
            {
                _CoreThread.Abort();
            }
            //if ((_StandardOutputThread != null) && (_StandardOutputThread.IsAlive))
            //{
            //    _StandardOutputThread.Abort();
            //}
            //if ((_StandardErrorThread != null) && (_StandardErrorThread.IsAlive))
            //{
            //    _StandardErrorThread.Abort();
            //}
            _FinishedExecuting = true;
        }

        public void KillProcessingThreads()
        {
            if (_ScriptProcessID > 0)
            {
                // make sure not to kill processes belonging to a new process that got a recycled ID
                bool killChildProcesses = true;

                try
                {
                    if (_ScriptProcess.HasExited)
                    {
                        try
                        {
                            Process t = Process.GetProcessById(_ScriptProcessID);
                            Regex rex = new Regex(DaVinciProcessNameRegex);
                            if (!rex.IsMatch(t.ProcessName))
                            {
                                killChildProcesses = false;
                            }
                        }
                        catch (Exception ex)
                        {
                            // do nothing - there is no existing process with the same ID
                        }
                        //killChildProcesses = false;
                    }
                    else
                    {
                        _ScriptProcess.Kill();
                    }
                }
                catch (Exception ex)
                {
                    // do nothing
                }

                //if (!_ScriptProcess.HasExited)
                //{
                //    try
                //    {
                //        Process[] childProcesses = ProcessLocator.GetProcesses(_ScriptProcessID);
                //        foreach (Process p in childProcesses)
                //        {
                //            try
                //            {
                //                p.Kill();
                //            }
                //            catch (Exception ex)
                //            {
                //                _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber,
                //                    "Unable to kill an existing child process with PID " + p.Id.ToString() +
                //                    ". The error description was '" +
                //                    ex.Message + "'.");
                //            }
                //        }
                //    }
                //    catch (Exception ex2)
                //    {
                //        _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber,
                //            "Unable to kill existing child processes. The error description was '" +
                //            ex2.Message + "'.");
                //    }
                //    try
                //    {
                //        _ScriptProcess.Kill();
                //    }
                //    catch (Exception ex3)
                //    {
                //        _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber,
                //            "Unable to kill the script process. The error description was '" +
                //            ex3.Message + "'.");
                //    }
                //}
                if (killChildProcesses)
                {
                    try
                    {
                        Process[] childProcesses = ProcessLocator.GetProcesses(_ScriptProcessID, DaVinciProcessNameRegex, TMSBEngineControllerConfiguration.BLUExecutionPlatformFromBLDVExecutionPlatform(_ParentEngine.EngineControllerConfig.ExecutionPlatform));
                        foreach (Process p in childProcesses)
                        {
                            Regex rex = new Regex(DaVinciProcessNameRegex);
                            if (rex.IsMatch(p.ProcessName))
                            {
                                try
                                {
                                    p.Kill();
                                }
                                catch (Exception ex)
                                {
                                    try
                                    {
                                        _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber,
                                            "Unable to kill an existing child process with PID " + p.Id.ToString() +
                                            ". The error description was '" +
                                            ex.Message + "'.");
                                    }
                                    catch (Exception ex3)
                                    {
                                        //do nothing
                                    }
                                }
                            }
                            else
                            {
                                _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "A process with ID " + p.Id.ToString() + " and name '" + p.ProcessName + "' reached the second kill check but failed to pass it.");
                            }
                        }
                    }
                    catch (Exception ex2)
                    {
                        try
                        {
                            _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber,
                                "Unable to kill existing child processes. The error description was '" +
                                ex2.Message + "'.");
                        }
                        catch (Exception ex3)
                        {
                            //do nothing
                        }
                    }
                    try
                    {
                        _ScriptProcess.Kill();
                    }
                    catch (Exception ex3)
                    {
                        try
                        {
                            try
                            {
                                _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber,
                                    "Unable to kill the script process. The error description was '" +
                                    ex3.Message + "'.");
                            }
                            catch (Exception ex5)
                            {
                                //do nothing
                            }
                        }
                        catch (Exception ex4)
                        {
                            // do nothing
                        }
                    }
                }
            }
        }

        public void ExecuteQuantumThreaded()
        {
            _CoreThread = new Thread(new ThreadStart(ExecuteQuantum));
            _CoreThread.Start();
        }

        public void ExecuteQuantum()
        {
            bool executed = false;
            int retryCount = 0;
            string scriptPath = "";
            TimeSpan elapsed = TimeSpan.FromSeconds(0);
            while (!executed)
            {
                try
                {
                    scriptPath = _ParentEngine.GetTemporaryScriptFilePath(_ParentEngine.BaseDirectory);
                    string logPrefix = Path.GetFileNameWithoutExtension(scriptPath);
                    string stdOutLogPath = _ParentEngine.BaseDirectory + Path.DirectorySeparatorChar + logPrefix + "-StdOut.txt";
                    string stdErrLogPath = _ParentEngine.BaseDirectory + Path.DirectorySeparatorChar + logPrefix + "-StdErr.txt";

                    FileStream oStream = new FileStream(scriptPath, FileMode.Create, FileAccess.Write);
                    StreamWriter oWriter = new StreamWriter(oStream);

                    oWriter.Write(_ScriptQuantum.ScriptText);

                    oWriter.Close();
                    oStream.Close();

                    try
                    {
                        oWriter.Dispose();
                        oStream.Dispose();
                    }
                    catch (Exception ex)
                    {
                        // do nothing
                    }

                    lock (_ParentEngine.Tracker)
                    {
                        if (_ScriptQuantum.TotalImages > 0)
                        {
                            _ParentEngine.Tracker.CurrentElementNumber++;
                        }
                        _GlobalImageNumber = _ParentEngine.Tracker.CurrentElementNumber;
                    }

                    _ParentEngine.GenerateImageStartedEvent(_GlobalThreadNumber, _ScriptQuantum);

                    DateTime startTime = DateTime.Now;

                    string standardOutput = "";
                    string standardError = "";

                    CallScript(scriptPath, stdOutLogPath, stdErrLogPath, out standardOutput, out standardError);

                    bool errorOccurred = false;
                    string[] scriptErrorMessages = new string[0];
                    if (ScriptGeneratedErrors(standardOutput, standardError, out scriptErrorMessages))
                    {
                        errorOccurred = true;
                    }

                    if (_ScriptQuantum.OutputImages.Length > 0)
                    {
                        ArrayList missingOutput = new ArrayList();
                        for (int i = 0; i < _ScriptQuantum.OutputImages.Length; i++)
                        {
                            string checkPath = _ParentEngine.BaseDirectory + Path.DirectorySeparatorChar +
                                _ScriptQuantum.OutputImages[i];
                            if (!File.Exists(checkPath))
                            {
                                missingOutput.Add(checkPath);
                            }
                        }
                        if (missingOutput.Count > 0)
                        {
                            errorOccurred = true;
                            Array.Resize(ref scriptErrorMessages, scriptErrorMessages.Length + 1);
                            StringBuilder mBuilder = new StringBuilder();
                            mBuilder.Append(missingOutput.Count.ToString());
                            mBuilder.Append(" images should have been generated, but were not: ");
                            for (int i = 0; i < missingOutput.Count; i++)
                            {
                                mBuilder.Append("'");
                                mBuilder.Append((string)missingOutput[i]);
                                mBuilder.Append("'");
                                if (i < (missingOutput.Count - 1))
                                {
                                    mBuilder.Append(", ");
                                }
                            }
                            scriptErrorMessages[scriptErrorMessages.Length - 1] = mBuilder.ToString();
                        }
                    }


                    if (errorOccurred)
                    {
                        _ParentEngine.GenerateScriptExecutionErrorEvent(_GlobalThreadNumber, _ParentEngine.BaseDirectory,
                            scriptPath, scriptErrorMessages);
                        if (_ParentEngine.EngineControllerConfig.HaltOnErrors)
                        {
                            _ParentEngine.GenerateScriptAbortedEvent(_GlobalThreadNumber, "One or more errors were logged during execution of the script '" + scriptPath + "'.");
                            _ContinueEngineExecution = false;
                        }
                    }

                    if (_ParentEngine.EngineControllerConfig.DeleteTemporaryFiles)
                    {
                        _ParentEngine.DeleteFile(scriptPath);
                        _ParentEngine.DeleteFile(stdOutLogPath);
                        _ParentEngine.DeleteFile(stdErrLogPath);
                    }

                    DateTime endTime = DateTime.Now;

                    elapsed = endTime - startTime;

                    if ((_ScriptWasGeneratedByTMSB) && (_ScriptQuantum.OutputImages.Length > 0))
                    {
                        TimeSpan el = elapsed;
                        if (_ScriptQuantum.FirstImageNumber != _ScriptQuantum.LastImageNumber)
                        {
                            el = TimeSpan.FromSeconds(el.TotalSeconds / _ScriptQuantum.TotalImages);
                        }

                        if (el.TotalSeconds >= TMSBEngine.MinimumElapsedSecondsForTrackerAdd)
                        {
                            _ParentEngine.Tracker.AddElapsedTime(el);
                        }
                    }

                    executed = true;

                    if (_ScriptQuantum.TotalImages > 1)
                    {
                        int addImageCount = (_ScriptQuantum.TotalImages - 1);
                        lock (_ParentEngine.Tracker)
                        {
                            _ParentEngine.Tracker.CurrentElementNumber += addImageCount;
                        }
                        _GlobalImageNumber += addImageCount;
                    }
                }
                catch (Exception ex)
                {
                    bool isFatalError = false;
                    string fatalErrorMessage = "";
                    if (ex is IOException)
                    {
                        if (TMSBEngine.IOExceptionIsOutOfDiskSpaceError((IOException)ex))
                        {
                            // out of disk space, fail the entire directory
                            isFatalError = true;
                            fatalErrorMessage = TMSBEngine.GetInsufficientDiskSpaceErrorMessage(_ParentEngine.BaseDirectory,
                                "TMSBEngineCore: ExecuteQuantum()", (IOException)ex);
                        }
                    }

                    if (!isFatalError)
                    {
                        string message = "An error occurred while executing the script '" + scriptPath +
                            ". The error description was '" + ex.Message.Trim() + "'.";
                        if (retryCount > _MaximumScriptQuantumRetries)
                        {
                            message += " The sub-script was re-executed " + _MaximumScriptQuantumRetries.ToString() +
                                " times, but failed each time.";
                            //_ParentEngine.GenerateScriptAbortedEvent(_GlobalThreadNumber, message);
                            //_ContinueEngineExecution = false;
                            //executed = true;
                            //break;
                            isFatalError = true;
                            fatalErrorMessage = message;
                        }
                        else
                        {
                            int iterationNum = retryCount + 1;
                            message += " The sub-script will be re-executed up to " + _MaximumScriptQuantumRetries.ToString() +
                                " times. This is attempt number " + iterationNum.ToString();
                        }
                    }

                    if (isFatalError)
                    {
                        _ParentEngine.GenerateScriptAbortedEvent(_GlobalThreadNumber, fatalErrorMessage);
                        _ContinueEngineExecution = false;
                        executed = true;
                        _ParentEngine.Stop();
                        break;
                    }

                    retryCount++;
                }
                finally
                {
                    //_FinishedExecuting = true;
                    //_ParentEngine.GenerateImageFinishedEvent(_GlobalThreadNumber, _ScriptQuantum, elapsed, _MemoryUseTracker);
                }
                if (_AbortImmediately)
                {
                    _ParentEngine.GenerateScriptAbortedEvent(_GlobalThreadNumber, "A request was received to abort immediately.");
                    _ContinueEngineExecution = false;
                    executed = true;
                }
            }
            _FinishedExecuting = true;
            _ParentEngine.GenerateImageFinishedEvent(_GlobalThreadNumber, _ScriptQuantum, elapsed, _MemoryUseTracker, _ProcessorTimeTracker,
                _GlobalImageNumber);
        }

        protected bool ScriptGeneratedErrors(string standardOutput, string standardError, out string[] scriptErrorMessages)
        {
            // no output of any kind is an error
            // (this was too buggy to leave in place - the processes would often exit before the output could be captured)
            //if (standardOutput.Trim() == "")
            //{
            //    scriptErrorMessages = new string[] { "No output of any kind was found." };
            //    return true;
            //}

            // Threnody for Failed Script Execution
            string[] regexPatterns = new string[]
            {
                "error",
                "fail",

                "abort",
                "bad ",
                "illegal",
                "improper",
                "invalid",
                "null",
                "out of ",
                "overflow",
                "syntax",
                "unable",
                "underflow",
                "unhandled",
                "unknown",
                "unrecognized",
                "unsupported",
                "unusable",

                "could not",
                "couldn't",
                "cannot",
                "can't",
                //"c[adlno' ]{1,8}t",                                         // "could not", "couldn't", "cannot", "can't", etc.
                "do not",
                "don't",
                "does not",
                "doesn't",
                //"do[enos' ]{1,6}t",                                         // "do not ", "doesn't ", "don't ", etc.
                "is not",
                "isn't",
                //"i[nos' ]{1,5}t",                                           // "isn't", "is not", etc.

                //"already",
                "are different",
                "both contain",
                "duplicate",
                "expect",
                "incomplete",
                "must",
                "need to",
                //"only",
                "redundant",
                "required",
                "sizes differ",
                "too many",

                "no match",
                "non-unique",
                "not a[n]{0,1} ",
                "not enabled",
                "not enough",
                "not found",
                "not specified", 
                "not supported",
                "not that many",
                "not valid",                                                // "not a valid" is caught by the first regex in this section
                

                "i'm not",                                              // This section is for catching obscure errors (that really do exist 
                                                                        // in the DaVinci source)
                "the hell",


                "example",
                @"\$[0-9]{1,3}",                                        // the presence of references to an argument number (usually means that 
                                                                        //          the help message was displayed for a function)

                @"no [0-9A-Za-z\(\) _\-]{1,512} spec"                   // This originally had "specified" spelled out, but then I discovered 
                                                                        // that it is spelled as "specfiied" in at least two locations in the 
                                                                        // DaVinci source code. It's at the end because it's the most complex
                                                                        // expression used here.
                //@"null [0-9A-Za-z\(\) _\-]{1,512} pass"

            };
            ArrayList errorMessages = new ArrayList();

            string[] logContents = standardOutput.Split(new string[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries);

            GetMatchingLines(logContents, regexPatterns, ref errorMessages);

            scriptErrorMessages = (string[])errorMessages.ToArray("".GetType());
            if (scriptErrorMessages.Length > 0)
            {
                return true;
            }
            return false;
        }

        protected void GetMatchingLines(string[] text, string[] regexPatterns, ref ArrayList matchList)
        {
            foreach (string line in text)
            {
                foreach (string rP in regexPatterns)
                {
                    Regex rex = new Regex(rP, RegexOptions.IgnoreCase);
                    if (rex.IsMatch(line))
                    {
                        if (_ParentEngine.EngineControllerConfig.OutputLevel > OutputLevel.Verbose)
                        {
                            matchList.Add(line + " [ Matched by regular expression '" + rP + "'. ]");
                        }
                        else
                        {
                            matchList.Add(line);
                        }
                        break;
                    }
                }
            }
        }

        protected void CallScript(string scriptPath, string stdOutLogPath, string stdErrLogPath, out string standardOutput, out string standardError)
        {
            try
            {
                //switch (_ParentEngine.EngineControllerConfig.ExecutionPlatform)
                //{
                //    case ExecutionPlatform.Windows:
                //        CallScriptWindows(scriptPath, stdOutLogPath, stdErrLogPath, out standardOutput, out standardError);
                //        break;
                //    //case ExecutionPlatform.UnixLike:
                //    //    break;
                //    default:
                //        throw new Exception(TMSBEngineController.GetUnsupportedOSMessage(_ParentEngine.EngineControllerConfig.ExecutionPlatform.ToString()));
                //        break;
                //}
                CallScriptWindows(scriptPath, stdOutLogPath, stdErrLogPath, out standardOutput, out standardError);
            }
            catch (Exception ex)
            {
                throw new Exception("An error occurred while executing a sub-script: '" + ex.Message + "'.");
            }
        }

        protected void CallScriptWindows(string scriptPath, string stdOutLogPath, string stdErrLogPath, out string standardOutput, out string standardError)
        {
            StringBuilder cBuilder = new StringBuilder();
            if (_ParentEngine.EngineControllerConfig.DaVinciCommandArgs.Trim() != "")
            {
                cBuilder.Append(_ParentEngine.EngineControllerConfig.DaVinciCommandArgs);
                cBuilder.Append(" ");
            }
            cBuilder.Append(@"""");
            cBuilder.Append(Path.GetFileName(scriptPath));
            cBuilder.Append(@"""");

            System.Diagnostics.ProcessStartInfo sPI = new ProcessStartInfo((_ParentEngine.EngineControllerConfig.DaVinciCommand), cBuilder.ToString());

            sPI.UseShellExecute = false;
            //sPI.CreateNoWindow = true;
            sPI.CreateNoWindow = true;
            sPI.WindowStyle = ProcessWindowStyle.Hidden;
            //sPI.WindowStyle = ProcessWindowStyle.Normal;
            sPI.WorkingDirectory = Path.GetDirectoryName(scriptPath);
            sPI.LoadUserProfile = true;

            sPI.RedirectStandardOutput = true;
            sPI.RedirectStandardError = true;

            _ScriptProcess = new Process();
            _ScriptProcess.StartInfo = sPI;

            ProcessOutputHandler poh = new ProcessOutputHandler(_ScriptProcess);
            _StandardOutputThread = new Thread(new ThreadStart(poh.StartStandardOutputRead));
            _StandardErrorThread = new Thread(new ThreadStart(poh.StartStandardErrorRead));

            int waitLength = 0;
            _ScriptProcess.Start();
            _ScriptProcess.PriorityClass = _ParentEngine.EngineControllerConfig.DaVinciProcessPriority;
            _StandardOutputThread.Start();
            _StandardErrorThread.Start();
            _ScriptProcessID = _ScriptProcess.Id;

            while ((!_ScriptProcess.HasExited) || (_StandardOutputThread.IsAlive) || (_StandardErrorThread.IsAlive))
            {
                GetPerformanceData();
                _ScriptProcessID = _ScriptProcess.Id;

                Thread.Sleep(TMSBEngine.SleepLengthWaitingForScript);
                waitLength += TMSBEngine.SleepLengthWaitingForScript;
                if ((_ParentEngine.EngineControllerConfig.ScriptExecutionTimeLimit >= 0) && (waitLength >= _ParentEngine.EngineControllerConfig.ScriptExecutionTimeLimit))
                {
                    _ScriptProcess.Kill();
                    throw new Exception("Error: timed out while waiting for the script '" + scriptPath +
                        "' to finish executing after " + waitLength + " milliseconds. The script was forcibly terminated.");
                }
                if (_AbortImmediately)
                {
                    break;
                }
            }

            standardOutput = poh.StandardOutput;
            standardError = poh.StandardError;

            _ParentEngine.WriteTextToFile(stdOutLogPath, standardOutput, false);
            _ParentEngine.WriteTextToFile(stdErrLogPath, standardError, false);

            _ParentEngine.WriteRelevantLogTextToMasterLog(_GlobalThreadNumber, standardOutput, out standardOutput);
        }

        protected void GetPerformanceData()
        {
            if (!_ParentEngine.EngineControllerConfig.CollectPerformanceData)
            {
                return;
            }
            _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "[GetPerformanceData] _ScriptProcess.ProcessName =  '" +
                _ScriptProcess.ProcessName + "'.");
            string[] processNames = new string[1];
            int[] processIDs = new int[1];
            processNames[0] = _ScriptProcess.ProcessName;
            processIDs[0] = _ScriptProcess.Id;
            // memory use
            try
            {
                double currentVS = 0;

                if (_ScriptProcess.VirtualMemorySize64 > double.MaxValue)
                {
                    currentVS = double.MaxValue;
                }
                else
                {
                    currentVS = (double)_ScriptProcess.VirtualMemorySize64;
                }

                Process[] childProcesses = ProcessLocator.GetProcesses(_ScriptProcess.Id, DaVinciProcessNameRegex, TMSBEngineControllerConfiguration.BLUExecutionPlatformFromBLDVExecutionPlatform(_ParentEngine.EngineControllerConfig.ExecutionPlatform));
                processNames = new string[childProcesses.Length + 1];
                processNames[0] = _ScriptProcess.ProcessName;

                processIDs = new int[childProcesses.Length + 1];
                processIDs[0] = _ScriptProcess.Id;

                int i = 1;
                for (int j = 0; j < childProcesses.Length; j++)
                {
                    _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "[GetPerformanceData] attempting to obtain memory use data from child process " + j.ToString());
                    _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "\tVirtualMemorySize64 = " + childProcesses[j].VirtualMemorySize64.ToString());
                    _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "\tPeakVirtualMemorySize64 = " + childProcesses[j].PeakVirtualMemorySize64.ToString());
                    if (childProcesses[j].VirtualMemorySize64 > double.MaxValue)
                    {
                        currentVS = double.MaxValue;
                    }
                    else
                    {
                        _MemoryUseTracker.UpdateMaximum(childProcesses[j].PeakVirtualMemorySize64);
                        currentVS = Math.Min(double.MaxValue, currentVS + (double)childProcesses[j].VirtualMemorySize64);
                    }
                    processNames[i] = childProcesses[j].ProcessName;
                    processIDs[i] = childProcesses[j].Id;
                    childProcesses[j].Dispose();
                    i++;
                }

                _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "[GetPerformanceData] processNames[] is now: ");
                foreach (string pName in processNames)
                {
                    _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "\t" + pName);
                }

                _MemoryUseTracker.AddValue(currentVS);
            }
            catch (Exception ex)
            {
                _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "Unable to measure current memory usage. The error description was '" +
                    ex.Message + "'.");
            }

            // processor time
            try
            {
                bool createCountObjects = false;
                if (_ProcessorTimeCounters.Length == 0)
                {
                    createCountObjects = true;
                }
                if (_IterationsWithZeroPercentProcessorTime >= _MaxIterationsWithZeroPercentProcessorTime)
                {
                    createCountObjects = true;
                    _IterationsWithZeroPercentProcessorTime = 0;
                }

                if (createCountObjects)
                {
                    switch (_ParentEngine.EngineControllerConfig.ExecutionPlatform)
                    {
                        case ExecutionPlatform.Windows:
                            _ProcessorTimeCounters = PerformanceGatherer.GetPerformanceCounterForProcesses(processNames, "% Processor Time");
                            break;
                        case ExecutionPlatform.UnixLike:
                            _ProcessorTimeCounters = PerformanceGatherer.GetPerformanceCounterForProcesses(processIDs, "% Processor Time", BenLincoln.DaVinci.TMSBEngineControllerConfiguration.BLUExecutionPlatformFromBLDVExecutionPlatform(_ParentEngine.EngineControllerConfig.ExecutionPlatform));
                            break;
                        default:
                            throw new Exception("Unsupported platform: " + _ParentEngine.EngineControllerConfig.ExecutionPlatform.ToString());
                            break;
                    }
                    _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "Created the counter objects. " + 
                    _ProcessorTimeCounters.Length.ToString() + " were obtained.");
                }
                else
                {
                    //double currentPT = 0.0d;
                    //foreach (PerformanceCounter pc in _ProcessorTimeCounters)
                    //{
                    //    float currentVal = pc.NextValue();
                    //    currentPT += currentVal;
                    //    //_ParentEngine.GenerateScriptInformationalMessage(_GlobalThreadNumber, "Debug: individual counter value is " + String.Format("{0:0.0000}", currentVal));
                    //}
                    //_ParentEngine.GenerateScriptInformationalMessage(_GlobalThreadNumber, "Debug: sum is " + String.Format("{0:0.0000}", currentPT));
                    double currentPT = GetCurrentPercentProcessorTime();
                    _ProcessorTimeTracker.AddValue(currentPT);
                    if (currentPT == 0.0d)
                    {
                        _IterationsWithZeroPercentProcessorTime++;
                        _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "Percent Processor Time recorded as zero - iteration " + 
                            _IterationsWithZeroPercentProcessorTime.ToString() + "/" + _MaxIterationsWithZeroPercentProcessorTime.ToString() +
                            ". The current number of performance counter objects being summed for this value is " + _ProcessorTimeCounters.Length.ToString() + 
                            ".");
                    }
                    else
                    {
                        _IterationsWithZeroPercentProcessorTime = 0;
                    }
                }
            }
            catch (Exception ex)
            {
                _ParentEngine.GenerateScriptInformationalMessage(_GlobalThreadNumber, "Unable to measure current processor time usage. The error description was '" +
                    ex.Message + "'.");
            }
        }

        public double GetCurrentPercentProcessorTime()
        {
            double currentPT = 0.0d;
            _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "Summing " + _ProcessorTimeCounters.Length.ToString() +
                " processor time values.");
            foreach (PerformanceCounter pc in _ProcessorTimeCounters)
            {
                double previousPT = currentPT;
                try
                {
                    float currentVal = pc.NextValue();
                    _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "\tCurrent component value: " + string.Format("{0:0.00000000}", currentVal));
                    currentPT += currentVal;
                }
                catch (Exception ex)
                {
                    _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "\tException thrown: " + ex.Message);
                    _ParentEngine.GenerateScriptDebugMessage(_GlobalThreadNumber, "\tSetting currentPT back to '" + previousPT + "'.");
                    currentPT = previousPT;
                }
            }
            return currentPT;
        }

        public string GetProcessingInformationString()
        {
            return _ParentEngine.GetExtendedQuantumInformationString(_ScriptQuantum);
        }

        public void AbortImmediately()
        {
            _AbortImmediately = true;
        }
    }
}
