// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;

namespace BenLincoln.DaVinci
{
    public class TMSBInputConfigurationFile : TMSBConfigurationFile
    {
        protected TMSBInputConfiguration _InputConfig;
        protected TransformationProfileCollection _TransformationProfiles;

        protected enum ConfigProcessingState
        {
            Normal,
            SpectralBandSection,
            SpectralBandMain,
            SpectralBandSourceList,
            SpectralBandSource
        }

        #region Properties

        public TMSBInputConfiguration InputConfig
        {
            get
            {
                return _InputConfig;
            }
            set
            {
                _InputConfig = value;
            }
        }

        #endregion

        public TMSBInputConfigurationFile(string configFileName)
            : this(configFileName, null)
        {
        }

        public TMSBInputConfigurationFile(string configFileName, TransformationProfileCollection transformationProfiles)
            : base()
        {
            _InputConfig = new TMSBInputConfiguration();
            _TransformationProfiles = transformationProfiles;
            ReadData(configFileName);
        }

        protected override void ReadTypeSpecificData(Stream sourceStream)
        {
            XmlTextReader xReader = new XmlTextReader(sourceStream);
            //xReader.Settings.Schemas.Add(_FileSchema);

            string currentElementName = "";
            ConfigProcessingState currentState = ConfigProcessingState.Normal;
            SpectralBand newBand = new SpectralBand();
            SpectralBandData newBandData = new SpectralBandData();
            TransformationProfileCollection newBandInputTransformations = new TransformationProfileCollection();
            int dataSourceNum = 0;

            while (xReader.Read())
            {
                switch (xReader.NodeType)
                {
                    case XmlNodeType.Element:
                        currentElementName = xReader.Name;
                        switch (currentState)
                        {
                            case ConfigProcessingState.Normal:
                                switch (currentElementName)
                                {
                                    case "SpectralBands":
                                        currentState = ConfigProcessingState.SpectralBandSection;
                                        break;
                                    //case "TrueColourMapping":
                                    //    currentState = ConfigProcessingState.TrueColourMapping;
                                    //    break;
                                }
                                break;

                            // beginning of a spectral band definition
                            case ConfigProcessingState.SpectralBandSection:
                                if (currentElementName == "SpectralBand")
                                {
                                    currentState = ConfigProcessingState.SpectralBandMain;
                                    newBand = new SpectralBand();
                                    newBandInputTransformations = new TransformationProfileCollection();
                                    dataSourceNum = 0;
                                }
                                break;

                            // beginning of the list of data sources for a spectral band
                            case ConfigProcessingState.SpectralBandMain:
                                if (currentElementName == "DataSources")
                                {
                                    currentState = ConfigProcessingState.SpectralBandSourceList;
                                }
                                break;

                            // a data source definition for a spectral band
                            case ConfigProcessingState.SpectralBandSourceList:
                                if (currentElementName == "DataSource")
                                {
                                    currentState = ConfigProcessingState.SpectralBandSource;
                                    newBandData = new SpectralBandData();
                                    dataSourceNum++;
                                }
                                break;
                        }
                        break;
                    case XmlNodeType.EndElement:
                        currentElementName = xReader.Name;
                        switch (currentState)
                        {
                            // cases that would result in jumping back to the "normal" state
                            case ConfigProcessingState.SpectralBandSection:
                                if (currentElementName == "SpectralBands")
                                {
                                    currentState = ConfigProcessingState.Normal;
                                }
                                break;
                            //case ConfigProcessingState.TrueColourMapping:
                            //    if (currentElementName == "TrueColourMapping")
                            //    {
                            //        currentState = ConfigProcessingState.Normal;
                            //    }
                            //    break;

                            // end of a spectral band definition
                            case ConfigProcessingState.SpectralBandMain:
                                if (currentElementName == "SpectralBand")
                                {
                                    currentState = ConfigProcessingState.SpectralBandSection;
                                    newBand.InputTransformations = newBandInputTransformations;
                                    InputConfig.AddSpectralBand(newBand);
                                }
                                break;

                            // end of the list of data sources for a spectral band
                            case ConfigProcessingState.SpectralBandSourceList:
                                if (currentElementName == "DataSources")
                                {
                                    currentState = ConfigProcessingState.SpectralBandMain;
                                }
                                break;

                            // end of a data source definition for a spectral band
                            case ConfigProcessingState.SpectralBandSource:
                                if (currentElementName == "DataSource")
                                {
                                    currentState = ConfigProcessingState.SpectralBandSourceList;
                                    switch (dataSourceNum)
                                    {
                                        case 1:
                                            newBand.SetPrimaryImageData(newBandData);
                                            break;
                                        case 2:
                                            newBand.SetSecondaryImageData(newBandData);
                                            break;
                                        default:
                                            // ignore data sources after the second
                                            break;
                                    }
                                }
                                break;
                        }
                        break;
                    case XmlNodeType.Text:
                        string currentText = xReader.Value.Trim();
                        switch (currentState)
                        {
                            // Spectral band-related data
                            case ConfigProcessingState.SpectralBandMain:
                                switch (currentElementName)
                                {
                                    case "BandName":
                                        newBand.BandName = currentText;
                                        break;
                                    case "BandCategory":
                                        newBand.AddToBandCategory(SpectralBand.BandCategoryFromString(currentText));
                                        break;
                                    case "SpecialBandName":
                                        newBand.SpecialBandNames = SpectralBand.SpecialBandNameFromString(currentText);
                                        break;
                                    case "UseAsLuminanceChannel":
                                        newBand.UseAsLumaChannel = StringToBool(currentText);
                                        break;
                                    case "UseAsThreeChannelChannel":
                                        newBand.UseAsThreeChannelChannel = StringToBool(currentText);
                                        break;
                                    //case "Normalization":
                                        //newBand.InputNormalization = OutputImage.NormalizationModeFromString(currentText);
                                    //    break;
                                    case "InputTransformation":
                                        string spName = currentText.Trim();
                                        TransformationProfileCollection.AddTransformationProfileIfItExists(
                                            ref _TransformationProfiles, ref newBandInputTransformations, spName,
                                            "The spectral band '" + newBand.BandName + "'");
                                        break;
                                    case "MutuallyExclusiveBand":
                                        newBand.AddMutuallyExclusiveBand(currentText);
                                        break;
                                }
                                break;
                            case ConfigProcessingState.SpectralBandSource:
                                switch (currentElementName)
                                {
                                    case "FileName":
                                        newBandData.FileName = currentText;
                                        break;
                                    case "Plane":
                                        newBandData.PlaneNumber = BenLincoln.Utility.FormatConverter.StringToInt(currentText);
                                        break;
                                }
                                break;

                            //// true colour-related data
                            //case ConfigProcessingState.TrueColourMapping:
                            //    switch (currentElementName)
                            //    {
                            //        case "RedBand":
                            //            InputConfig.RedBandName = currentText;
                            //            break;
                            //        case "GreenBand":
                            //            InputConfig.GreenBandName = currentText;
                            //            break;
                            //        case "BlueBand":
                            //            InputConfig.BlueBandName = currentText;
                            //            break;
                            //    }
                            //    break;


                        }
                        break;
                }
            }
        }
    }
}
