// BenLincoln.DaVinci - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.DaVinci.

// BenLincoln.DaVinci is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DaVinci is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DaVinci (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.DaVinci
{
    public class TransformationProfileCollection : TransformationProfile
    {
        protected TransformationProfile[] _TransformationProfiles;
        protected bool _RequireUniqueNames;

        #region Properties

        public TransformationProfile[] TransformationProfiles
        {
            get
            {
                return _TransformationProfiles;
            }
            set
            {
                _TransformationProfiles = value;
            }
        }

        /// <summary>
        /// This property (which is set to false by default) controls whether or not 
        /// profile names must be unique for each profile in a given collection. This 
        /// is a good thing for e.g. a list of available settings, but not for e.g. 
        /// a list of profiles to be applied to a given image
        /// </summary>
        public bool RequireUniqueNames
        {
            get
            {
                return _RequireUniqueNames;
            }
            set
            {
                _RequireUniqueNames = value;
            }
        }

        #endregion

        public TransformationProfileCollection()
        {
            TransformationProfiles = new TransformationProfile[0];
            RequireUniqueNames = false;
        }

        public void AddTransformationProfile(TransformationProfile newProfile)
        {
            if (RequireUniqueNames)
            {
                if (ContainsTransformationProfile(newProfile.Name))
                {
                    throw new ArgumentException("The collection already contains a transformation profile named '" +
                        newProfile.Name + "', and duplicate names are not allowed under the current configuration.");
                }
            }
            if ((_TransformationProfiles == null) || (_TransformationProfiles.Length == 0))
            {
                TransformationProfiles = new TransformationProfile[1];
            }
            else
            {
                Array.Resize(ref _TransformationProfiles, (_TransformationProfiles.Length + 1));
            }
            TransformationProfiles[TransformationProfiles.Length - 1] = newProfile;
        }

        public void AddTransformationProfile(TransformationProfile[] newProfiles)
        {
            for (int i = 0; i < newProfiles.Length; i++)
            {
                if (newProfiles[i] != null)
                {
                    AddTransformationProfile(newProfiles[i]);
                }
            }
        }

        public bool ContainsTransformationProfile(string profileName)
        {
            return (ContainsTransformationProfile(profileName, true));
        }

        public bool ContainsTransformationProfile(string profileName, bool caseSensitive)
        {
            if (TransformationProfiles == null)
            {
                return false;
            }
            string pName = profileName;
            if (!caseSensitive)
            {
                pName = pName.ToLower();
            }
            for (int i = 0; i < TransformationProfiles.Length; i++)
            {
                string cName = TransformationProfiles[i].Name;
                if (!caseSensitive)
                {
                    cName = cName.ToLower();
                }
                if (pName == cName)
                {
                    return true;
                }
            }
            return false;
        }

        public string GetCaseInsensitiveTransformationProfileName(string profileName)
        {
            if (TransformationProfiles != null)
            {
                string pName = profileName.ToLower();
                for (int i = 0; i < TransformationProfiles.Length; i++)
                {
                    string cName = TransformationProfiles[i].Name.ToLower();
                    if (pName == cName)
                    {
                        return TransformationProfiles[i].Name;
                    }
                }
            }
            throw new ArgumentException("The collection does not contain a transformation profile named '" + profileName + "'.");
        }

        public TransformationProfile GetTransformationProfile(string profileName)
        {
            if (TransformationProfiles != null)
            {
                for (int i = 0; i < TransformationProfiles.Length; i++)
                {
                    if (TransformationProfiles[i].Name == profileName)
                    {
                        return TransformationProfiles[i];
                    }
                }
            }
            throw new ArgumentException("The collection does not contain a transformation profile named '" + profileName + "'.");
        }

        public int GetTransformationProfileNumber(string profileName)
        {
            if (TransformationProfiles != null)
            {
                for (int i = 0; i < TransformationProfiles.Length; i++)
                {
                    if (TransformationProfiles[i].Name == profileName)
                    {
                        return i;
                    }
                }
            }
            throw new ArgumentException("The collection does not contain a transformation profile named '" + profileName + "'.");
        }

        public TransformationProfile GetTransformationProfile(int profileNumber)
        {
            return TransformationProfiles[profileNumber];
        }

        public void RemoveTransformationProfile(int profileNumber)
        {
            int maxIndex = TransformationProfiles.Length - 1;
            if (profileNumber >= TransformationProfiles.Length)
            {
                throw new ArgumentOutOfRangeException("The maximum index of the collection is currently " +
                    maxIndex.ToString() + ", which is lower than the specified index (" +
                    profileNumber.ToString() + ").");
            }
            TransformationProfile[] newArray = new TransformationProfile[maxIndex];
            for (int i = 0; i < TransformationProfiles.Length; i++)
            {
                if (i != profileNumber)
                {
                    int targetIndex = i;
                    if (i > profileNumber)
                    {
                        targetIndex--;
                    }
                    newArray[targetIndex] = TransformationProfiles[i];
                }
            }
            TransformationProfiles = newArray;
        }

        public void RemoveTransformationProfile(string profileName)
        {
            RemoveTransformationProfile(GetTransformationProfileNumber(profileName));
        }

        public override string WrapScriptObjectInTransformationCode(string scriptObject)
        {
            string result = scriptObject;
            if (TransformationProfiles != null)
            {
                for (int i = 0; i < TransformationProfiles.Length; i++)
                {
                    result = TransformationProfiles[i].WrapScriptObjectInTransformationCode(result);
                }
            }
            return result;
        }

        /// <summary>
        /// Note: this method is intended for use when the collection is being used in the context of 
        /// e.g. a list of available options. Calling it when the collection is being used in the context
        /// of e.g. which profiles should be applied to a particular image will put the profiles
        /// in the wrong order.
        /// </summary>
        public void SortByName()
        {
            Array.Sort(TransformationProfiles);
        }


        public static void AddTransformationProfileIfItExists(ref TransformationProfileCollection sourceCollection,
            ref TransformationProfileCollection destinationCollection, string profileName, string referenceSourceName)
        {
            if (sourceCollection.ContainsTransformationProfile(profileName))
            {
                destinationCollection.AddTransformationProfile(sourceCollection.GetTransformationProfile(profileName));
            }
            else
            {
                string errorMessage = referenceSourceName + " contains a reference to a transformation profile named '" +
                    profileName + ", but this profile was not found in the the source collection.";
                if (sourceCollection.ContainsTransformationProfile(profileName, false))
                {
                    errorMessage += " the source collection does contain a profile named '" +
                        sourceCollection.GetCaseInsensitiveTransformationProfileName(profileName) +
                        "'. Profile names are case-sensitive. If this is the profile you intended " +
                        "to reference, please correct the capitalization of the profile names to match.";
                }
                errorMessage += " Please correct or remove the reference.";
                throw new ArgumentException(errorMessage);
            }
        }
    }
}
