// BenLincoln.Utility - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.Utility.

// BenLincoln.Utility is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.Utility is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.Utility (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace BenLincoln.Utility
{
    public class NumericValueTracker
    {
        //protected const int _DefaultMaxValueCount = 1024;
        //protected const int _DefaultMaxWindowSize = 512;
        //protected const int _DefaultMaxValueCount = 64;
        //protected const int _DefaultMaxWindowSize = 32;
        protected const int _DefaultMaxValueCount = 128;
        protected const int _DefaultMaxWindowSize = 64;

        protected double[] _Values;
        protected double _Maximum;
        protected double _Minimum;

        protected int _MaxValueCount;
        protected int _MaxValueCountWindowSize;

        // the numeric value to return if there is no data
        protected double _DefaultValue;

        // the string to return if a value cannot be calculated
        public const string DefaultString = "Waiting For Data...";

        protected static readonly object UpdateLockObject = new object();

        #region Properties

        public double DefaultValue
        {
            get
            {
                return _DefaultValue;
            }
        }

        public double[] Values
        {
            get
            {
                return _Values;
            }
        }

        public double Maximum
        {
            get
            {
                return _Maximum;
            }
        }

        public double Minimum
        {
            get
            {
                return _Minimum;
            }
        }

        /// <summary>
        /// The maximum number of discreet values to retain in the tracker.
        /// </summary>
        public int MaxValueCount
        {
            get
            {
                return _MaxValueCount;
            }
            set
            {
                _MaxValueCount = value;
            }
        }

        /// <summary>
        /// When the maximum number of values is exceeded, the oldest 
        /// n items will be removed, where n is this value.
        /// </summary>
        public int MaxValueCountWindowSize
        {
            get
            {
                return _MaxValueCountWindowSize;
            }
            set
            {
                _MaxValueCountWindowSize = value;
            }
        }

        public double CurrentValue
        {
            get
            {
                if (_Values.Length > 0)
                {
                    return _Values[_Values.Length - 1];
                }
                else
                {
                    return _DefaultValue;
                }
            }
        }

        #endregion

        public NumericValueTracker()
        {
            _Values = new double[0];
            _MaxValueCount = _DefaultMaxValueCount;
            _MaxValueCountWindowSize = _DefaultMaxWindowSize;
            _Minimum = 0;
            _Maximum = 0;
            _DefaultValue = -1.0d;
        }

        public void AddValue(Double newValue)
        {
            try
            {
                Thread addThread = new Thread(new ParameterizedThreadStart(AddValueThreadedWrapper));
                addThread.Start(newValue);
            }
            catch (Exception ex)
            {
                //Console.WriteLine("Error adding value '" + newValue.ToString() + "': " + ex.Message);
                // this is not a critical function, so fail silently
            }
        }

        protected void AddValueThreadedWrapper(object obj)
        {
            if (obj is double)
            {
                try
                {
                    AddValueThreaded((double)obj);
                }
                catch (Exception ex)
                {
                    //Console.WriteLine("Error adding value '" + obj.ToString() + "': " + ex.Message);
                    // this is not a critical function, so fail silently
                }
            }
        }

        protected void AddValueThreaded(Double newValue)
        {
            bool isFirstValue = false;
            lock (UpdateLockObject)
            {
                if ((_Values == null) || (_Values.Length == 0))
                {
                    _Values = new double[1];
                    isFirstValue = true;
                }
                else
                {
                    Array.Resize(ref _Values, _Values.Length + 1);
                }
                _Values[_Values.Length - 1] = newValue;

                // these are inside the "lock" block to prevent multiple threads
                // stomping on each others' triggering of this function
                if (isFirstValue)
                {
                    _Maximum = newValue;
                    _Minimum = newValue;
                }
                else
                {
                    if (newValue < _Minimum)
                    {
                        _Minimum = newValue;
                    }
                    if (newValue > _Maximum)
                    {
                        _Maximum = newValue;
                    }
                }
            }

            CheckArraySize();
        }

        public void UpdateMaximum(double newMax)
        {
            if (newMax > _Maximum)
            {
                _Maximum = newMax;
            }
        }

        public void UpdateMinimum(double newMin)
        {
            if (newMin < _Minimum)
            {
                _Minimum = newMin;
            }
        }

        protected void CheckArraySize()
        {
            lock (_Values)
            {
                if (_Values.Length > _MaxValueCount)
                {
                    int newSize = _MaxValueCount - _MaxValueCountWindowSize;
                    double[] trimmed = new double[newSize];
                    Array.Copy(_Values, _MaxValueCountWindowSize, trimmed, 0, newSize);
                    _Values = trimmed;
                }
            }
        }

        protected static float GetPercentChange(double val1, double val2)
        {
            if (val1 == val2)
            {
                return 0.0f;
            }
            double max = val1;
            double min = val2;
            if (max < min)
            {
                min = val1;
                max = val2;
            }

            float result = (float)Math.Abs(min / max);

            // these two things should never happen
            if (result < 0.0f)
            {
                result = 0.0f;
            }

            if (result > 1.0f)
            {
                result = 1.0f;
            }

            return (1.0f - result);
        }

        //public double GetMedianValue()
        //{
        //    double result = _DefaultValue;
        //    if (_Values.Length == 0)
        //    {
        //        return result;
        //    }
        //    if (_Values.Length == 1)
        //    {
        //        return _Values[0];
        //    }
        //    lock (_Values)
        //    {
        //        double[] sortedValues = new double[_Values.Length];
        //        Array.Copy(_Values, sortedValues, _Values.Length);
        //        Array.Sort(sortedValues);
        //        if ((sortedValues.Length % 2) == 1)
        //        {
        //            int midPoint = ((sortedValues.Length - 1) / 2);
        //            result = (long)sortedValues[midPoint];
        //        }
        //        else
        //        {
        //            int sP = (sortedValues.Length / 2) - 1;
        //            result = (long)((sortedValues[sP] + sortedValues[sP + 1]) / 2);
        //        }
        //    }
        //    return result;
        //}

        public double GetMedianValue()
        {
            //lock (_Values)
            //{
                return GetMedianValue(_Values);
            //}
        }

        public double GetMedianValue(double[] values)
        {
            double result = 0;
            if (values.Length == 0)
            {
                return _DefaultValue;
            }
            if (values.Length == 1)
            {
                return _Values[0];
            }
            double[] sortedValues = new double[values.Length];
            Array.Copy(values, sortedValues, values.Length);
            Array.Sort(sortedValues);
            if ((sortedValues.Length % 2) == 1)
            {
                int midPoint = ((sortedValues.Length - 1) / 2);
                result = (long)sortedValues[midPoint];
            }
            else
            {
                int sP = (sortedValues.Length / 2) - 1;
                result = (long)((sortedValues[sP] + sortedValues[sP + 1]) / 2);
            }
            return result;
        }

        public double GetAverageValue()
        {
            //lock (_Values)
            //{
                return GetAverageValue(_Values);
            //}
        }

        public double GetAverageValue(double[] values)
        {
            double result = 0;
            if (values.Length == 0)
            {
                return _DefaultValue;
            }
            if (values.Length == 1)
            {
                return values[0];
            }
            foreach (double d in values)
            {
                result += d;
            }
            result /= values.Length;

            return result;
        }

        public static string GetValueAsBytesInPowerOfTenStringFormat(double value)
        {
            double revisedVal = value;
            string suffix = "";
            if (revisedVal > 1000)
            {
                revisedVal /= 1000.0d;
                suffix = " KB";
            }
            if (revisedVal > 1000)
            {
                revisedVal /= 1000.0d;
                suffix = " MB";
            }
            if (revisedVal > 1000)
            {
                revisedVal /= 1000.0d;
                suffix = " GB";
            }
            if (revisedVal > 1000)
            {
                revisedVal /= 1000.0d;
                suffix = " TB";
            }
            if (revisedVal > 1000)
            {
                revisedVal /= 10004.0d;
                suffix = " PB";
            }
            if (revisedVal > 1000)
            {
                revisedVal /= 1000.0d;
                suffix = " EB";
            }
            if (revisedVal > 1000)
            {
                revisedVal /= 1000.0d;
                suffix = " ZB";
            }
            if (revisedVal > 1000)
            {
                revisedVal /= 1000.0d;
                suffix = " YB";
            }

            return GetDecimalValueAsString(revisedVal) + suffix;
        }

        public static string GetValueAsBytesInPowerOfTwoStringFormat(double value)
        {
            double revisedVal = value;
            string suffix = " Bytes";
            if (revisedVal > 1024)
            {
                revisedVal /= 1024.0d;
                suffix = " KiB";
            }
            if (revisedVal > 1024)
            {
                revisedVal /= 1024.0d;
                suffix = " MiB";
            }
            if (revisedVal > 1024)
            {
                revisedVal /= 1024.0d;
                suffix = " GiB";
            }
            if (revisedVal > 1024)
            {
                revisedVal /= 1024.0d;
                suffix = " TiB";
            }
            if (revisedVal > 1024)
            {
                revisedVal /= 1024.0d;
                suffix = " PiB";
            }
            if (revisedVal > 1024)
            {
                revisedVal /= 1024.0d;
                suffix = " EiB";
            }
            if (revisedVal > 1024)
            {
                revisedVal /= 1024.0d;
                suffix = " ZiB";
            }
            if (revisedVal > 1024)
            {
                revisedVal /= 1024.0d;
                suffix = " YiB";
            }

            return GetDecimalValueAsString(revisedVal) + suffix;
        }

        protected static string GetDecimalValueAsString(double d)
        {
            if (d < 10)
            {
                return String.Format("{0:0.#}", d);
            }
            else
            {
                return String.Format("{0:0}", d);
            }
        }
    }
}
