// BenLincoln.Utility - A shared library used by The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of BenLincoln.Utility.

// BenLincoln.Utility is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.Utility is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.Utility (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Text.RegularExpressions;

namespace BenLincoln.Utility
{
    public enum ExecutionPlatform
    {
        Windows,
        UnixLike
    }

    public class ProcessLocator
    {

        public static Process[] GetProcessTree(int parentProcessID)
        {
            ArrayList processList = new ArrayList();
            try
            {
                Process parent = Process.GetProcessById(parentProcessID);
                processList.Add(parent);
                ArrayList fallback = processList;
                try
                {
                    Process[] childProcesses = GetProcessTree(parentProcessID);
                    foreach (Process child in childProcesses)
                    {
                        ArrayList fallback2 = processList;
                        try
                        {
                            processList.Add(child);
                        }
                        catch (Exception ex)
                        {
                            processList = fallback2;
                        }
                    }
                }
                catch (Exception ex)
                {
                    processList = fallback;
                }
            }
            catch (Exception ex)
            {
                throw new Exception("Unable to build the process tree. The error description was '" + ex.Message + "'.", ex);
            }

            return (Process[])processList.ToArray(Process.GetCurrentProcess().GetType());
        }

        public static Process[] GetProcesses(int parentProcessID, ExecutionPlatform platform)
        {
            return GetProcesses(parentProcessID, "", platform);
        }

        public static Process[] GetProcesses(int parentProcessID, string instanceNameRegex, ExecutionPlatform platform)
        {
            return GetProcesses(parentProcessID, instanceNameRegex, true, platform);
        }

        public static Process[] GetProcesses(string instanceNameRegex, bool ignoreCase, ExecutionPlatform platform)
        {
            return GetProcesses(-1, instanceNameRegex, ignoreCase, platform);
        }

        public static Process[] GetProcesses(int parentProcessID, string instanceNameRegex, bool ignoreCase, ExecutionPlatform platform)
        {
            ArrayList idList = new ArrayList();
            PerformanceCounterCategory pcc = new PerformanceCounterCategory("Process");
            string[] instanceNames = pcc.GetInstanceNames();

            Regex rex = new Regex(instanceNameRegex, RegexOptions.IgnoreCase);
            if (!ignoreCase)
            {
                rex = new Regex(instanceNameRegex);
            }

            ArrayList parentIDs = new ArrayList();
            parentIDs.Add(parentProcessID);

            foreach (string iName in instanceNames)
            {
                //Console.WriteLine("Debug [ProcessLocator]: current iName is '" + iName + "'.");
                if ((instanceNameRegex == "") || (rex.IsMatch(iName)))
                {
                    //Console.WriteLine("Debug [ProcessLocator]: iName matches regex '" + instanceNameRegex + "'.");
                    int pProcessID = 0;
                    int processID = 0;

                    switch (platform)
                    {
                        case ExecutionPlatform.Windows:
                            foreach (PerformanceCounter counter in pcc.GetCounters(iName))
                            {
                                string counterName = counter.CounterName.ToLower();
                                if (counterName == "creating process id")
                                {
                                    pProcessID = (int)counter.RawValue;
                                }
                                if (counterName == "id process")
                                {
                                    processID = (int)counter.RawValue;
                                }
                                //Console.WriteLine("Debug [ProcessLocator]: Counter Name '" + counter.CounterName + "', Instance Name '" +
                                //    counter.InstanceName + "', Category Name '" + counter.CategoryName + "', Raw Value " +
                                //    counter.RawValue.ToString());
                            }
                            break;
                        case ExecutionPlatform.UnixLike:
                            // thanks for making this completely different, Mono Project!
                            //Console.WriteLine("Debug [ProcessLocator]: attempting to obtain a process ID from the string '" + iName + "'.");
                            string[] iNameComponents = iName.Split('/');
                            int tempParentProcessID = pProcessID;
                            int tempProcessID = processID;
                            try
                            {
                                int pid = int.Parse(iNameComponents[0]);
                                pProcessID = pid;
                                // can't seem to get the parent process ID, so we'll use -1 for now
                                // so that since the regex matched, this process will be included in the results
                                //pProcessID = -1;
                                processID = pid;
                            }
                            catch (Exception ex)
                            {
                                //Console.WriteLine("Debug [ProcessLocator]: exception thrown while trying to convert '" + iNameComponents[0].ToString() + "' to an integer: " + ex.Message);
                                pProcessID = tempParentProcessID;
                                processID = tempProcessID;
                            }
                            //Console.WriteLine("Debug [ProcessLocator]: result -  pProcessID = '" + pProcessID.ToString() + "', processID = '" + processID.ToString() + "'.");
                            foreach (PerformanceCounter counter in pcc.GetCounters(iName))
                            {
                                string counterName = counter.CounterName.ToLower();
                                //if (counterName == "creating process id")
                                //{
                                //    pProcessID = (int)counter.RawValue;
                                //}
                                //if (counterName == "id process")
                                //{
                                //    processID = (int)counter.RawValue;
                                //}
                                //Console.WriteLine("Debug [ProcessLocator]: Counter Name '" + counter.CounterName + "', Instance Name '" +
                                //    counter.InstanceName + "', Category Name '" + counter.CategoryName + "', Raw Value " +
                                //    counter.RawValue.ToString());
                            } 
                            break;
                    }
                    //if ((parentProcessID == -1) || (pProcessID == parentProcessID))
                    //Console.WriteLine("Debug [ProcessLocator]: parentProcessID is " + parentProcessID.ToString());
                    //Console.WriteLine("Debug [ProcessLocator]: parentIDs contains: ");
                    //foreach (int parentID in parentIDs)
                    //{
                    //    Console.WriteLine("\t" + parentID.ToString());
                    //}

                    if ((parentProcessID == -1) || (parentIDs.Contains(pProcessID)))
                    {
                        //Console.WriteLine("Debug [ProcessLocator]: parentProcessID is a match.");

                        if (!idList.Contains(processID))
                        {
                            //Console.WriteLine("Debug [ProcessLocator]: adding " + processID.ToString() + " to the list of process IDs.");
                            idList.Add(processID);
                        }
                        else
                        {
                            //Console.WriteLine("Debug [ProcessLocator]: the list of process IDs already contains " + processID.ToString() + ".");
                        }
                        if (!parentIDs.Contains(processID))
                        {
                            //Console.WriteLine("Debug [ProcessLocator]: adding " + processID.ToString() + " to the list of parent IDs.");
                            parentIDs.Add(processID);
                        }
                        else
                        {
                            //Console.WriteLine("Debug [ProcessLocator]: the list of parent IDs already contains " + processID.ToString() + ".");
                        }
                    }
                }
                else
                {
                    //Console.WriteLine("Debug [ProcessLocator]: iName does NOT match regex '" + instanceNameRegex + "'.");
                }
            }

            int[] IDs = (int[])idList.ToArray(1.GetType());

            if (IDs.Length == 0)
            {
                return new Process[0];
            }
            if (IDs.Length == 1)
            {
                return new Process[] { Process.GetProcessById(IDs[0]) };
            }         

            ArrayList pList = new ArrayList();

            foreach (int pid in IDs)
            {
                Process p = Process.GetProcessById(pid);
                pList.Add(p);
            }

            return (Process[])pList.ToArray(Process.GetCurrentProcess().GetType());
        }
    }
}
