// The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of The Mirror's Surface Breaks.

// The Mirror's Surface Breaks is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Mirror's Surface Breaks is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Mirror's Surface Breaks (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using BLDV = BenLincoln.DaVinci;
using BenLincoln.Utility;

namespace TMSB
{
    class Program
    {
        static void Main(string[] args)
        {
            BLDV.TMSBEngineControllerConfiguration ecConfig = GetEngineControllerConfigFromArguments(args);

            BLDV.TMSBEngineController controller = new BLDV.TMSBEngineController(ecConfig);
            BLDV.TMSBCommandLineEngineMonitor eM = new BLDV.TMSBCommandLineEngineMonitor(controller, ecConfig.OutputLevel);
            controller.ExecuteProcessing();
        }

        protected static BLDV.TMSBEngineControllerConfiguration GetEngineControllerConfigFromArguments(string[] args)
        {
            BLDV.TMSBEngineControllerConfiguration ecConfig = new BLDV.TMSBEngineControllerConfiguration();

            if (args.Length == 0)
            {
                DisplayHelpMessage();
                Environment.Exit(0);
            }

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                switch (args[i])
                {
                    case "-i":
                        try
                        {
                            ecConfig.InputConfigFilePath = args[i + 1];
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the input configuration file path", ex);
                        }
                        break;
                    case "-p":
                        try
                        {
                            ecConfig.ProcessingConfigFilePath = args[i + 1];
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the processing configuration file path", ex);
                        }
                        break;
                    case "-o":
                        try
                        {
                            ecConfig.OutputConfigFilePath = args[i + 1];
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the output configuration file path", ex);
                        }
                        break;
                    case "-m":
                        try
                        {
                            ecConfig.MinimumSourceBandsInInputDirectory = BenLincoln.Utility.FormatConverter.StringToInt(args[i + 1]);
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the minimum number of files required in an input directory", ex);
                        }
                        break;
                    case "-d":
                        try
                        {
                            string dvPath = args[i + 1];
                            if (!Path.IsPathRooted(dvPath))
                            {
                                dvPath = Path.GetFullPath(dvPath);
                            }
                            ecConfig.DaVinciCommand = dvPath;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the script processor path", ex);
                        }
                        break;
                    case "-da":
                        try
                        {
                            ecConfig.DaVinciCommandArgs = args[i + 1];
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the script processor arguments", ex);
                        }
                        break;
                    case "-pri":
                        try
                        {
                            ecConfig.DaVinciProcessPriority = BLDV.TMSBEngineControllerConfiguration.ProcessPriorityClassFromString(args[i + 1]);
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the script processor priority setting", ex);
                        }
                        break;
                    case "-td":
                        try
                        {
                            int directoryThreadCount = BenLincoln.Utility.FormatConverter.StringToInt(args[i + 1]);
                            if (directoryThreadCount < 1)
                            {
                                ExitWithError("The number of simultaneous directories to process cannot be less than 1.", 1);
                            }
                            ecConfig.EngineThreadCount = directoryThreadCount;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the simultaneous directory processing setting", ex);
                        }
                        break;
                    case "-tt":
                        try
                        {
                            int globalThreadCount = BenLincoln.Utility.FormatConverter.StringToInt(args[i + 1]);
                            if (globalThreadCount < 1)
                            {
                                ExitWithError("The number of simultaneous processing threads cannot be less than 1.", 1);
                            }
                            ecConfig.GlobalThreadCount = globalThreadCount;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the simultaneous processing thread count", ex);
                        }
                        break;
                    case "-ol":
                        try
                        {
                            int outputLevel = BenLincoln.Utility.FormatConverter.StringToInt(args[i + 1]);
                            if (outputLevel < 0)
                            {
                                ExitWithError("The output level cannot be less than 0.", 1);
                            }
                            if (outputLevel > 4)
                            {
                                ExitWithError("The output level cannot be greater than 4.", 1);
                            }
                            ecConfig.OutputLevel = BLDV.TMSBEngineControllerConfiguration.OutputLevelFromInt(outputLevel);
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the output level setting", ex);
                        }
                        break;
                    case "-tl":
                        try
                        {
                            int timeLimit = BenLincoln.Utility.FormatConverter.StringToInt(args[i + 1]);
                            ecConfig.ScriptExecutionTimeLimit = timeLimit;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the subscript execution time limit", ex);
                        }
                        break;
                    case "-log":
                        try
                        {
                            ecConfig.WriteOutputToLogFile = true;
                            string logFilePath = args[i + 1];
                            if (!Path.IsPathRooted(logFilePath))
                            {
                                //string fileName = Path.GetFileName(logFilePath);
                                //logFilePath = Directory.GetCurrentDirectory() + Path.DirectorySeparatorChar +
                                //    fileName;
                                logFilePath = Path.GetFullPath(logFilePath);
                            }
                            ecConfig.LogFilePath = logFilePath;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitDueToException("reading the output log file path", ex);
                        }
                        break;
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            ArrayList remainingArgs = new ArrayList(arguments);

            ArrayList inputDirectories = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    DisplayHelpMessage();
                    Environment.Exit(0);
                }
                switch (arg)
                {
                    case "-r":
                        remainingArgs.Remove(arg);
                        ecConfig.ScanInputDirectoriesRecursively = true;
                        break;
                    case "-x":
                        remainingArgs.Remove(arg);
                        ecConfig.ExecuteScripts = true;
                        break;
                    case "-ise":
                        remainingArgs.Remove(arg);
                        ecConfig.HaltOnErrors = false;
                        break;
                    case "-gn":
                        remainingArgs.Remove(arg);
                        ecConfig.GenerateScripts = BLDV.ScriptGenerationMode.Never;
                        ecConfig.ExecuteScripts = true;
                        break;
                    case "-gi":
                        remainingArgs.Remove(arg);
                        ecConfig.GenerateScripts = BLDV.ScriptGenerationMode.IfNotAlreadyPresent;
                        ecConfig.ExecuteScripts = true;
                        break;
                    case "-kt":
                        remainingArgs.Remove(arg);
                        ecConfig.DeleteTemporaryFiles = false;
                        break;
                    case "-aw":
                        remainingArgs.Remove(arg);
                        ecConfig.WriteScriptCodeWhenNoImagesWillBeGenerated = true;
                        break;
                    case "-al":
                        remainingArgs.Remove(arg);
                        ecConfig.LogFileMode = BLDV.LogMode.Append;
                        break;
                    case "-nco":
                        remainingArgs.Remove(arg);
                        ecConfig.WriteOutputToConsole = false;
                        break;
                    case "-dec":
                        remainingArgs.Remove(arg);
                        ecConfig.MemoryUnits = BLDV.MemoryRepresentation.PowerOfTen;
                        break;
                    case "-npd":
                        remainingArgs.Remove(arg);
                        ecConfig.CollectPerformanceData = false;
                        break;
                    case "-fow":
                        remainingArgs.Remove(arg);
                        ecConfig.ExecutionPlatformMode = BLDV.ExecutionPlatformMode.ForceWindows;
                        break;
                    case "-fou":
                        remainingArgs.Remove(arg);
                        ecConfig.ExecutionPlatformMode = BLDV.ExecutionPlatformMode.ForceUnixLike;
                        break;
                    default:
                        inputDirectories.Add(arg);
                        remainingArgs.Remove(arg);
                        break;
                }
            }

            ecConfig.InputDirectories = (string[])inputDirectories.ToArray("Proto".GetType());

            try
            {
                ecConfig.CheckRequiredValues();
            }
            catch (Exception ex)
            {
                ExitDueToException("processing the specified command-line arguments", ex);
            }

            if (ecConfig.GlobalThreadCount < ecConfig.EngineThreadCount)
            {
                Console.WriteLine("Warning: the maximum number of simultaneous directories to process was specified as " + 
                    ecConfig.EngineThreadCount.ToString() + ", but the total number of simultaneous processing threads " + 
                    "was either not specified, or was specified as less than the number of simultaneous directories. " +
                    "The number of simultaneous processing threads has been increased to match the number of " + 
                    "simultaneous directories.");
                ecConfig.GlobalThreadCount = ecConfig.EngineThreadCount;
            }

            return ecConfig;
        }

        protected static void ExitDueToException(string exceptionLocation, Exception ex)
        {
            ExitWithError("An error occurred while " + exceptionLocation + ". The error message was '" + ex.Message + "'.", 1);
        }

        protected static void ExitWithError(string message, int exitCode)
        {
            Console.WriteLine(message);
            System.Environment.ExitCode = exitCode;
            Environment.Exit(exitCode);
        }

        protected static void DisplayHelpMessage()
        {
            Console.WriteLine("The Mirror's Surface Breaks (Command-Line Interface)");
            Console.WriteLine("Version " + GetVersion());
            Console.WriteLine("Copyright 2010 Ben Lincoln");
            Console.WriteLine(Environment.NewLine);
            try
            {
                Stream inStream = BLDV.TMSBConfigurationFile.GetEmbeddedFileStream(Assembly.GetExecutingAssembly(), "HelpMessage.txt");
                StreamReader inReader = new StreamReader(inStream);
                Console.Write(inReader.ReadToEnd());
                inReader.Close();
                inStream.Close();
            }
            catch (Exception ex)
            {
                Console.WriteLine("An error occurred while trying to access the embedded help message. The error message was '" + 
                    ex.Message + "'.");
            }
            Console.WriteLine();
        }

        protected static string GetVersion()
        {
            Assembly asm = Assembly.GetExecutingAssembly();
            AssemblyName name = asm.GetName();
            string[] versionRaw = name.Version.ToString().Split('.');
            return versionRaw[0] + "." + versionRaw[1] + "." + versionRaw[2];
        }
    }
}
