// The Mirror's Surface Breaks
// Copyright 2010-2011 Ben Lincoln
// http://www.beneaththewaves.net/
//

// This file is part of The Mirror's Surface Breaks.

// The Mirror's Surface Breaks is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// The Mirror's Surface Breaks is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with The Mirror's Surface Breaks (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows.Forms;

using BenLincoln.DaVinci;
using BenLincoln.DaVinci.GUI;
using BenLincoln.Utility;

namespace TMSBGUI
{
    public partial class frmProcessingStatus : Form, IDisposable
    {
        protected enum ProcessingFormState
        {
            Running,
            Stopped
        }

        // the interval to wait (in milliseconds) between polling engines for perf data
        protected const int _PerformanceDataUpdateInterval = 5000;

        // the interval to wait (in milliseconds) after issuing a force stop, before checking for remaining DaVinci processes
        protected const int _SleepLengthWaitForForceStop = 5000;

        // the minimum interval (in milliseconds) that should elapse between recalculating total estimated remaining time
        protected const int _MinimumTimeEstimationInterval = 10000;

        // the number of pixels to subtract from the status display panel's width to 
        // determine the width of the status displays within it
        protected const int _StatusDisplayVersusPanelWidthFudgeFactorWithScrollBars = 25;
        protected const int _StatusDisplayVersusPanelWidthFudgeFactorWithoutScrollBars = 5;

        protected string[] _ExtensionPreferenceThumbnails = new string[] { "jpg", "tif", "bmp", "png" };
        protected string[] _ExtensionPreferenceDefaultFile = new string[] { "jpg", "tif", "bmp", "png" };

        protected ProcessingFormState _CurrentState;

        protected frmMain _ParentForm;

        protected TMSBGUIEngineMonitor _GUIEngineMonitor;
        protected TMSBCommandLineEngineMonitor _CLEngineMonitor;

        protected EngineStatusDisplay[] _StatusDisplays;

        protected int _TotalDirectories;
        protected int _ProcessedDirectories;
        protected string _EstimatedTimeRemainingString;
        protected float _EstimatedTimeRemainingConfidence;
        protected int _ProcessCount;
        protected string _CurrentMemoryUsageString;
        protected string _MaxMemoryUsageString;
        protected string _AverageMemoryUsageString;
        protected string _MedianMemoryUsageString;
        protected string _CurrentProcessorUsageString;
        protected string _MaxProcessorUsageString;
        protected string _AverageProcessorUsageString;
        protected string _MedianProcessorUsageString;

        protected System.Windows.Forms.Timer _UpdateTimer;
        protected DateTime _LastPerformanceDataUpdate;
        protected DateTime _LastTimeEstimationUpdate;

        protected bool _UpdatingTotalTimeRemaining;
        protected bool _TimeEstimationDataExists;

        public static readonly object StatusPanelLockObject = new object();
        public static readonly object PerformanceDataLockObject = new object();
        public static readonly object UpdatingTotalTimeRemainingLockObject = new object();

        #region Properties

        public EngineStatusDisplay[] StatusDisplays
        {
            get
            {
                return _StatusDisplays;
            }
            set
            {
                _StatusDisplays = value;
                UpdateStatusDisplayPanel();
            }
        }

        public ProcessableItemQueueGrid QueueGrid
        {
            get
            {
                return pqgQueue;
            }
            set
            {
                pqgQueue = value;
            }
        }

        public int TotalDirectories
        {
            get
            {
                return _TotalDirectories;
            }
            set
            {
                _TotalDirectories = value;
                UpdateDirectoriesProcessedLabel();
            }
        }

        public int ProcessedDirectories
        {
            get
            {
                return _ProcessedDirectories;
            }
            set
            {
                _ProcessedDirectories = value;
                UpdateDirectoriesProcessedLabel();
            }
        }

        public string EstimatedTimeRemainingString
        {
            get
            {
                return _EstimatedTimeRemainingString;
            }
            set
            {
                _EstimatedTimeRemainingString = value;
                UpdateTimeRemainingLabel();
            }
        }

        public float EstimatedTimeRemainingConfidence
        {
            get
            {
                return _EstimatedTimeRemainingConfidence;
            }
            set
            {
                _EstimatedTimeRemainingConfidence = value;
                UpdateTimeRemainingLabel();
            }
        }

        public int ProcessCount
        {
            get
            {
                return _ProcessCount;
            }
            set
            {
                _ProcessCount = value;
                lblProcessCount.Text = "Total Process Count: " + _ProcessCount.ToString();
            }
        }

        public string CurrentMemoryUsageString
        {
            get
            {
                return _CurrentMemoryUsageString;
            }
            set
            {
                _CurrentMemoryUsageString = value;
                UpdateMemoryStatsLabel();
            }
        }

        public string MaxMemoryUsageString
        {
            get
            {
                return _MaxMemoryUsageString;
            }
            set
            {
                _MaxMemoryUsageString = value;
                UpdateMemoryStatsLabel();
            }
        }

        public string AverageMemoryUsageString
        {
            get
            {
                return _AverageMemoryUsageString;
            }
            set
            {
                _AverageMemoryUsageString = value;
                UpdateMemoryStatsLabel();
            }
        }

        public string MedianMemoryUsageString
        {
            get
            {
                return _MedianMemoryUsageString;
            }
            set
            {
                _MedianMemoryUsageString = value;
                UpdateMemoryStatsLabel();
            }
        }

        public string CurrentProcessorUsageString
        {
            get
            {
                return _CurrentProcessorUsageString;
            }
            set
            {
                _CurrentProcessorUsageString = value;
                UpdateProcessorStatsLabel();
            }
        }

        public string MaxProcessorUsageString
        {
            get
            {
                return _MaxProcessorUsageString;
            }
            set
            {
                _MaxProcessorUsageString = value;
                UpdateProcessorStatsLabel();
            }
        }

        public string AverageProcessorUsageString
        {
            get
            {
                return _AverageProcessorUsageString;
            }
            set
            {
                _AverageProcessorUsageString = value;
                UpdateProcessorStatsLabel();
            }
        }

        public string MedianProcessorUsageString
        {
            get
            {
                return _MedianProcessorUsageString;
            }
            set
            {
                _MedianProcessorUsageString = value;
                UpdateProcessorStatsLabel();
            }
        }

        #endregion

        public frmProcessingStatus(frmMain parentForm, TMSBGUIEngineMonitor monitor, TMSBCommandLineEngineMonitor clMonitor)
        {
            InitializeComponent();

            _UpdatingTotalTimeRemaining = false;
            _TimeEstimationDataExists = false;

            _ParentForm = parentForm;

            _CurrentState = ProcessingFormState.Running;

            this.FormClosing += new FormClosingEventHandler(frmProcessingStatus_FormClosing);

            ResetState();

            _GUIEngineMonitor = monitor;
            _CLEngineMonitor = clMonitor;
            BuildQueue();

            _GUIEngineMonitor.ScriptBeingGenerated += OnScriptBeingGenerated;
            _GUIEngineMonitor.ScriptGenerated += OnScriptGenerated;
            _GUIEngineMonitor.ImageProcessingStarted += OnImageStarted;
            _GUIEngineMonitor.ImageProcessingFinished += OnImageFinished;
            _GUIEngineMonitor.DirectoryProcessingStarted += OnDirectoryStarted;
            _GUIEngineMonitor.DirectoryProcessingFinished += OnDirectoryFinished;
            _GUIEngineMonitor.EngineExecutionComplete += OnEngineExecutionComplete;
            _GUIEngineMonitor.EngineControllerExecutionComplete += OnEngineControllerExecutionComplete;
            _GUIEngineMonitor.WaitingForProcessingThread += OnWaitingForProcessingThread;
            _GUIEngineMonitor.DirectoryProcessingError += OnDirectoryProcessingError;

            TotalDirectories = _GUIEngineMonitor.Controller.BaseDirectories.Length;

            ResetStatusDisplays(_GUIEngineMonitor.Controller.EngineControllerConfiguration.EngineThreadCount);

            _GUIEngineMonitor.Controller.ExecuteProcessingThreaded();

            _LastPerformanceDataUpdate = DateTime.Now;
            _LastTimeEstimationUpdate = DateTime.Now - TimeSpan.FromMilliseconds(_MinimumTimeEstimationInterval);

            _UpdateTimer = new System.Windows.Forms.Timer();
            _UpdateTimer.Interval = _PerformanceDataUpdateInterval;
            _UpdateTimer.Tick += _UpdateTimer_Tick;
            _UpdateTimer.Enabled = true;
        }

        //public void Dispose()
        //{
        //    Dispose(true);

        //    // Use SupressFinalize in case a subclass
        //    // of this type implements a finalizer.
        //    //GC.SuppressFinalize(this);
        //}

        ~frmProcessingStatus()
        {
            this.Dispose();
        }

        protected override void Dispose(bool disposing)
        {
            // If you need thread safety, use a lock around these 
            // operations, as well as in your methods that use the resource.
            //if (!_Disposed)
            //{
                //if (disposing)
                //{
                    _ParentForm = null;
                    _UpdateTimer.Tick -= _UpdateTimer_Tick;
                    try
                    {
                        if (_UpdateTimer != null)
                        {
                            _UpdateTimer.Dispose();
                        }
                    }
                    catch (Exception ex)
                    {
                        // do nothing
                    }
                    try
                    {
                        if (_GUIEngineMonitor != null)
                        {
                            _GUIEngineMonitor.Dispose();
                        }
                    }
                    catch (Exception ex)
                    {
                        // do nothing
                    }
                    try
                    {
                        if (_CLEngineMonitor != null)
                        {
                            _CLEngineMonitor.Dispose();
                        }
                    }
                    catch (Exception ex)
                    {
                        // do nothing
                    }

                    try
                    {
                        this.Events.Dispose();
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Error disposing of delegate list: " + ex.Message);
                        // do nothing
                    }
                //}

                //_Disposed = true;
            //}
            base.Dispose(disposing);
        }


        protected void _UpdateTimer_Tick(object sender, EventArgs e)
        {
            TimerTickFunction();
        }


        delegate void TimerTickCallback();

        protected void TimerTickFunction()
        {
            if (this.InvokeRequired)
            {
                TimerTickCallback c = new TimerTickCallback(TimerTickFunction);
                this.Invoke(c);
            }
            else
            {
                lock (PerformanceDataLockObject)
                {
                    TimeSpan elapsed = DateTime.Now - _LastPerformanceDataUpdate;

                    if (elapsed.TotalMilliseconds < _PerformanceDataUpdateInterval)
                    {
                        // do nothing - something caused a logjam of tick events
                    }
                    else
                    {
                        _GUIEngineMonitor.Controller.UpdateEngineCoreStatus();
                        int activeProcesses = 0;
                        for (int i = 0; i < _GUIEngineMonitor.Controller.Engines.Length; i++)
                        {
                            int activeProcessesForEngine = 0;
                            if (_GUIEngineMonitor.Controller.Engines[i].Engine != null)
                            {
                                _GUIEngineMonitor.Controller.Engines[i].Engine.UpdateEngineCoreStatus();
                                activeProcessesForEngine = _GUIEngineMonitor.Controller.Engines[i].Engine.ActiveEngineCoreCount;
                            }
                            StatusDisplays[i].ProcessCount = activeProcessesForEngine;
                            activeProcesses += activeProcessesForEngine;
                        }
                        ProcessCount = activeProcesses;

                        UpdateOverallPerformanceStats();

                        _LastPerformanceDataUpdate = DateTime.Now;
                    }
                }
            }
        }

        public int IndexOfStatusDisplay(string directoryName)
        {
            for (int i = 0; i < StatusDisplays.Length; i++)
            {
                if (StatusDisplays[i].DirectoryName == directoryName)
                {
                    return i;
                }
            }
            return -1;
        }

        protected EngineStatusDisplay GetNewEngineStatusDisplay()
        {
            EngineStatusDisplay newDisplay = new EngineStatusDisplay();
            if (_GUIEngineMonitor.Controller.EngineControllerConfiguration.DisplayImageThumbnails)
            {
                newDisplay.DisplayThumbnails = true;
            }
            else
            {
                newDisplay.DisplayThumbnails = false;
            }
            if (_GUIEngineMonitor.Controller.EngineControllerConfiguration.CollectPerformanceData)
            {
                newDisplay.DisplayPerformanceData = true;
            }
            else
            {
                newDisplay.DisplayPerformanceData = false;
            }
            if (pnlEngineStatus.VerticalScroll.Visible)
            {
                newDisplay.Width = (pnlEngineStatus.Width - _StatusDisplayVersusPanelWidthFudgeFactorWithScrollBars);
            }
            else
            {
                newDisplay.Width = (pnlEngineStatus.Width - _StatusDisplayVersusPanelWidthFudgeFactorWithoutScrollBars);
            }

            return newDisplay;
        }

        public void ResetStatusDisplays(int engineCount)
        {
            StatusDisplays = new EngineStatusDisplay[engineCount];

            int currentY = 0;

            for (int i = 0; i < engineCount; i++)
            {
                EngineStatusDisplay newDisplay = GetNewEngineStatusDisplay();
                newDisplay.Top = currentY;

                _StatusDisplays[i] = newDisplay;
                currentY += newDisplay.Height;
            }

            UpdateStatusDisplayPanel();
        }

        protected void UpdateStatusDisplayPanel()
        {
            try
            {
                lock (StatusPanelLockObject)
                {
                    try
                    {
                        pnlEngineStatus.Controls.Clear();
                    }
                    catch (Exception ex)
                    {
                        throw new Exception("Couldn't clear the controls of pnlEngineStatus.", ex);
                    }
                   
                    int currentY = 0;

                    if (_StatusDisplays != null)
                    {
                        for (int i = 0; i < _StatusDisplays.Length; i++)
                        {
                            if (_StatusDisplays[i] == null)
                            {
                                try
                                {
                                    _StatusDisplays[i] = GetNewEngineStatusDisplay();
                                }
                                catch (Exception ex)
                                {
                                    throw new Exception("Couldn't create the engine status display with number " + i.ToString() + ".", ex);
                                }
                            }
                            _StatusDisplays[i].Top = currentY;
                            if (pnlEngineStatus.VerticalScroll.Visible)
                            {
                                try
                                {
                                    _StatusDisplays[i].Width = (pnlEngineStatus.Width - _StatusDisplayVersusPanelWidthFudgeFactorWithScrollBars);
                                }
                                catch (Exception ex)
                                {
                                    throw new Exception("Couldn't set the width of engine status display with number " + i.ToString() + " (VerticalScroll is Visible).", ex);
                                }
                            }
                            else
                            {
                                try
                                {
                                    _StatusDisplays[i].Width = (pnlEngineStatus.Width - _StatusDisplayVersusPanelWidthFudgeFactorWithoutScrollBars);
                                }
                                catch (Exception ex)
                                {
                                    throw new Exception("Couldn't set the width of engine status display with number " + i.ToString() + " (VerticalScroll is not Visible).", ex);
                                }
                            }
                            try
                            {
                                pnlEngineStatus.Controls.Add(_StatusDisplays[i]);
                            }
                            catch (Exception ex)
                            {
                                throw new Exception("Couldn't add the engine status display with number " + i.ToString() + " to pnlEngineStatus.", ex);
                            }

                            try
                            {
                                currentY += _StatusDisplays[i].Height;
                            }
                            catch (Exception ex)
                            {
                                throw new Exception("Couldn't add the height of engine status display with number " + i.ToString() + " to currentY.", ex);
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                throw new Exception("Couldn't update the status displays.", ex);
            }
        }

        public void ResetState()
        {
            _TotalDirectories = 0;
            _ProcessedDirectories = 0;
            UpdateDirectoriesProcessedLabel();

            _EstimatedTimeRemainingString = "Waiting For Data...";
            _EstimatedTimeRemainingConfidence = 1.0f;
            UpdateTimeRemainingLabel();

            _CurrentMemoryUsageString = "";
            _MaxMemoryUsageString = "";
            _AverageMemoryUsageString = "";
            _MedianMemoryUsageString = "";
            UpdateMemoryStatsLabel();

            _CurrentProcessorUsageString = "";
            _MaxProcessorUsageString = "";
            _AverageProcessorUsageString = "";
            _MedianProcessorUsageString = "";
            UpdateProcessorStatsLabel();

            ProcessCount = 0;
        }

        protected void UpdateDirectoriesProcessedLabel()
        {
            lblDirectoriesProcessed.Text = "Directories Processed: " + _ProcessedDirectories.ToString() + " / " + _TotalDirectories.ToString();
            //pbProgress.Maximum = _TotalImages;
            //pbProgress.Value = _ProcessedImages;
        }

        protected void UpdateTimeRemainingLabel()
        {
            // Time estimation is too buggy for now
            lblEstimatedTimeRemaining.Text = "";
            return;

            lblEstimatedTimeRemaining.Text = "Total Estimated Time Remaining For All Directories: " + _EstimatedTimeRemainingString;
            // use of the confidence value to vary the label colour would go here
        }

        protected void UpdateMemoryStatsLabel()
        {
            //Memory Use: 47 MiB [Current] / 512MiB [Max] / 298 MiB [Avg] / 256MiB [Med]
            if ((_CurrentMemoryUsageString != "") &&
                (_MaxMemoryUsageString != "") &&
                (_AverageMemoryUsageString != "") &&
                (_MedianMemoryUsageString != ""))
            {
                StringBuilder builder = new StringBuilder();
                builder.Append("Overall Memory Use: ");
                builder.Append(Environment.NewLine);
                builder.Append(_CurrentMemoryUsageString);
                builder.Append(" [Last]");
                builder.Append(Environment.NewLine);
                builder.Append(_MaxMemoryUsageString);
                builder.Append(" [Max]");
                builder.Append(Environment.NewLine);
                builder.Append(_AverageMemoryUsageString);
                builder.Append(" [Avg]");
                builder.Append(Environment.NewLine);
                builder.Append(_MedianMemoryUsageString);
                builder.Append(" [Med]");

                txtMemoryStats.Text = builder.ToString();
            }
            else
            {
                txtMemoryStats.Text = "";
            }
        }

        protected void UpdateProcessorStatsLabel()
        {
            //Processor Use: 25% [Current] / 100% [Max] / 50% [Avg] / 47% [Med]
            if ((_CurrentProcessorUsageString != "") &&
                    (_MaxProcessorUsageString != "") &&
                    (_AverageProcessorUsageString != "") &&
                    (_MedianProcessorUsageString != ""))
            {
                StringBuilder builder = new StringBuilder();
                builder.Append("Overall Processor Use: ");
                builder.Append(Environment.NewLine);
                builder.Append(_CurrentProcessorUsageString);
                builder.Append(" [Last]");
                builder.Append(Environment.NewLine);
                builder.Append(_MaxProcessorUsageString);
                builder.Append(" [Max]");
                builder.Append(Environment.NewLine);
                builder.Append(_AverageProcessorUsageString);
                builder.Append(" [Avg]");
                builder.Append(Environment.NewLine);
                builder.Append(_MedianProcessorUsageString);
                builder.Append(" [Med]");

                txtProcessorStats.Text = builder.ToString();
            }
            else
            {
                txtProcessorStats.Text = "";
            }
        }

        #region Buttons

        private void btnPause_Click(object sender, EventArgs e)
        {
            btnPause.Enabled = false;
            btnResume.Enabled = true;
            _GUIEngineMonitor.Controller.Pause();
        }

        private void btnResume_Click(object sender, EventArgs e)
        {
            btnPause.Enabled = true;
            btnResume.Enabled = false;
            _GUIEngineMonitor.Controller.Resume();
        }

        private void btnStop_Click(object sender, EventArgs e)
        {
            Stop();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        #endregion

        void frmProcessingStatus_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (_CurrentState == ProcessingFormState.Running)
            {
                DialogResult d = MessageBox.Show("Warning: processing activity is still taking place. If you choose to close this window, " +
                    "that activity will be forcibly ended. This may leave the image directories in an inconsistent state, " +
                    "requiring you to perform manual cleanup of their contents. Are you sure you want to close this window? " +
                    "If you are unsure, you should choose No.", "Processing Incomplete", MessageBoxButtons.YesNo, MessageBoxIcon.Warning);
                if (d == DialogResult.Yes)
                {
                    Stop();
                    Thread.Sleep(_SleepLengthWaitForForceStop);
                    CheckForDaVinciProcesses();
                }
                else
                {
                    e.Cancel = true;
                }
            }
            else
            {
                CheckForDaVinciProcesses();
            }
        }

        protected void Stop()
        {
            bool shouldStop = false;
            if (_CurrentState == ProcessingFormState.Running)
            {
                DialogResult d = MessageBox.Show("Warning: processing activity is still taking place. If you choose to stop processing, " +
                    "that activity will be forcibly ended. This may leave the image directories in an inconsistent state, " +
                    "requiring you to perform manual cleanup of their contents. Are you sure you want to stop processing? " +
                    "If you are unsure, you should choose No.", "Processing Incomplete", MessageBoxButtons.YesNo, MessageBoxIcon.Warning);
                if (d == DialogResult.Yes)
                {
                    shouldStop = true;
                }
            }
            else
            {
                shouldStop = true;
            }

            if (shouldStop)
            {
                _CurrentState = ProcessingFormState.Stopped;
                btnPause.Enabled = false;
                btnStop.Text = "Stopping";
                btnStop.Enabled = false;
                btnClose.Enabled = true;
                _GUIEngineMonitor.Controller.ForceStop();
                FailAllStatusDisplays();
                SetAllStatusDisplays(EngineStatusDisplay.EngineDisplayStatus.Failed);
            }
        }

        protected void SetAllStatusDisplays(EngineStatusDisplay.EngineDisplayStatus newStatus)
        {
            for (int i = 0; i < StatusDisplays.Length; i++)
            {
                if (StatusDisplays[i] != null)
                {
                    StatusDisplays[i].DisplayStatus = newStatus;
                }
            }
        }

        protected void FailAllStatusDisplays()
        {
            for (int i = 0; i < StatusDisplays.Length; i++)
            {
                if (StatusDisplays[i] != null)
                {
                    StatusDisplays[i].HasRegisteredError = true;
                }
            }
        }

        protected void CheckForDaVinciProcesses()
        {
            try
            {
                Process[] dvProcesses = ProcessLocator.GetProcesses(TMSBEngineCore.DaVinciProcessNameRegex, true, TMSBEngineControllerConfiguration.BLUExecutionPlatformFromBLDVExecutionPlatform(_GUIEngineMonitor.Controller.EngineControllerConfiguration.ExecutionPlatform));
                if (dvProcesses.Length > 0)
                {
                    DialogResult d = MessageBox.Show("Warning: " + dvProcesses.Length.ToString() + " DaVinci processes are still running on this system. " +
                        "These processes may or may not have been launched by TMSB. Do you want TMSB to forcibly end these processes? " +
                        "Choosing Yes may be dangerous. Choose No if you are unsure.", "DaVinci Processes Detected", MessageBoxButtons.YesNo, MessageBoxIcon.Warning);
                    if (d == DialogResult.Yes)
                    {
                        int successCount = 0;
                        int failureCount = 0;
                        foreach (Process p in dvProcesses)
                        {
                            try
                            {
                                p.Kill();
                                successCount++;
                            }
                            catch (Exception ex)
                            {
                                failureCount++;
                            }
                        }
                        if (failureCount > 0)
                        {
                            MessageBox.Show(successCount.ToString() + " process(es) were successfully killed. " +
                                failureCount.ToString() +
                                " processs(es) could not be killed. You should manually kill these processes, or reboot your computer.",
                                "Error Killing Processes", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                // do nothing for now
            }
        }

        #region Event Handlers

        public void OnImageStarted(object sender, ImageEventArgs a)
        {
            ImageStartedFunction(sender, a);
        }

        delegate void ImageStartedCallback(object sender, ImageEventArgs a);

        protected void ImageStartedFunction(object sender, ImageEventArgs a)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    ImageStartedCallback c = new ImageStartedCallback(ImageStartedFunction);
                    this.Invoke(c, new object[] { sender, a });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                if (_GUIEngineMonitor.Controller.Engines[a.EngineNumber].Engine != null)
                {
                    int activeCores = _GUIEngineMonitor.Controller.Engines[a.EngineNumber].Engine.ActiveEngineCoreCount;
                    if (activeCores > 1)
                    {
                        StatusDisplays[a.EngineNumber].CurrentStatus = "Processing " + activeCores.ToString() + " Images";
                    }
                    else
                    {
                        StatusDisplays[a.EngineNumber].CurrentStatus = "Processing Image '" + a.ImageName + "'";
                    }
                }
            }
        }

        public void OnImageFinished(object sender, ImageFinishedEventArgs a)
        {
            ImageFinishedFunction(sender, a);
        }

        delegate void ImageFinishedCallback(object sender, ImageFinishedEventArgs a);

        protected void ImageFinishedFunction(object sender, ImageFinishedEventArgs a)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    ImageFinishedCallback c = new ImageFinishedCallback(ImageFinishedFunction);
                    this.Invoke(c, new object[] { sender, a });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                if (StatusDisplays[a.EngineNumber].TotalImages < a.TotalImagesInDirectory)
                {
                    StatusDisplays[a.EngineNumber].TotalImages = a.TotalImagesInDirectory;
                }
                //StatusDisplays[a.EngineNumber].ProcessedImages = _GUIEngineMonitor.Controller.Engines[a.EngineNumber].Engine.Tracker.CurrentElementNumber;
                StatusDisplays[a.EngineNumber].ProcessedImages += a.ImageCount;
                //UpdateQueueItemSuccessCount(a.BaseDirectory, _GUIEngineMonitor.Controller.Engines[a.EngineNumber].Engine.Tracker.CurrentElementNumber);
                IncrementQueueItemSuccessCount(a.BaseDirectory, a.ImageCount);

                TimeRemainingInformation tri = _GUIEngineMonitor.Controller.Engines[a.EngineNumber].Engine.Tracker.GetTimeRemainingInformation();

                StatusDisplays[a.EngineNumber].EstimatedTimeRemainingString = tri.EstimatedTimeRemainingLongString;
                StatusDisplays[a.EngineNumber].EstimatedTimeRemainingConfidence = tri.EstimatedTimeRemainingAverageConfidence;

                StatusDisplays[a.EngineNumber].MaxMemoryUsageString = _GUIEngineMonitor.Controller.GetMemoryUsageString(a.MemoryUseTracker.Maximum, true);
                StatusDisplays[a.EngineNumber].MedianMemoryUsageString = _GUIEngineMonitor.Controller.GetMemoryUsageString(a.MemoryUseTracker.GetMedianValue(), true);
                StatusDisplays[a.EngineNumber].AverageMemoryUsageString = _GUIEngineMonitor.Controller.GetMemoryUsageString(a.MemoryUseTracker.GetAverageValue(), true);
                StatusDisplays[a.EngineNumber].CurrentMemoryUsageString = _GUIEngineMonitor.Controller.GetMemoryUsageString(a.MemoryUseTracker.CurrentValue, true);

                StatusDisplays[a.EngineNumber].MaxProcessorUsageString = _GUIEngineMonitor.Controller.GetPercentProcessorTimeString(a.ProcessorTimeTracker.Maximum, true);
                StatusDisplays[a.EngineNumber].MedianProcessorUsageString = _GUIEngineMonitor.Controller.GetPercentProcessorTimeString(a.ProcessorTimeTracker.GetMedianValue(), true);
                StatusDisplays[a.EngineNumber].AverageProcessorUsageString = _GUIEngineMonitor.Controller.GetPercentProcessorTimeString(a.ProcessorTimeTracker.GetAverageValue(), true);
                StatusDisplays[a.EngineNumber].CurrentProcessorUsageString = _GUIEngineMonitor.Controller.GetPercentProcessorTimeString(a.ProcessorTimeTracker.CurrentValue, true);

                if (a.OutputImages.Length > 0)
                {
                    // figure out how many distinct images there are
                    ArrayList distinctImageNames = new ArrayList();
                    for (int i = 0; i < a.OutputImages.Length; i++)
                    {
                        string imageName = Path.GetFileNameWithoutExtension(a.OutputImages[i]);
                        if (!distinctImageNames.Contains(imageName))
                        {
                            distinctImageNames.Add(imageName);
                        }
                    }

                    foreach (string currentImage in distinctImageNames)
                    {
                        ArrayList currentImageList = new ArrayList();
                        for (int i = 0; i < a.OutputImages.Length; i++)
                        {
                            string fullImagePath = a.OutputImages[i];
                            if (currentImage == Path.GetFileNameWithoutExtension(fullImagePath))
                            {
                                if (!currentImageList.Contains(fullImagePath))
                                {
                                    currentImageList.Add(fullImagePath);
                                }
                            }
                        }
                        string[] currentImageArray = (string[])currentImageList.ToArray("".GetType());
                        StatusDisplays[a.EngineNumber].QueueThumbnail(currentImageArray,
                            currentImageArray[GetThumbnailSourceIndex(currentImageArray, _ExtensionPreferenceDefaultFile)],
                            currentImageArray[GetThumbnailSourceIndex(currentImageArray, _ExtensionPreferenceThumbnails)]);
                    }

                    //StatusDisplays[a.EngineNumber].QueueThumbnail(a.OutputImages, a.OutputImages[GetThumbnailSourceIndex(a.OutputImages, _ExtensionPreferenceDefaultFile)], a.OutputImages[GetThumbnailSourceIndex(a.OutputImages, _ExtensionPreferenceThumbnails)]);
                }

                if (_GUIEngineMonitor.Controller.Engines[a.EngineNumber].Engine != null)
                {
                    int activeCores = _GUIEngineMonitor.Controller.Engines[a.EngineNumber].Engine.ActiveEngineCoreCount;
                    if (activeCores > 0)
                    {
                        string t = " Images";
                        if (activeCores == 1)
                        {
                            t = " Image";
                        }
                        StatusDisplays[a.EngineNumber].CurrentStatus = "Processing " + activeCores.ToString() + t;
                    }
                    else
                    {
                        StatusDisplays[a.EngineNumber].CurrentStatus = "Finished Image '" + a.ImageName + "'";
                    }
                    StatusDisplays[a.EngineNumber].UpdateState(_GUIEngineMonitor.Controller.Engines[a.EngineNumber].Engine.CurrentState);
                }

                _TimeEstimationDataExists = true;

                UpdateTotalTimeRemaining();
            }
        }

        protected int GetThumbnailSourceIndex(string[] imagePaths, string[] extensionPreference)
        {
            int sourceIndex = -1;

            for (int i = 0; i < extensionPreference.Length; i++)
            {
                sourceIndex = GetThumbnailSourceIndexIteration(imagePaths, extensionPreference[i]);
                if (sourceIndex >= 0)
                {
                    return sourceIndex;
                }
            }

            return 0;
        }

        protected int GetThumbnailSourceIndexIteration(string[] imagePaths, string fileExtension)
        {
            for (int i = 0; i < imagePaths.Length; i++)
            {
                if (File.Exists(imagePaths[i]))
                {
                    string ext = Path.GetExtension(imagePaths[i]).Trim().ToLower();
                    if (ext == fileExtension)
                    {
                        return i;
                    }
                }
            }
            return -1;
        }

        public void OnDirectoryStarted(object sender, DirectoryProcessingArgs d)
        {
            DirectoryStartedFunction(sender, d);
        }

        delegate void DirectoryStartedCallback(object sender, DirectoryProcessingArgs d);

        protected void DirectoryStartedFunction(object sender, DirectoryProcessingArgs d)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    DirectoryStartedCallback c = new DirectoryStartedCallback(DirectoryStartedFunction);
                    this.Invoke(c, new object[] { sender, d });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                StatusDisplays[d.EngineNumber].ResetState();
                StatusDisplays[d.EngineNumber].DirectoryName = d.DirectoryPath;
                StatusDisplays[d.EngineNumber].CurrentStatus = "Beginning Processing";
                UpdateQueueItemStatus(d.DirectoryPath, ProcessableItemStatus.Processing);
            }
        }

        public void OnDirectoryFinished(object sender, DirectoryProcessingArgs d)
        {
            DirectoryFinishedFunction(sender, d);
        }

        delegate void DirectoryFinishedCallback(object sender, DirectoryProcessingArgs d);

        protected void DirectoryFinishedFunction(object sender, DirectoryProcessingArgs d)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    DirectoryFinishedCallback c = new DirectoryFinishedCallback(DirectoryFinishedFunction);
                    this.Invoke(c, new object[] { sender, d });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                StatusDisplays[d.EngineNumber].CurrentStatus = "Finishing Processing";
            }
        }



        public void OnEngineExecutionComplete(object sender, DirectoryProcessingArgs d)
        {
            EngineExecutionCompleteFunction(sender, d);
        }

        delegate void EngineExecutionCompleteCallback(object sender, DirectoryProcessingArgs d);

        protected void EngineExecutionCompleteFunction(object sender, DirectoryProcessingArgs d)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    EngineExecutionCompleteCallback c = new EngineExecutionCompleteCallback(EngineExecutionCompleteFunction);
                    this.Invoke(c, new object[] { sender, d });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                ProcessedDirectories++;

                UpdateTotalTimeRemaining();

                StatusDisplays[d.EngineNumber].SetCompleted();
                StatusDisplays[d.EngineNumber].CurrentStatus = "Finished Processing";

                StatusDisplays[d.EngineNumber].UpdateState(_GUIEngineMonitor.Controller.Engines[d.EngineNumber].Engine.CurrentState);

                if (StatusDisplays[d.EngineNumber].DisplayStatus == EngineStatusDisplay.EngineDisplayStatus.Normal)
                {
                    StatusDisplays[d.EngineNumber].DisplayStatus = EngineStatusDisplay.EngineDisplayStatus.Finished;
                }

                if (pqgQueue.Items.Contains(d.DirectoryPath))
                {
                    ProcessableItem pi = pqgQueue.Items[d.DirectoryPath];

                    if ((pi.Status == ProcessableItemStatus.Error) || (StatusDisplays[d.EngineNumber].DisplayStatus == EngineStatusDisplay.EngineDisplayStatus.Failed))
                    {
                        UpdateQueueItemStatus(d.DirectoryPath, ProcessableItemStatus.Failed);
                    }
                    else
                    {
                        UpdateQueueItemStatus(d.DirectoryPath, ProcessableItemStatus.Finished);
                    }
                }
            }
        }

        protected void UpdateTotalTimeRemaining()
        {
            // Time estimation is too buggy for now
            return;

            try
            {
                Thread updateThread = new Thread(new ThreadStart(UpdateTotalTimeRemainingFunction));
                updateThread.Start();
            }
            catch (Exception ex)
            {
                // this is not a critical function, so fail silently
            }
        }

        delegate void UpdateTotalTimeRemainingCallback();

        protected void UpdateTotalTimeRemainingFunction()
        {
            if (this.InvokeRequired)
            {
                try
                {
                    UpdateTotalTimeRemainingCallback c = new UpdateTotalTimeRemainingCallback(UpdateTotalTimeRemainingFunction);
                    this.Invoke(c);
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                TimeSpan elapsed = DateTime.Now - _LastTimeEstimationUpdate;
                if (elapsed.TotalMilliseconds < _MinimumTimeEstimationInterval)
                {
                    // do nothing
                }
                else
                {
                    bool shouldUpdate = true;
                    lock (UpdatingTotalTimeRemainingLockObject)
                    {
                        if ((_UpdatingTotalTimeRemaining) || (!_TimeEstimationDataExists))
                        {
                            shouldUpdate = false;
                        }
                        else
                        {
                            _UpdatingTotalTimeRemaining = true;
                        }
                    }
                    if (shouldUpdate)
                    {
                        string previousETRS = EstimatedTimeRemainingString;
                        float previousETRC = EstimatedTimeRemainingConfidence;
                        try
                        {
                            double multiplier = _GUIEngineMonitor.GetTimeRemainingMultiplier();
                            TimeRemainingInformation tri = _GUIEngineMonitor.MetaTracker.GetTimeRemainingInformation(multiplier);

                            TimeSpan longestAverage = TimeSpan.MinValue;
                            TimeSpan longestMedian = TimeSpan.MinValue;
                            TimeSpan longestMax = TimeSpan.MinValue;

                            string ets = tri.EstimatedTimeRemainingLongString;
                            float etc = tri.EstimatedTimeRemainingAverageConfidence;

                            bool recalculate = false;
                            for (int i = 0; i < _GUIEngineMonitor.Controller.Engines.Length; i++)
                            {
                                if ((_GUIEngineMonitor.Controller.Engines[i].Engine != null) && (_GUIEngineMonitor.Controller.Engines[i].Engine.CurrentState == TMSBEngineState.Running))
                                {
                                    TimeRemainingInformation tri2 = _GUIEngineMonitor.Controller.Engines[i].Engine.Tracker.GetTimeRemainingInformation();
                                    if (tri2.EstimatedTimeRemainingAverage > longestAverage)
                                    {
                                        longestAverage = tri2.EstimatedTimeRemainingAverage;
                                    }
                                    if (tri2.EstimatedTimeRemainingMedian > longestMedian)
                                    {
                                        longestMedian = tri2.EstimatedTimeRemainingMedian;
                                    }
                                    if (tri2.EstimatedTimeRemainingMax > longestMax)
                                    {
                                        longestMax = tri2.EstimatedTimeRemainingMax;
                                    }
                                }
                            }

                            // make sure the estimate is never longer than the longest remaining engine (assuming no more items in queue)
                            if (_GUIEngineMonitor.MetaTracker.GetRemainingElementCount() <= _GUIEngineMonitor.Controller.EngineControllerConfiguration.EngineThreadCount)
                            {
                                //if ((longestAverage < tri.EstimatedTimeRemainingAverage)
                                //    || (longestMedian < tri.EstimatedTimeRemainingMedian)
                                //    || (longestMax < tri.EstimatedTimeRemainingMax))
                                //{
                                //    needsRecalculation = true;
                                //}
                                if ((longestMax < tri.EstimatedTimeRemainingMax))
                                {
                                    recalculate = true;
                                }
                            }
                            else
                            {
                                // make sure the estimate is at least as long as the longest current estimate for a single engine (assuming active processing)
                                if ((longestMax > tri.EstimatedTimeRemainingMax))
                                {
                                    if (tri.EstimatedTimeRemainingAverage > longestAverage)
                                    {
                                        longestAverage = tri.EstimatedTimeRemainingAverage;
                                    }
                                    if (tri.EstimatedTimeRemainingMedian > longestMedian)
                                    {
                                        longestMedian = tri.EstimatedTimeRemainingMedian;
                                    }
                                    if (tri.EstimatedTimeRemainingMax > longestMax)
                                    {
                                        longestMax = tri.EstimatedTimeRemainingMax;
                                    }
                                    recalculate = true;
                                }
                            }

                            if (recalculate)
                            {
                                ets = _GUIEngineMonitor.MetaTracker.GetEstimatedTimeRemainingLongString(longestAverage, longestMedian, longestMax);
                                //etc = 1.0f;
                            }

                            EstimatedTimeRemainingString = ets;
                            EstimatedTimeRemainingConfidence = etc;
                        }
                        catch (Exception ex)
                        {
                            EstimatedTimeRemainingString = previousETRS;
                            EstimatedTimeRemainingConfidence = previousETRC;
                        }

                        _LastTimeEstimationUpdate = DateTime.Now;

                        _UpdatingTotalTimeRemaining = false;
                    }
                }
            }
        }

        //protected string GetDirectoryErrorSummaryText(string directoryPath, MessageCollection mc)
        //{
        //    string statusMessage = "One or more error messages";
        //    if (mc.Contains(directoryPath))
        //    {
        //        KeyedMessageList kml = mc[directoryPath];
        //        if (kml.Messages.Length == 1)
        //        {
        //            statusMessage = "1 error message";
        //        }
        //        if (kml.Messages.Length > 1)
        //        {
        //            statusMessage = kml.Messages.Length.ToString() + " error messages";
        //        }
        //    }
        //    return statusMessage;
        //}

        //protected string GetDirectorySuccessSummaryText(string directoryPath, int numberGenerated)
        //{
        //    string statusMessage = "No images generated";

        //    if (numberGenerated == 1)
        //    {
        //        statusMessage = "1 image generated";
        //    }
        //    if (numberGenerated > 1)
        //    {
        //        statusMessage = numberGenerated.ToString() + " images generated";
        //    }

        //    return statusMessage;
        //}

        protected void UpdateOverallPerformanceStats()
        {
            MaxMemoryUsageString = _GUIEngineMonitor.Controller.GetMemoryUsageString(_GUIEngineMonitor.Controller.MemoryUseTracker.Maximum, true);
            MedianMemoryUsageString = _GUIEngineMonitor.Controller.GetMemoryUsageString(_GUIEngineMonitor.Controller.MemoryUseTracker.GetMedianValue(), true);
            AverageMemoryUsageString = _GUIEngineMonitor.Controller.GetMemoryUsageString(_GUIEngineMonitor.Controller.MemoryUseTracker.GetAverageValue(), true);
            CurrentMemoryUsageString = _GUIEngineMonitor.Controller.GetMemoryUsageString(_GUIEngineMonitor.Controller.MemoryUseTracker.CurrentValue, true);

            MaxProcessorUsageString = _GUIEngineMonitor.Controller.GetPercentProcessorTimeString(_GUIEngineMonitor.Controller.ProcessorTimeTracker.Maximum, true);
            MedianProcessorUsageString = _GUIEngineMonitor.Controller.GetPercentProcessorTimeString(_GUIEngineMonitor.Controller.ProcessorTimeTracker.GetMedianValue(), true);
            AverageProcessorUsageString = _GUIEngineMonitor.Controller.GetPercentProcessorTimeString(_GUIEngineMonitor.Controller.ProcessorTimeTracker.GetAverageValue(), true);
            CurrentProcessorUsageString = _GUIEngineMonitor.Controller.GetPercentProcessorTimeString(_GUIEngineMonitor.Controller.ProcessorTimeTracker.CurrentValue, true);
            this.Invalidate();
        }


        public void OnDirectoryProcessingError(object sender, DirectoryProcessingArgs d)
        {
            //StatusDisplays[d.EngineNumber].IsInErrorState = true;
            DirectoryProcessingErrorFunction(sender, d);
        }

        delegate void DirectoryProcessingErrorCallback(object sender, DirectoryProcessingArgs d);

        protected void DirectoryProcessingErrorFunction(object sender, DirectoryProcessingArgs d)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    DirectoryProcessingErrorCallback c = new DirectoryProcessingErrorCallback(DirectoryProcessingErrorFunction);
                    this.Invoke(c, new object[] { sender, d });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                StatusDisplays[d.EngineNumber].DisplayStatus = EngineStatusDisplay.EngineDisplayStatus.Failed;
                StatusDisplays[d.EngineNumber].CurrentStatus = "Error";
                UpdateQueueItemStatus(d.DirectoryPath, ProcessableItemStatus.Error);
                IncrementQueueItemFailureCount(d.DirectoryPath, 1);
            }
        }


        public void OnEngineControllerExecutionComplete(object sender, EngineManagementEventArgs e)
        {
            //ProcessedDirectories = TotalDirectories;
            //btnClose.Enabled = true;
            EngineControllerExecutionCompleteFunction(sender, e);
        }

        delegate void EngineControllerExecutionCompleteCallback(object sender, EngineManagementEventArgs e);

        protected void EngineControllerExecutionCompleteFunction(object sender, EngineManagementEventArgs e)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    EngineControllerExecutionCompleteCallback c = new EngineControllerExecutionCompleteCallback(EngineControllerExecutionCompleteFunction);
                    this.Invoke(c, new object[] { sender, e });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                // catch any stragglers
                for (int i = 0; i < StatusDisplays.Length; i++)
                {
                    StatusDisplays[i].SetCompleted();
                }
                ProcessedDirectories = TotalDirectories;
                _CurrentState = ProcessingFormState.Stopped;
                btnClose.Enabled = true;
                btnStop.Enabled = false;
                btnStop.Text = "Stop";
                btnPause.Enabled = false;
                btnResume.Enabled = false;
                if (_ParentForm != null)
                {
                    pqgQueue.SortBy(ProcessableItemQueueGrid.SortColumn.Position);
                    _ParentForm.UpdateQueueItems(pqgQueue.Items);
                }
                if (_GUIEngineMonitor.ErrorMessages.Count > 0)
                {
                    frmErrorSummary errorDisplay = new frmErrorSummary(_GUIEngineMonitor.ErrorMessages);
                    errorDisplay.Show();
                }
                _UpdateTimer.Enabled = false;
            }
        }

        public void OnScriptBeingGenerated(object sender, ThreadedEventArgs d)
        {
            ScriptBeingGeneratedFunction(sender, d);
        }

        delegate void ScriptBeingGeneratedCallback(object sender, ThreadedEventArgs d);

        protected void ScriptBeingGeneratedFunction(object sender, ThreadedEventArgs d)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    ScriptBeingGeneratedCallback c = new ScriptBeingGeneratedCallback(ScriptBeingGeneratedFunction);
                    this.Invoke(c, new object[] { sender, d });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                StatusDisplays[d.EngineNumber].CurrentStatus = "Generating Script";
            }
        }

        public void OnScriptGenerated(object sender, CubeScriptGeneratedEventArgs d)
        {
            ScriptGeneratedFunction(sender, d);
        }

        delegate void ScriptGeneratedCallback(object sender, CubeScriptGeneratedEventArgs d);

        protected void ScriptGeneratedFunction(object sender, CubeScriptGeneratedEventArgs d)
        {
            if (this.InvokeRequired)
            {
                try
                {
                    ScriptGeneratedCallback c = new ScriptGeneratedCallback(ScriptGeneratedFunction);
                    this.Invoke(c, new object[] { sender, d });
                }
                catch (Exception ex)
                {
                    // do nothing
                }
            }
            else
            {
                StatusDisplays[d.EngineNumber].CurrentStatus = "Script Generated";
            }
        }

        #endregion

        private void pnlEngineStatus_Resize(object sender, EventArgs e)
        {
            UpdateStatusDisplayPanel();
        }

        protected void BuildQueue()
        {
            for (int i = 0; i < _GUIEngineMonitor.Controller.BaseDirectories.Length; i++)
            {
                ProcessableItem pi = new ProcessableItem(_GUIEngineMonitor.Controller.BaseDirectories[i], ProcessableItemStatus.Queued, "");
                pqgQueue.Items.Add(pi);
            }
            //pqgQueue.SetSortColumn("Status", ListSortDirection.Ascending);
            pqgQueue.SortBy(ProcessableItemQueueGrid.SortColumn.Status);
        }

        protected void UpdateQueueItemSuccessCount(string directoryName, int newCount)
        {
            if (pqgQueue.Items.Contains(directoryName))
            {
                pqgQueue.Items[directoryName].SuccessCount = newCount;
            }
            UpdateQueueItemStatusMessages();
        }

        protected void UpdateQueueItemFailureCount(string directoryName, int newCount)
        {
            if (pqgQueue.Items.Contains(directoryName))
            {
                pqgQueue.Items[directoryName].FailureCount = newCount;
            }
            UpdateQueueItemStatusMessages();
        }

        protected void IncrementQueueItemSuccessCount(string directoryName, int incrementAmount)
        {
            if (pqgQueue.Items.Contains(directoryName))
            {
                pqgQueue.Items[directoryName].SuccessCount += incrementAmount;
            }
            UpdateQueueItemStatusMessages();
        }

        protected void IncrementQueueItemFailureCount(string directoryName, int incrementAmount)
        {
            if (pqgQueue.Items.Contains(directoryName))
            {
                pqgQueue.Items[directoryName].FailureCount += incrementAmount;
            }
            UpdateQueueItemStatusMessages();
        }

        protected void UpdateQueueItemStatus(string directoryName, ProcessableItemStatus newStatus)
        {
            UpdateQueueItemStatus(directoryName, newStatus, "");
        }

        protected void UpdateQueueItemStatus(string directoryName, ProcessableItemStatus newStatus, string message)
        {
            if (pqgQueue.Items.Contains(directoryName))
            {
                pqgQueue.Items[directoryName].Status = newStatus;
                if (message != "")
                {
                    pqgQueue.Items[directoryName].Information = message;
                }
                UpdateQueueItemStatusMessages();
            }
        }

        protected void UpdateQueueItemStatusMessages()
        {
            for (int i = 0; i < pqgQueue.Items.Count; i++)
            {
                ProcessableItem pi = pqgQueue.Items[i];

                StringBuilder builder = new StringBuilder();
                string statusMessage = "No images generated";
                if (pi.SuccessCount == 1)
                {
                    statusMessage = "1 image generated";
                }
                if (pi.SuccessCount > 1)
                {
                    statusMessage = pi.SuccessCount.ToString() + " images generated";
                }
                builder.Append(statusMessage);

                if ((pi.FailureCount > 0) || (pi.Status == ProcessableItemStatus.Error) ||
                    (pi.Status == ProcessableItemStatus.Failed))
                {
                    statusMessage = "One or more error messages";
                    if (pi.FailureCount == 1)
                    {
                        statusMessage = pi.FailureCount.ToString() + " error message";
                    }
                    if (pi.FailureCount > 1)
                    {
                        statusMessage = pi.FailureCount.ToString() + " error messages";
                    }
                    builder.Append(", ");
                    builder.Append(statusMessage);
                }
                pqgQueue.Items[i].SetDefaultInformation(builder.ToString());
            }
            pqgQueue.ForceQueueUpdate();
        }

        public void OnWaitingForProcessingThread(object sender, EngineManagementEventArgs e)
        {
            if (StatusDisplays[e.EngineNumber].ProcessCount == 0)
            {
                StatusDisplays[e.EngineNumber].CurrentStatus = "Waiting For An Available Processing Thread";
            }
        }
    }
}